# Copyright (c) 2019 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cinder Volume driver for Fujitsu ETERNUS DX S3 series."""

import six

from cinder.i18n import _
from cinder import ssh_utils
from cinder.volume.drivers.fujitsu.eternus_dx import constants as CONSTANTS


class FJDXCLI(object):
    """ETERNUS CLI Code."""

    def __init__(self, user, storage_ip, password=None, keyfile=None):
        """Constructor."""
        self.user = user
        self.storage_ip = storage_ip
        if password and keyfile:
            raise Exception(_('can not specify both password and keyfile'))

        self.use_ipv6 = False
        if storage_ip.find(':') != -1:
            self.use_ipv6 = True

        if password:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              password=password, max_size=2)

        if keyfile:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              privatekey=keyfile, max_size=2)

        self.ce_support = False
        self.CMD_dic = {
            'check_user_role': self._check_user_role,
            'expand_volume': self._expand_volume,
            'format_volume': self._format_volume,
            'format_tpv': self._format_tpv,
            'set_volume_qos': self._set_volume_qos,
            'show_volume_progress': self._show_volume_progress,
            'show_tpv_progress': self._show_tpv_progress,
            'show_host_wwn_names': self._show_host_wwn_names,
            'show_host_iscsi_names': self._show_host_iscsi_names,
            'create_host_wwn_name': self._create_host_wwn_name,
            'create_host_iscsi_name': self._create_host_iscsi_name,
            'create_affinity_group': self._create_affinity_group,
            'delete_affinity_group': self._delete_affinity_group,
            'show_affinity_group': self._show_affinity_group,
            'show_iscsi_parameters': self._show_iscsi_parameters,
            'show_fc_parameters': self._show_fc_parameters,
            'set_affinity_group': self._set_affinity_group,
            'set_host_affinity': self._set_host_affinity,
            'show_pool_provision': self._show_pool_provision,
            'show_host_affinity': self._show_host_affinity,
            'discover_host_wwn_names': self._discover_wwn_names,
            'release_host_affinity': self._release_host_affinity,
            'show_tfo_groups': self._show_tfo_groups,
            'show_boxid': self._show_boxid,
            'show_copy_sessions': self._show_copy_sessions,
            'forced_tfo_group_activate': (
                self._forced_tfo_group_activate),
            'set_tfog_pair': self._set_tfog_pair,
            'show_model_supplement_info': (
                self._show_model_supplement_info),
            'show_qos_bandwidth_limit': self._show_qos_bandwidth_limit,
            'set_qos_bandwidth_limit': self._set_qos_bandwidth_limit,
            'show_volume_qos': self._show_volume_qos,
            'set_qos_mode': self._set_qos_mode,
            'show_enclosure_status': self._show_enclosure_status
        }

        self.SMIS_dic = {
            '0000': '0',  # Success.
            '0060': '32787',  # The device is in busy state.
            '0100': '4097'
        }  # Size not supported.

    def done(self, command, **option):
        """Done ETERNUS CLI."""
        func = self.CMD_dic.get(command, self._default_func)
        return func(**option)

    def _exec_cli(self, cmd, StrictHostKeyChecking=True, **option):

        exec_cmdline = cmd + self._get_option(**option)
        stdoutdata = self._exec_cli_with_eternus(exec_cmdline)
        output = []
        message = []
        stdoutlist = stdoutdata.split('\r\n')
        output_header = ""

        for no, outline in enumerate(stdoutlist):
            if len(outline) <= 0 or outline is None:
                continue

            if not output_header.endswith(exec_cmdline):
                output_header += outline
                continue

            if 0 <= outline.find('Error'):
                raise Exception(_("Output: %(outline)s: "
                                  "Command: %(cmdline)s")
                                % {'outline': outline,
                                   'cmdline': exec_cmdline})

            if not self._is_status(outline):
                continue

            status = int(outline, 16)
            lineno = no + 1
            break
        else:
            raise Exception(_(
                "Invalid CLI output: %(exec_cmdline)s, %(stdoutlist)s")
                % {'exec_cmdline': exec_cmdline,
                   'stdoutlist': stdoutlist})

        if status == 0:
            rc = '0'
            for outline in stdoutlist[lineno:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                message.append(outline)
        else:
            code = stdoutlist[lineno]
            for outline in stdoutlist[lineno + 1:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                output.append(outline)

            rc, message = self._create_error_message(code, output)

        return {'result': 0, 'rc': rc, 'message': message}

    def _exec_cli_with_eternus(self, exec_cmdline):
        """Execute CLI command with arguments."""
        ssh = None
        try:
            ssh = self.ssh_pool.get()
            chan = ssh.invoke_shell()
            chan.send(exec_cmdline + '\n')
            stdoutdata = ''
            while True:
                stdoutdata += str(chan.recv(65535))
                # CLI command end with 'CLI>'.
                if stdoutdata == '\r\nCLI> ':
                    continue
                if (stdoutdata[len(stdoutdata) - 5: len(stdoutdata) - 1] ==
                        'CLI>'):
                    break
        except Exception as e:
            raise Exception(_("Execute CLI "
                              "command error. Error: %s") % six.text_type(e))
        finally:
            if ssh:
                self.ssh_pool.put(ssh)
                self.ssh_pool.remove(ssh)
        return stdoutdata

    def _create_error_message(self, code, msg):
        """Create error code and message using arguements."""
        message = None
        if code in self.SMIS_dic:
            rc = self.SMIS_dic[code]
        else:
            rc = 'E' + code

            if rc in CONSTANTS.CLIRETCODE_dic:
                message = CONSTANTS.CLIRETCODE_dic[rc]
            else:
                message = msg

        return rc, message

    @staticmethod
    def _is_status(value):
        """Check whether input value is status value or not."""
        try:
            if len(value) != 2:
                return False

            int(value, 16)
            int(value[0], 16)
            int(value[1], 16)

            return True
        except ValueError:
            return False

    @staticmethod
    def _get_option(**option):
        """Create option strings from dictionary."""
        ret = ""
        for key, value in six.iteritems(option):
            ret += " -%(key)s %(value)s" % {'key': key, 'value': value}
        return ret

    def _default_func(self, **option):
        """Default function."""
        raise Exception(_("Invalid function is specified"))

    def _check_user_role(self, **option):
        """Check user role."""
        try:
            output = self._exec_cli("show users",
                                    StrictHostKeyChecking=False,
                                    **option)
            # Return error.
            rc = output['rc']
            if rc != "0":
                return output

            userlist = output.get('message')
            role = None
            for userinfo in userlist:
                username = userinfo.split('\t')[0]
                if username == self.user:
                    role = userinfo.split('\t')[1]
                    break

            output['message'] = role
        except Exception as ex:
            if 'show users' in six.text_type(ex):
                msg = ("Specified user(%s) does not have Software role"
                       % self.user)
            elif 'Error connecting' in six.text_type(ex):
                msg = (six.text_type(ex)[34:] +
                       ', Please check fujitsu_private_key_path or .xml file')
            else:
                msg = six.text_type(ex)
            output = {
                'result': 0,
                'rc': '4',
                'message': msg
            }
        return output

    def _expand_volume(self, **option):
        """Exec expand volume."""
        return self._exec_cli("expand volume", **option)

    def _format_volume(self, **option):
        """Exec format volume."""
        return self._exec_cli("format volume", **option)

    def _format_tpv(self, **option):
        """Exec format tpv."""
        thick_option = {
            'volume-name': option['volume-name'],
            'allocation': 'thick'
        }

        output = self._exec_cli("set volume-parameters", **thick_option)
        rc = output['rc']
        if rc != "0":
            return output

        return self._exec_cli("format volume", **option)

    def _set_volume_qos(self, **option):
        """Exec set volume-qos."""
        return self._exec_cli("set volume-qos", **option)

    def _show_volume_progress(self, **option):
        """Exec show vollume-progress."""
        output = self._exec_cli("show volume-progress", **option)

        # Return error
        rc = output['rc']
        if rc != "0":
            return output

        # Get remain time.
        message = output['message']
        volume_count = message[0]
        if volume_count == "00":
            # Format completion.
            output['message'] = '0'
            return output
        else:
            remain_minute = 1
            remain_time_str = message[1].split("\t")[9]

            if remain_time_str != "FFFFFFFF":
                remain_minute = int(remain_time_str, 16) + 1

            output['message'] = str(remain_minute * 60)

        return output

    def _show_tpv_progress(self, **option):
        """Exec show tpv-progress."""
        output = self._exec_cli("show thin-pro-pools", **option)
        rc = output['rc']
        if rc != "0":
            return output

        message = output['message']
        pool_count = int(message[1], 16)
        raid_count_lineno = 2 + pool_count
        raid_count = int(message[raid_count_lineno], 16)

        for num in range(raid_count):
            raid_info = message[raid_count_lineno + num + 1]
            raid_status = int(raid_info.split('\t')[4], 16)
            cnfg_lu_status_bit_qf = raid_status & 0x0100

            if cnfg_lu_status_bit_qf != 0:
                output['message'] = '600'
                break
        else:
            output['message'] = '0'

        return output

    def _show_host_wwn_names(self, **option):
        """Get host wwn names."""
        try:
            output = self._exec_cli("show host-wwn-names", **option)

            # Return error.
            rc = output['rc']

            if rc != "0":
                return output

            hostdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                hostdatalist.append({'Host Num': int(clidata[0], 16),
                                     'Host Name': clidata[2]})

            output['message'] = hostdatalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host wwn names error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_host_iscsi_names(self, **option):
        """Get host iscsi names."""
        try:
            output = self._exec_cli("show host-iscsi-names", **option)

            # Return error.
            rc = output['rc']

            if rc != "0":
                return output

            hostdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                hostdatalist.append({'Host Num': int(clidata[0], 16),
                                     'Host Name': clidata[5],
                                     'IP Version': clidata[6]})

            output['message'] = hostdatalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host iscsi names error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_affinity_group(self, **option):
        """Get affinity group information."""
        try:
            output = self._exec_cli("show affinity-group", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[3:]:
                clidata = clidataline.split('\t')
                datalist.append({'Lun Num': int(clidata[0], 16),
                                 'Lun Name': clidata[2]})

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host iscsi names error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_pool_provision(self, **option):
        """Get TPP provision capacity information."""
        try:
            output = self._exec_cli("show volumes", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')

            data = 0
            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                if clidata[0] == 'FFFF':
                    break
                data += int(clidata[7], 16)
            # MB unit.
            provision = data / 2048

            output['message'] = provision
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show pool provision capacity error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_iscsi_parameters(self, **option):
        """Get iscsi port information."""

        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show iscsi-parameters", **option)

            rc = output['rc']

            if rc != "0":
                return output

            ipv4_data_list = []
            ipv6_data_list = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                tmpport = clidata[0] + clidata[1]
                port_str = self._get_port(clidata[0], clidata[1])
                if self.use_ipv6:
                    if clidata[31] != '::':
                        ipv6_data_list.append({'port num': tmpport,
                                               'port iqn': clidata[4],
                                               'port ip': '[' + clidata[31] +
                                                          ']',
                                               'port str': port_str})
                else:
                    if clidata[10] != '0.0.0.0':
                        ipv4_data_list.append({'port num': tmpport,
                                               'port iqn': clidata[4],
                                               'port ip': clidata[10],
                                               'port str': port_str})

            if self.use_ipv6:
                output['message'] = ipv6_data_list
            else:
                output['message'] = ipv4_data_list
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show iscsi-parameters error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_fc_parameters(self, **option):
        """Get fc port information."""
        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show fc-parameters", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                tmpport = clidata[0] + clidata[1]
                port_str = self._get_port(clidata[0], clidata[1])
                datalist.append({'port num': tmpport,
                                 'port wwpn': clidata[18],
                                 'port str': port_str})

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show fc-parameters error: %s"
                           % six.text_type(ex)
            }

        return output

    def _get_port(self, ca_module_id, port_number):
        if self.ce_support:
            return self._get_port_str_v2(ca_module_id, port_number)
        else:
            return self._get_port_str(ca_module_id, port_number)

    @staticmethod
    def _get_port_str(ca_module_id, port_number):
        cm = int(ca_module_id[1], 16) % 8
        ca = int(ca_module_id[0], 16) % 4
        port = ((int(ca_module_id[1], 16) // 4) // 2) * 2 + (
            int(port_number[1], 16) % 2)
        temp_str = str(hex(cm)) + str(hex(ca)) + str(hex(port))
        return temp_str.replace('0x', '').upper()

    @staticmethod
    def _get_port_str_v2(ca_module_id, port_number):
        # Port_str format: CE + CM + CA + PORT
        # CE: 0,1,2,3,4,5,6,7,8,9,10,11
        # CM: 0,1
        # CA: 0,1,2,3
        # PORT: 0,1,2,3
        base = 4
        ce = ((int(ca_module_id[1], 16) % 8) // 2) + (
            ((int(ca_module_id[0], 16) - base) // 4) * 4)
        cm = int(ca_module_id[1], 16) % 2
        ca = int(ca_module_id[0], 16) % 4
        port = ((int(ca_module_id[1], 16) // 4) // 2) * 2 + (
            int(port_number[1], 16) % 2)
        temp_str = str(hex(ce)) + str(hex(cm)) + str(hex(ca)) + str(hex(port))
        return temp_str.replace('0x', '').upper()

    def _create_host_wwn_name(self, **option):
        """Create host wwn name."""
        return self._exec_cli("create host-wwn-name", **option)

    def _create_host_iscsi_name(self, **option):
        """Create host iscsi name."""
        return self._exec_cli("create host-iscsi-name", **option)

    def _create_affinity_group(self, **option):
        """Create affinity group."""
        return self._exec_cli("create affinity-group", **option)

    def _delete_affinity_group(self, **option):
        """Delete affinity group."""
        return self._exec_cli("delete affinity-group", **option)

    def _set_host_affinity(self, **option):
        """Set host affinity."""
        return self._exec_cli("set host-affinity", **option)

    def _set_affinity_group(self, **option):
        """Set affinity group."""
        return self._exec_cli("set affinity-group", **option)

    def _show_host_affinity(self, **option):
        """Get host affinity group information."""
        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show host-affinity", **option)

            rc = output['rc']

            if rc != "0":
                return output

            # [{Port:xxx,Host Num:xxx,Host Name:xxx,AG Num:xxx,AG Name:xxx}].
            datalist = []
            clidatalist = output.get('message')

            port_str = None
            for line in clidatalist:
                infos = line.split()
                if len(infos) == 2:
                    port_str = self._get_port(
                        infos[0][-4:-2], infos[0][-2:])
                if len(infos) == 7 or len(infos) == 6:
                    host_ag = {
                        'Port': port_str,
                        'Host Num': int(infos[0], 16),
                        'Host Name': infos[1],
                        'AG Num': int(infos[2], 16),
                        'AG Name': infos[3]
                    }
                    datalist.append(host_ag)

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "discover host-wwn-names: %s"
                           % six.text_type(ex)
            }

        return output

    def _discover_wwn_names(self, **option):
        """Get FC zone info on ETERNUS."""
        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("discover host-wwn-names", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = {}  # {Host WWPN: [port1, port2]}.
            clidatalist = output.get('message')

            port_str = None
            for idx, line in enumerate(clidatalist):
                infos = line.split()
                if len(infos) == 3:
                    port_str = self._get_port(infos[0], infos[1])
                    port_idx = idx
                    continue
                if port_str and idx != port_idx + 1:
                    wwpn = infos[0]
                    temp_list = datalist.get(wwpn, [])
                    temp_list.append(port_str)
                    datalist[wwpn] = temp_list

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "discover host-wwn-names: %s"
                           % six.text_type(ex)
            }

        return output

    def _release_host_affinity(self, **option):
        """Release host affinity."""
        return self._exec_cli("release host-affinity", **option)

    def _show_tfo_groups(self, **option):
        """Show tfo groups."""
        try:
            local_ce_support = option['local_ce_support']
            remote_ce_support = option['remote_ce_support']
            option.pop('local_ce_support')
            option.pop('remote_ce_support')
            output = self._exec_cli("show tfo-groups", **option)

            rc = output['rc']

            if rc != "0":
                return output

            tfog_list = []
            clidatalist = output.get('message')
            datalist = clidatalist[1:]

            for index, item in enumerate(datalist):
                if len(item) == 4:
                    tfog_port_list = []
                    tmp = []
                    tfog_dict = {}
                    # Get TFO group information.
                    tfog_info = datalist[index - 1]
                    tfog_line = tfog_info.strip().split('\t')
                    tfog_dict['no'] = tfog_line[0]
                    tfog_dict['name'] = tfog_line[1]
                    tfog_dict['type'] = tfog_line[2]
                    tfog_dict['status'] = tfog_line[4]
                    tfog_dict['phase'] = tfog_line[5]
                    tfog_dict['condition'] = tfog_line[6]
                    tfog_dict['pair_boxid'] = tfog_line[8]

                    # Get each port-pair information.
                    tmp = [
                        datalist[index + i + 1] for i in range(int(item))
                    ]
                    for sub_tmp in tmp:
                        tfog_port_dict = {}
                        tfog_ports = sub_tmp.strip().split('\t')
                        if local_ce_support:
                            tfog_port_dict['own_port'] = tfog_ports[0]
                        else:
                            tfog_port_dict['own_port'] = tfog_ports[0][-3:]
                        # Local storage only has one remote storage.
                        if remote_ce_support:
                            tfog_port_dict['pair_port'] = tfog_ports[1]
                        else:
                            tfog_port_dict['pair_port'] = tfog_ports[1][-3:]
                        if tfog_ports[2] == '02':
                            tfog_port_dict['port_prtcl'] = 'iSCSI'
                        else:
                            tfog_port_dict['port_prtcl'] = 'FC'
                        tfog_port_list.append(tfog_port_dict)

                    tfog_dict['tfog_port_list'] = tfog_port_list
                    tfog_list.append(tfog_dict)

            output['message'] = tfog_list
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show tfo group error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_boxid(self, **option):
        """Show boxid."""
        try:
            output = self._exec_cli("show boxid", **option)

            rc = output['rc']

            if rc != "0":
                return output

            output['message'] = output.get('message')[0]
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show boxid error: %s"
                           % six.text_type(ex)
            }

        return output

    def _forced_tfo_group_activate(self, **option):
        """Failover or Failback"""
        return self._exec_cli("forced tfo-group-activate", **option)

    def _set_tfog_pair(self, **option):
        """set tfog pair and create copy sessions"""
        return self._exec_cli("set tfo-pair", **option)

    def _show_model_supplement_info(self, **option):
        """Show model supplement info."""
        try:
            output = self._exec_cli("show model-supplement-info", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')
            infos = clidatalist[0].split()
            output['message'] = {
                'machine_info': infos[0],
                'model_info': infos[1]}

        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show model-supplement-info error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_copy_sessions(self, **option):
        """Get copy sessions."""
        try:
            output = self._exec_cli("show copy-sessions", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            cpsdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                # Eg. 0b10010000 0b11000000
                if bin(int(clidata[3], 16) & 16) != 0:
                    temp_type = 'Snap'
                elif bin(int(clidata[3], 16) & 64) != 0:
                    temp_type = 'Snap+'
                else:
                    temp_type = 'Other'
                cpsdatalist.append({'Source Num': int(clidata[13], 16),
                                    'Dest Num': int(clidata[14], 16),
                                    'Type': temp_type})

            output['message'] = cpsdatalist
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show copy sessions error: %s"
                                 % six.text_type(ex)}

        return output

    def _show_qos_bandwidth_limit(self, **option):
        """Get qos bandwidth limit."""
        try:
            output = self._exec_cli("show qos-bandwidth-limit", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            qoslist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                qoslist.append({'total_limit': int(clidata[0], 16),
                                'total_iops_sec': int(clidata[1], 16),
                                'total_bytes_sec': int(clidata[2], 16),
                                'read_limit': int(clidata[0], 16),
                                'read_iops_sec': int(clidata[3], 16),
                                'read_bytes_sec': int(clidata[4], 16),
                                'write_limit': int(clidata[0], 16),
                                'write_iops_sec': int(clidata[5], 16),
                                'write_bytes_sec': int(clidata[6], 16)})

            output['message'] = qoslist

        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show qos bandwidth limit error: %s"
                                 % six.text_type(ex)}

        return output

    def _set_qos_bandwidth_limit(self, **option):
        """Set qos bandwidth limit"""
        return self._exec_cli("set qos-bandwidth-limit", **option)

    def _show_volume_qos(self, **option):
        """Get volumes with qos."""
        try:
            output = self._exec_cli("show volume-qos", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            vqosdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                vqosdatalist.append({'total_limit': int(clidata[2], 16),
                                     'read_limit': int(clidata[3], 16),
                                     'write_limit': int(clidata[4], 16)})

            output['message'] = vqosdatalist

        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show volume qos error: %s"
                                 % six.text_type(ex)}

        return output

    def _set_qos_mode(self, **option):
        """Set qos bandwidth limit."""
        return self._exec_cli("set qos-mode", **option)

    def _show_enclosure_status(self, **option):
        """Get the version of machine."""
        try:
            output = self._exec_cli("show enclosure-status", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')
            clidata = clidatalist[0].split('\t')
            versioninfo = {'version': clidata[11]}

            output['message'] = versioninfo

        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show enclosure status error: %s"
                                 % six.text_type(ex)}

        return output
