# Copyright (c) 2019 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import ctypes
import functools
import inspect
import threading

from oslo_concurrency import lockutils
from oslo_log import log as logging

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.volume import volume_types
import oslo_context.context as context_req

LOG = logging.getLogger(__name__)


class FJDXInterruptableThread(threading.Thread):
    def __init__(self, func, *args, **kwargs):
        self.ctxt = None
        if '_context' in kwargs:
            self.ctxt = kwargs['_context']
            kwargs.pop('_context')
        threading.Thread.__init__(self)
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._result = None

    def run(self):
        try:
            if self.ctxt:
                context_req._request_store.context = self.ctxt
            self._result = self._func(*self._args, **self._kwargs)
        except Exception as e:
            self._result = e

    @property
    def result(self):
        return self._result

    def _get_thread_id(self):
        """judge self's thread status"""
        if not self.isAlive():
            raise threading.ThreadError("The thread is not alive.")

        # Thread id whether is exist.
        if hasattr(self, "_thread_id"):
            return self._thread_id

        # Find thread id in _active.items.
        for tid, tobj in threading._active.items():
            if tobj is self:
                self._thread_id = tid
                return tid

        raise AssertionError("Thread's status cannot be judged.")

    def terminate(self):
        """terminate the thread"""
        _async_raise(self._get_thread_id(), SystemExit)


def _async_raise(tid, exctype):
    """terminate the thread, post handle"""
    if not inspect.isclass(exctype):
        raise TypeError("Types incorrect.")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(tid), ctypes.py_object(exctype))
    if res == 0:
        LOG.debug("The thread is not alive.")
    elif res != 1:
        """terminate failed"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
        raise SystemError("Invoke PyThreadState_SetAsyncExc failed.")


class FJDXTimeout(object):
    def __init__(self):
        self._sec = None

    def __call__(self, func):
        def wrapped_f(*args, **kwargs):
            self._sec = kwargs.get('timeout', None)
            self._sec = self._sec if self._sec != 0 else None

            # Pass the logs req value in different threads.
            if getattr(context_req._request_store, 'context', None):
                kwargs['_context'] = context_req._request_store.context
            it = FJDXInterruptableThread(func, *args, **kwargs)
            it.start()
            it.join(self._sec)
            if not it.is_alive():
                if isinstance(it.result, Exception):
                    raise it.result
                else:
                    return it.result
            msg = _('Execution timeout!')
            if it.is_alive():
                it.terminate()
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return wrapped_f


def FJDXLockutils(name, lock_file_prefix, external=False, lock_path=None):
    """Lockutils for Fujitsu ETERNUS DX."""
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            lockname = 'ETERNUS_DX-' + name + '-' + self.storage_name

            @lockutils.synchronized(lockname, lock_file_prefix,
                                    external, lock_path)
            @functools.wraps(func)
            def caller():
                return func(self, *args, **kwargs)
            return caller()
        return wrapper
    return decorator


def get_metadata(volume):
    """Get metadata using volume information."""
    LOG.debug('get_metadata, volume id: %s.',
              volume['id'])

    d_metadata = {}

    metadata = volume.get('volume_metadata')

    # value = {} enters the if branch, value = None enters the else.
    if metadata is not None:
        d_metadata = {
            data['key']: data['value'] for data in metadata
        }
    else:
        metadata = volume.get('metadata')
        if metadata:
            d_metadata = {
                key: metadata[key] for key in metadata
            }

    LOG.debug('get_metadata, metadata is: %s.',
              d_metadata)
    return d_metadata


def get_format_property(volume, format_key):
    format_in_type = get_extra_specs(volume,
                                     key=format_key,
                                     default='False')
    format_in_vol = get_metadata(volume).get(format_key, 'False')

    if 'true' in [format_in_type.lower(), format_in_vol.lower()]:
        return 'True'
    else:
        return 'False'


def get_extra_specs(volume, key=None, default=None):
    """Get extra specs from volume.

    If key is none, return all extra specs.
    """
    LOG.debug('get_extra_specs, '
              'volume id: %(v_id)s, '
              'key: %(key)s, '
              'default: %(default)s.',
              {'v_id': volume['id'],
               'key': key,
               'default': default})

    extra_specs = None
    ret = default

    volume_type_id = volume.get('volume_type_id')
    if volume_type_id:
        ctxt = context.get_admin_context()
        volume_type = volume_types.get_volume_type(ctxt, volume_type_id)
        extra_specs = volume_type.get('extra_specs')

    if extra_specs:
        if not key:
            ret = extra_specs
        else:
            ret = extra_specs.get(key, default)

    LOG.debug('get_extra_specs: %s.', extra_specs)
    return ret
