import log
import os
import re
from threading import Timer
import time
from xml.etree.ElementTree import parse

from cinder.i18n import _
from cinder.volume.drivers.fujitsu.eternus_dx.osvd_monitor \
    import eternus_dx_monitor_cli
from oslo_concurrency import lockutils
from oslo_config import cfg

MAX_AG_LUN_NO = 256
MIN_POLLING_TIME = 3600
MAX_POLLING_TIME = MIN_POLLING_TIME * 24
DRIVER_TYPE_ISCSI = "iSCSI"
DRIVER_TYPE_FC = "FC"
SUPPORT_DRIVER_TYPE_LIST = [DRIVER_TYPE_FC, DRIVER_TYPE_ISCSI]
POOL_TYPE_RG = "standard"
POOL_TYPE_TPP = "tpv"
SUPPORT_POOL_TYPE_LIST = [POOL_TYPE_RG, POOL_TYPE_TPP]
DEFAULT_TEMP_VOL_NAME = "!@#_OSVD_temp#@!"
DEFAULT_PRIVATE_KEY = "/var/lib/cinder/eternus"

MONITOR_CFG_FILE = "osvd_host_affinity_monitor.conf"
DRIVER_TYPE = "protocol"
TEMP_VOL_NAME = "temp_volume_name"
POOL_NAME = "pool_name"
POLLING_TIME = "polling_time"
OSVD_CONF = "osvd_conf"
HOST_INFO_FILE = "host_info_conf"
HOST_AG_MAP_FILE = "host_affinity_mapping_file"
PASSWORDLESS = "passwordless"
PRIVATE_KEY = "private_key"


DOLPHIN = '30'
BELUGA = ('C1', 'C2')
PENGUIN = '40'
ADELIE = ('A1')

RETCODE_dic = {
    '0': 'Success',
    '1': 'Method Not Supported',
    '4': 'Failed',
    '5': 'Invalid Parameter',
    '4096': 'Method Parameters Checked - Job Started',
    '4097': 'Size Not Supported',
    '4101': 'Target/initiator combination already exposed',
    '4102': 'Requested logical unit number in use',
    '32769': 'Maximum number of Logical Volume in a RAID group '
             'has been reached',
    '32770': 'Maximum number of Logical Volume in the storage device '
             'has been reached',
    '32771': 'Maximum number of registered Host WWN '
             'has been reached',
    '32772': 'Maximum number of affinity group has been reached',
    '32773': 'Maximum number of host affinity has been reached',
    '32781': 'Not available under current system configuration',
    '32782': 'Controller firmware update in process',
    '32785': 'The RAID group is in busy state',
    '32786': 'The Logical Volume is in busy state',
    '32787': 'The device is in busy state',
    '32788': 'Element Name is in use',
    '32791': 'Maximum number of copy session has been reached',
    '32792': 'No Copy License',
    '32793': 'Session does not exist',
    '32794': 'Phase is not correct',
    '32796': 'Quick Format Error',
    '32801': 'The CA port is in invalid setting',
    '32802': 'The Logical Volume is Mainframe volume',
    '32803': 'The RAID group is not operative',
    '32804': 'The Logical Volume is not operative',
    '32805': 'The Logical Element is Thin provisioning Pool Volume',
    '32806': 'The Logical Volume is pool for copy volume',
    '32807': 'The Logical Volume is unknown volume',
    '32808': 'No Thin Provisioning License',
    '32809': 'The Logical Element is ODX volume',
    '32810': 'The specified volume is under use as NAS volume',
    '32811': 'This operation cannot be performed to the NAS resources',
    '32813': 'This operation cannot be performed to the '
             'VVOL resources',
    '32816': 'Generic fatal error',
    '32817': 'Inconsistent State with 1Step Restore '
             'operation',
    '32818': 'REC Path failure during copy',
    '32819': 'RAID failure during EC/REC copy',
    '32820': 'Previous command in process',
    '32821': 'Cascade local copy session exist',
    '32822': 'Cascade EC/REC session is not suspended',
    '35302': 'Invalid LogicalElement',
    '35304': 'LogicalElement state error',
    '35316': 'Multi-hop error',
    '35318': 'Maximum number of multi-hop has been reached',
    '35324': 'RAID is broken',
    '35331': 'Maximum number of session has been reached(per device)',
    '35333': 'Maximum number of session has been reached(per SourceElement)',
    '35334': 'Maximum number of session has been reached(per TargetElement)',
    '35335': 'Maximum number of Snapshot generation has been reached '
             '(per SourceElement)',
    '35346': 'Copy table size is not setup',
    '35347': 'Copy table size is not enough',
}

LOG = log.FinalLogger().getLogger(__name__)
CONF = cfg.CONF


class FJDXMonitor(object):
    """Monitor program for OSVD1.6.0."""

    def __init__(self):
        self.cfg_dic = get_value_from_cfg_file(MONITOR_CFG_FILE)
        self.config_file = self.cfg_dic.get(OSVD_CONF, None)
        vol_name = self.cfg_dic.get(TEMP_VOL_NAME, None)
        passwordless = self.cfg_dic.get(PASSWORDLESS, None)
        private_key_path = self.cfg_dic.get(PRIVATE_KEY, None)

        self.use_ipv6 = False
        storage_ip = self._get_drvcfg('EternusIP', filename=self.config_file)
        if storage_ip.find(':') != -1:
            self.use_ipv6 = True
        if vol_name:
            self.temp_vol_name = vol_name
        else:
            self.temp_vol_name = DEFAULT_TEMP_VOL_NAME

        if passwordless == 'False':
            self.passwordless = False
        else:
            self.passwordless = True

        if private_key_path:
            self.private_key_path = private_key_path
        else:
            self.private_key_path = DEFAULT_PRIVATE_KEY

        self.fjdxcli = None
        self._check_user()
        self._create_tmp_volume()
        self.f_hosts = None
        self.host_ag_dict = None

        model_info = self._get_model_supplement_info()
        self.ce_support = False
        if model_info.get('machine_info') == DOLPHIN and \
                model_info.get('model_info') in BELUGA:
            self.ce_support = True
        if model_info.get('machine_info') == PENGUIN and \
                model_info.get('model_info') in ADELIE:
            self.ce_support = True

    def _get_model_supplement_info(self):
        """Get model supplement info. """

        LOG.debug('_get_model_supplement_info start. ')

        rc, emsg, clidata = self._exec_eternus_cli(
            'show_model_supplement_info')
        if rc != 0:
            msg = (_('_get_model_supplement_info, '
                     '_show_model_supplement_info failed. '
                     'Return code:%(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: {machine_info:x,model_info:y}
        LOG.debug('_get_model_supplement_info, '
                  'model supplement info: %(info)s.',
                  {'info': clidata})
        return clidata

    def _check_user(self):
        """Check whether user's role is accessible to ETERNUS and Software."""
        LOG.info('_check_user, Enter method.')
        ret = True
        rc, errordesc, job = self._exec_eternus_cli('check_user_role')
        if rc != 0:
            msg = (_('_check_user, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Job: %(job)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise Exception(msg)

        if job != 'Software':
            msg = (_('_check_user, '
                     'Specified user(%(u)s) does not have '
                     'Software role: %(role)s.')
                   % {'u': self._get_drvcfg('EternusUser', self.config_file),
                      'role': job})
            LOG.error(msg)
            raise Exception(msg)

        return ret

    def _get_host_ag_from_cfg_file(self, conf_file_name):
        if not os.path.exists(conf_file_name):
            msg = (_('_get_host_ag_from_cfg_file, '
                     'cfg file(%(path)s) is not exist.')
                   % {'path': conf_file_name})
            LOG.error(msg)
            raise Exception(msg)

        # Every line in cfg_file likes follow:
        # port1,port2,...=iqn=lungroup
        # port: xyz,wxyz
        # xyz: Fixed 3 digits, CM number + CA number + Port number
        # wxyz: Fixed 4 digits,
        #       CE number(hex) + CM number + CA number + Port number
        # iqn: iqn.domaindate.reverse.domain.name:optional name
        # lungroup: lungroup name

        # {host: {portlist: [port1, port2]}, lungroup: AG_#01}
        host_ag_dict = {}
        port_list = []
        host_list = []
        pattern = '^(\s*[0-9A-Fa-f]{3,4}\s*,)*\s*[0-9A-Fa-f]{3,4}\s*=\s*\w'
        with open(conf_file_name) as f:
            line = f.readline()
            while line:
                result = re.match(pattern, line)
                if result:
                    infos = [x.strip() for x in line.split('=', 2)]
                    if len(infos) != 3:
                        msg = (_('_get_host_ag_from_cfg_file, '
                                 'pattern of cfg file(%(path)s) '
                                 'should be port1,port2,...=iqn=lungroup.')
                               % {'path': conf_file_name})
                        LOG.error(msg)
                        raise Exception(msg)

                    port_list = [x.strip() for x in infos[0].split(',')]
                    host_ag_dict[infos[1].lower()] = {'portlist': port_list,
                                                      'lungroup': infos[2]}
                    host_list.append(infos[1])
                line = f.readline()

        if not host_list:
            msg = (_('_get_host_ag_from_cfg_file, '
                     'cfg file: %(path)s, '
                     'no host affinity group info in this file. '
                     'Please check the content of this file!')
                   % {'path': conf_file_name})
            LOG.error(msg)
            raise Exception(msg)

        LOG.info('_get_host_ag_from_cfg_file, '
                 'host_ag_dict: %s, Exit method.', host_ag_dict)
        return host_list, host_ag_dict

    def _get_pool_type(self, pool_name):
        LOG.debug('_get_pool_type, enter method. Pool name: %s.', pool_name)

        rc, emsg, clidata = self._exec_eternus_cli('show_raid_groups')
        if rc != 0:
            msg = (_('_get_pool_type, '
                     'show_raid_groups failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{Name:x,...}]
        rg_names = [x.get('Name') for x in clidata]
        pool_type = None
        if pool_name in rg_names:
            pool_type = POOL_TYPE_RG

        if not pool_type:
            rc, emsg, clidata = self._exec_eternus_cli('show_thin_pro_pools')
            if rc != 0:
                msg = (_('_get_pool_type, '
                         'show_thin_pro_pools failed. '
                         'Return code: %(rc)lu, '
                         'Error code:%(clidata)s, '
                         'Message:%(emsg)s.')
                       % {'rc': rc,
                          'clidata': clidata,
                          'emsg': emsg})
                LOG.error(msg)
                raise Exception(msg)

            # clidata: [{Name:x,...}]
            tpp_names = [x.get('Name') for x in clidata]
            if pool_name in tpp_names:
                pool_type = POOL_TYPE_TPP
            else:
                msg = (_('_get_pool_type, '
                         'pool is not exist on ETERNUS or '
                         'pool type is not supported. '
                         'pool name: %(name)s, '
                         'supported pool types: %(types)s.')
                       % {'name': pool_name,
                          'types': SUPPORT_POOL_TYPE_LIST})
                LOG.error(msg)
                raise Exception(msg)

        LOG.debug('_get_pool_type, exit method. '
                  'Pool name: %(name)s, pool type: %(type)s.',
                  {'name': pool_name, 'type': pool_type})
        return pool_type

    def _create_tmp_volume(self):
        '''Create temp volume.'''
        LOG.info('_create_tmp_volume, Enter method.')
        pool_name = self.cfg_dic.get(POOL_NAME, None)
        pool_type = self._get_pool_type(pool_name)

        option = {'size': '24mb',
                  'type': pool_type,
                  'name': self.temp_vol_name}
        if pool_type == POOL_TYPE_RG:
            option['rg-name'] = pool_name
        else:
            option['pool-name'] = pool_name

        # RG: create volume -name xxx -rg-name xxx -type standard -size 24mb
        # TPP: create volume -name xxx -pool-name xxx -type tpv -size 24mb
        rc, emsg, clidata = self._exec_eternus_cli('create_volume', **option)

        if rc != 0 and clidata == "E0120":
            LOG.debug('_create_tmp_volume, '
                      'temp volume has been created.')
            return
        elif rc != 0:
            msg = (_('_create_tmp_volume, '
                     'Return code:%(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)
        return

    def _period_task(self):
        for host_iqn in self.f_hosts:
            self._check_host_affinity_group(host_iqn)
        return

    def polling_check(self, polling_time):
        Timer(polling_time, self.polling_check, [polling_time]).start()
        self._period_task()

    def _check_host_affinity_group(self, host_iqn):
        LOG.info('_check_host_affinity_group, '
                 'host IQN: %s, Enter method.', host_iqn)

        try:
            host_num = self._get_host_num_by_iqn(host_iqn)
            if host_num == 'Error':
                return
            elif host_num is not None:
                self._find_host_affinity_group(host_iqn, host_num)
            else:
                host_num = self._create_host_num_by_iqn(host_iqn)
                if host_num is not None:
                    self._create_host_affinity_group(host_iqn, host_num)
                else:
                    return
        except Exception:
            msg = (_('_check_host_affinity_group, '
                     'host IQN: %(IQN)s, '
                     'Exception or Error is occurred.')
                   % {'IQN': host_iqn})
            LOG.error(msg)

    def _get_host_num_by_iqn(self, iqn):
        '''Get host number by host iscsi name from ETERNUS.'''
        rc, emsg, clidata = self._exec_eternus_cli('show_host_iscsi_names')

        if rc != 0:
            msg = (_('_get_host_num_by_iqn, '
                     'Return code:%(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            host_num = 'Error'
            return host_num

        host_num = None
        for hostdata in clidata:
            if hostdata['Host Name'].lower() == iqn.lower():
                host_num = hostdata['Host Num']
                break

        LOG.info('_get_host_num_by_iqn, '
                 'host number: %s, Exit method.', host_num)
        return host_num

    def _create_host_num_by_iqn(self, iqn):
        '''Create host number by host iscsi name in ETERNUS.'''
        if self.use_ipv6:
            option = {'iscsi-name': iqn,
                      'ip-version': 'ipv6'}
        else:
            option = {'iscsi-name': iqn}
        rc, emsg, clidata = self._exec_eternus_cli(
            'create_host_iscsi_name', **option)

        if rc != 0:
            msg = (_('_create_host_num_by_iqn, '
                     'Return code:%(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            return None

        host_num = int(clidata[0], 16)
        LOG.info('_create_host_num_by_iqn, '
                 'host number: %s, Exit method.', host_num)
        return host_num

    def _create_host_affinity_group(self, host_iqn, host_num):
        LOG.info('_create_host_affinity_group, '
                 'host_iqn: %(iqn)s, '
                 'host_num: %(num)s, Enter method.',
                 {'iqn': host_iqn, 'num': host_num})
        f_ports = self.host_ag_dict[host_iqn.lower()]['portlist']
        lungroup = self.host_ag_dict[host_iqn.lower()]['lungroup']

        # Get affinity group list
        rc, errordesc, ag_list = self._exec_eternus_cli(
            'show_affinity_group_list')

        if rc != 0:
            msg = (_('_create_host_affinity_group, '
                     'get_affinity_group list failed, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Job: %(retdata)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'retdata': ag_list})
            LOG.error(msg)
            return

        host_affinity_dic = {}
        if lungroup in ag_list:
            option = {'ag-name': lungroup}
            rc, errordesc, host_affinity_dic = self._exec_eternus_cli(
                'show_host_affinity', **option)

            if rc != 0:
                msg = (_('_create_host_affinity_group, '
                         'get ports of affinity group failed, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Job: %(retdata)s.')
                       % {'rc': rc,
                          'errordesc': errordesc,
                          'retdata': host_affinity_dic})
                LOG.error(msg)
                return

            # host_affinity_dic: {host number: {port: affinity group name}}
            ag_ports = []
            for item in host_affinity_dic.values():
                # ag_ports: ['000', '001']
                ag_ports.extend(item.keys())

            # Get intersection of ag_ports and f_ports
            port_int = [val for val in ag_ports if val in f_ports]
            if port_int:
                msg = (_('_create_host_affinity_group, '
                         'some ports %(ports)s in file '
                         'have already been mapped '
                         'with lungroup: %(lungroup)s.')
                       % {'ports': port_int,
                          'lungroup': lungroup})
                LOG.error(msg)
                return

            # Get luns in lungroup
            ag_luns_info = []
            rc, errordesc, ag_luns_info = self._exec_eternus_cli(
                'show_affinity_group', **option)
            if rc != 0:
                msg = (_('_create_host_affinity_group, '
                         'get luns in lungroup failed, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Job: %(retdata)s.')
                       % {'rc': rc,
                          'errordesc': errordesc,
                          'retdata': ag_luns_info})
                LOG.error(msg)
                return

            # Get unused lun no
            idx = 0
            used_nos = []
            for data in ag_luns_info:
                no = int(data['Lun Num'])
                used_nos.append(no)
                idx = no if idx < no else idx

            lun_no = idx + 1
            if MAX_AG_LUN_NO <= lun_no:
                temp = range(MAX_AG_LUN_NO)
                unused_num = None
                for t in temp:
                    if t not in used_nos:
                        unused_num = t
                        break

                lun_no = unused_num
                if lun_no is None:
                    msg = (_('_create_host_affinity_group, '
                             'the volumes in lungroup(%(ag_name)s) '
                             'is full(MAX count: %(count)s).')
                           % {'ag_name': lungroup,
                              'count': MAX_AG_LUN_NO})
                    LOG.error(msg)
                    return

            # Put lun into lungroup
            put_lun_option = {'ag-name': lungroup,
                              'volume-name': self.temp_vol_name,
                              'lun': lun_no}
            rc, errordesc, clidata = self._exec_eternus_cli(
                'set_affinity_group', **put_lun_option)
            if rc != 0:
                # TEMP_VOL_NAME has already been put in lungroup
                if clidata == 'E0071':
                    LOG.warning('_create_host_affinity_group, '
                                'lun %(lun_name)s is already in '
                                'lungroup %(ag_name)s.',
                                {'ag_name': lungroup,
                                 'lun_name': self.temp_vol_name})
                else:
                    msg = (_('_create_host_affinity_group, '
                             'put lun into lungroup %(ag_name)s failed, '
                             'Return code: %(rc)lu, '
                             'Error: %(errordesc)s, '
                             'Job: %(retdata)s.')
                           % {'ag_name': lungroup,
                              'rc': rc,
                              'errordesc': errordesc,
                              'retdata': clidata})
                    LOG.error(msg)
                    return

            # Finally set host affinity with existed lungroup
            self._set_host_affinity(lungroup, host_num, f_ports)
        else:
            # Affinity group not in Eternus
            option = {'name': lungroup,
                      'volume-name': self.temp_vol_name,
                      'lun': 0}
            rc, errordesc, clidata = self._exec_eternus_cli(
                'create_affinity_group', **option)

            if rc != 0:
                msg = (_('_create_host_affinity_group, '
                         'create lungroup %(ag_name)s failed, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Job: %(retdata)s.')
                       % {'ag_name': lungroup,
                          'rc': rc,
                          'errordesc': errordesc,
                          'retdata': clidata})
                LOG.error(msg)
                return

            # Set host affinity with newly created lungroup
            self._set_host_affinity(lungroup, host_num, f_ports)

        return

    def _set_host_affinity(self, ag_name, host_num, ports):
        # Set host affinity with newly created lungroup
        set_host_affinity_option = {'ag-name': ag_name,
                                    'host-number': str(host_num),
                                    'port': ','.join(ports)}
        rc, emsg, clidata = self._exec_eternus_cli(
            'set_host_affinity', **set_host_affinity_option)

        if rc != 0:
            msg = (_('_create_host_affinity_group, '
                     'create host affinity failed, '
                     'lungroup: %(ag_name)s, '
                     'hostnumber: %(hostno)s, '
                     'portlist: %(portlist)s, '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'ag_name': ag_name,
                      'hostno': str(host_num),
                      'portlist': ','.join(ports),
                      'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            return False
        return True

    def _find_host_affinity_group(self, host_iqn, host_num):
        LOG.info('_find_host_affinity_group, '
                 'host_iqn: %(iqn)s, '
                 'host_num: %(num)s, Enter method.',
                 {'iqn': host_iqn, 'num': host_num})

        option = {'host-number': host_num}
        rc, emsg, clidata = self._exec_eternus_cli(
            'show_host_affinity', **option)

        if rc != 0:
            msg = (_('_find_host_affinity_group, '
                     'show_host_affinity failed. '
                     'Return code:%(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            return

        # host_affinity_dic: {host number: {port: affinity group name}}
        e_aglist = clidata.get(host_num, {})
        f_ports = self.host_ag_dict[host_iqn.lower()]['portlist']

        # Check if ports in map file is available for iSCSI.
        not_avai_ports = [x for x in f_ports if x not in self.iscsi_ports]
        if not_avai_ports:
            msg = (_('_find_host_affinity_group, '
                     'some ports in map file is not available for iSCSI. '
                     'Please check the map file. '
                     'Host IQN: %(host_iqn)s, '
                     'ports in map file: %(f_ports)s, '
                     'iSCSI ports on ETERNUS: %(iscsi_ports)s, '
                     'not available ports: %(not_avai_ports)s.')
                   % {'host_iqn': host_iqn,
                      'f_ports': f_ports,
                      'iscsi_ports': self.iscsi_ports,
                      'not_avai_ports': not_avai_ports})
            LOG.error(msg)
            return

        f_lungroup = self.host_ag_dict[host_iqn.lower()]['lungroup']

        OK_ports = []
        NG_ports = []
        unused_ports = []
        if e_aglist:
            for port in f_ports:
                if port in e_aglist:
                    if e_aglist[port] == f_lungroup:
                        OK_ports.append(port)
                    else:
                        NG_ports.append(port)
                else:
                    unused_ports.append(port)
        else:
            self._create_host_affinity_group(host_iqn, host_num)
            return

        if NG_ports:
            msg = (_('_find_host_affinity_group, '
                     'some port lungroup is different with config file, '
                     'host iqn:%(iqn)s, '
                     'ETERNUS affinity group:%(clidata)s, '
                     'config file affinity group:%(emsg)s.')
                   % {'iqn': host_iqn,
                      'clidata': clidata,
                      'emsg': self.host_ag_dict[host_iqn.lower()]})
            LOG.error(msg)
            return

        if unused_ports and OK_ports:
            msg = (_('_find_host_affinity_group, '
                     'some port are not used in ETERNUS, '
                     'host iqn:%(iqn)s, '
                     'ETERNUS affinity group:%(clidata)s, '
                     'config file affinity group:%(emsg)s.')
                   % {'iqn': host_iqn,
                      'clidata': clidata,
                      'emsg': self.host_ag_dict[host_iqn.lower()]})
            LOG.error(msg)
            return

        if len(unused_ports) == len(f_ports):
            self._create_host_affinity_group(host_iqn, host_num)
            return

        self._check_affinity_group(self.temp_vol_name, f_lungroup)

        if (len(OK_ports) == len(f_ports)) and (len(e_aglist) > len(OK_ports)):
            msg = (_('_find_host_affinity_group, '
                     'ETERNUS host affinity group is more than config file, '
                     'host iqn:%(iqn)s, '
                     'ETERNUS affinity group:%(clidata)s, '
                     'config file affinity group:%(emsg)s.')
                   % {'iqn': host_iqn,
                      'clidata': clidata,
                      'emsg': self.host_ag_dict[host_iqn.lower()]})
            LOG.warning(msg)
            return

        LOG.info('_find_host_affinity_group, '
                 'host affinity: %s, Exit method.', clidata)

    def _check_affinity_group(self, vol_name, ag_name):
        option = {'ag-name': ag_name}
        rc, errordesc, ag_luns_info = self._exec_eternus_cli(
            'show_affinity_group', **option)

        if rc != 0:
            msg = (_('_check_affinity_group, '
                     'get luns in lungroup failed, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Job: %(retdata)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'retdata': ag_luns_info})
            LOG.error(msg)
            return False

        idx = 0
        used_nos = []
        for data in ag_luns_info:
            if data['Lun Name'] == vol_name:
                LOG.debug('_check_affinity_group,'
                          'the specified volume(%(vol_name)s) '
                          'is already in affinity group(%(ag_name)s).',
                          {'vol_name': vol_name, 'ag_name': ag_name})
                return True

            no = int(data['Lun Num'])
            used_nos.append(no)
            idx = no if idx < no else idx

        lun_no = idx + 1
        if MAX_AG_LUN_NO <= lun_no:
            temp = range(MAX_AG_LUN_NO)
            unused_num = None
            for t in temp:
                if t not in used_nos:
                    unused_num = t
                    break

            lun_no = unused_num
            if lun_no is None:
                msg = (_('_check_affinity_group, '
                         'the volumes in lungroup(%(ag_name)s) '
                         'is full(MAX count: %(count)s).')
                       % {'ag_name': ag_name,
                          'count': MAX_AG_LUN_NO})
                LOG.error(msg)
                return False

        # Put lun into lungroup
        put_lun_option = {'ag-name': ag_name,
                          'volume-name': vol_name,
                          'lun': lun_no}
        rc, errordesc, clidata = self._exec_eternus_cli(
            'set_affinity_group', **put_lun_option)

        if rc != 0:
            # TEMP_VOL_NAME has already been put in lungroup
            if clidata == 'E0071':
                LOG.warning('_check_affinity_group, '
                            'lun %(lun_name)s is already in '
                            'lungroup %(ag_name)s.',
                            {'ag_name': ag_name,
                             'lun_name': vol_name})
            else:
                msg = (_('_check_affinity_group, '
                         'put lun into lungroup %(ag_name)s failed, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Job: %(retdata)s.')
                       % {'ag_name': ag_name,
                          'rc': rc,
                          'errordesc': errordesc,
                          'retdata': clidata})
                LOG.error(msg)
                return False
        return True

    @lockutils.synchronized('CLI-exec', 'OSVD-monitor-', True, './')
    def _exec_eternus_cli(self, command, retry=20, retry_interval=5,
                          retry_code=[32787], filename=None, **param_dict):
        """Execute ETERNUS CLI."""
        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'parameters: %(b)s.',
                  {'a': command,
                   'b': param_dict})

        out = None
        _err = None
        result = None
        rc = None
        retdata = None
        errordesc = None
        user = self._get_drvcfg('EternusUser', filename=self.config_file)
        storage_ip = self._get_drvcfg('EternusIP', filename=self.config_file)

        if not self.fjdxcli:
            if self.passwordless:
                self.fjdxcli = eternus_dx_monitor_cli.FJDXCLI(
                    user, storage_ip, keyfile=self.private_key_path)
            else:
                self.password = self._get_drvcfg('EternusPassword', filename)
                self.fjdxcli = eternus_dx_monitor_cli.FJDXCLI(
                    user, storage_ip, password=self.password)

        for retry_num in range(retry):
            # Execute ETERNUS CLI and get return value.
            try:
                if param_dict:
                    out = self.fjdxcli.done(command, **param_dict)
                else:
                    out = self.fjdxcli.done(command)

                out_dict = out
                result = out_dict.get('result')
                rc_str = out_dict.get('rc')
                retdata = out_dict.get('message')
            except Exception as ex:
                msg = (_('_exec_eternus_cli, '
                         'stdout: %(out)s, '
                         'stderr: %(err)s, '
                         'unexpected error: %(ex)s.')
                       % {'out': out,
                          'err': _err,
                          'ex': ex})
                LOG.error(msg)
                raise Exception(msg)

            # Check ssh result.
            if result == 255:
                LOG.info('_exec_eternus_cli, retry, '
                         'command: %(command)s, '
                         'option: %(option)s, '
                         'ip: %(ip)s, '
                         'SSH Result: %(result)s, '
                         'retdata: %(retdata)s, '
                         'TryNum: %(rn)s.',
                         {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata,
                          'rn': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
            elif result != 0:
                msg = (_('_exec_eternus_cli, '
                         'unexpected error, '
                         'command: %(command)s, '
                         'option: %(option)s, '
                         'ip: %(ip)s, '
                         'result: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata})
                LOG.error(msg)
                raise Exception(msg)

            # Check CLI return code.
            if rc_str.isdigit():
                # SMI-S style return code
                rc = int(rc_str)

                try:
                    errordesc = RETCODE_dic[str(rc)]
                except Exception:
                    errordesc = 'Undefined Error!!'

                if rc in retry_code:
                    LOG.info('_exec_eternus_cli, retry, '
                             'ip: %(ip)s, '
                             'RetryCode: %(rc)s, '
                             'TryNum: %(rn)s.',
                             {'ip': storage_ip,
                              'rc': rc,
                              'rn': (retry_num + 1)})
                    time.sleep(retry_interval)
                    continue

                break
            else:
                # CLI style return code
                LOG.warning('_exec_eternus_cli, '
                            'WARNING!! '
                            'ip: %(ip)s, '
                            'ReturnCode: %(rc_str)s, '
                            'ReturnData: %(retdata)s.',
                            {'ip': storage_ip,
                             'rc_str': rc_str,
                             'retdata': retdata})

                errordesc = retdata
                retdata = rc_str
                rc = 4  # Failed
                break
        else:
            if 0 < result:
                msg = (_('_exec_eternus_cli, '
                         'cannot connect to ETERNUS. '
                         'SSH Result: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'result': result,
                          'retdata': retdata})

                LOG.error(msg)
                raise Exception(msg)
            else:
                LOG.warning('_exec_eternus_cli, Retry was exceeded.')

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'parameters: %(b)s, '
                  'ip: %(ip)s, '
                  'Return code: %(rc)s, '
                  'Error: %(errordesc)s.',
                  {'a': command,
                   'b': param_dict,
                   'ip': storage_ip,
                   'rc': rc,
                   'errordesc': errordesc})
        return ret

    def _get_drvcfg(self, tagname, filename, multiple=False):
        """Read from driver configuration file."""
        tree = parse(filename)
        elem = tree.getroot()

        ret = None
        if not multiple:
            ret = elem.findtext(".//" + tagname)
        else:
            ret = []
            for e in elem.findall(".//" + tagname):
                if e.text and (e.text not in ret):
                    ret.append(e.text)

        if not ret:
            msg = (_('_get_drvcfg, '
                     'filename: %(filename)s, '
                     'tagname: %(tagname)s, '
                     'data is None!! '
                     'Please edit driver configuration file and correct.')
                   % {'filename': filename,
                      'tagname': tagname})
            LOG.error(msg)
            raise Exception(msg)

        return ret


class FJDXiSCSIMonitor(FJDXMonitor):
    def __init__(self):
        super(FJDXiSCSIMonitor, self).__init__()
        iscsi_map_file = self.cfg_dic.get(HOST_AG_MAP_FILE, None)
        self.f_hosts, self.host_ag_dict = self._get_host_ag_from_cfg_file(
            iscsi_map_file)
        self.iscsi_ports = self._get_iscsi_port_list()

    def _get_iscsi_port_list(self):
        LOG.debug('_get_iscsi_port_list, enter method.')

        param_dict = {'ce_support': self.ce_support}
        rc, emsg, clidata = self._exec_eternus_cli('show_iscsi_parameters',
                                                   **param_dict)
        if rc != 0:
            msg = (_('_get_iscsi_port_list, '
                     'show_iscsi_parameters failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{port num:x,port iqn:x,port str:x,port ip:x}]
        port_list = [x.get('port str') for x in clidata]
        LOG.debug('_get_iscsi_port_list, exit method. '
                  'iSCSI port list on ETERNUS: %s.', port_list)
        return port_list


class FCHostAffinityGroup(object):
    def __init__(self):
        self.wwpn = None
        self.wwpn_name = None
        self.wwpn_num = None
        self.port_list = []
        self.port_list_for_create = []
        self.ag_name = None
        self.ag_num = None
        self.is_new_wwpn = False
        self.ag_name_list = []
        self.is_effective = False


class FJDXFCMonitor(FJDXMonitor):
    def __init__(self):
        super(FJDXFCMonitor, self).__init__()

        fc_host_wwpn_file = self.cfg_dic.get(HOST_INFO_FILE, None)
        self.f_hosts, self.host_wwpns_dict = (
            self._get_host_wwpn_from_cfg_file(fc_host_wwpn_file))
        self.fc_ports = self._get_fc_port_list()

    def _get_fc_port_list(self):
        LOG.debug('_get_fc_port_list, enter method.')

        param_dict = {'ce_support': self.ce_support}
        rc, emsg, clidata = self._exec_eternus_cli('show_fc_parameters',
                                                   **param_dict)
        if rc != 0:
            msg = (_('_get_fc_port_list, '
                     'show_fc_parameters failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{port num:x,port wwpn:x,port str:x}]
        port_list = [x.get('port str') for x in clidata]
        LOG.debug('_get_fc_port_list, exit method. '
                  'FC port list on ETERNUS: %s.', port_list)
        return port_list

    def create_fc_map_file(self):
        LOG.debug('create_fc_map_file, enter method.')

        # Clear the FC map file.
        map_file = self.cfg_dic.get(HOST_AG_MAP_FILE, None)
        self._write_configuration(map_file, None)

        # Get FC zone info.
        fc_zone_dict = self._get_enternus_fc_zone_info()

        for hostname in self.f_hosts:
            self._check_host_affinity(hostname, fc_zone_dict)

        LOG.debug('create_fc_map_file, exit method.')

    def _get_enternus_fc_zone_info(self):
        LOG.debug('_get_enternus_fc_zone_info, enter method.')

        param_dict = {'ce_support': self.ce_support}
        rc, emsg, clidata = self._exec_eternus_cli('discover_host_wwn_names',
                                                   **param_dict)
        if rc != 0:
            msg = (_('_get_enternus_fc_zone_info, '
                     'discover_host_wwn_names failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: {host WWPN: [port1, port2]}
        LOG.debug('_get_enternus_fc_zone_info, exit method. '
                  'FC zone info on ETERNUS: %s.', clidata)
        return clidata

    def _get_host_wwpn_from_cfg_file(self, conf_file_name):
        LOG.debug('_get_host_wwpn_from_cfg_file, enter method. '
                  'cfg file name: %s.', conf_file_name)

        if not os.path.exists(conf_file_name):
            msg = (_('_get_host_wwpn_from_cfg_file, '
                     'cfg file(%(path)s) is not exist.')
                   % {'path': conf_file_name})
            LOG.error(msg)
            raise Exception(msg)

        # cfg file line format: host name=WWPN,WWPN.
        host_ag_dict = {}  # {host: [WWPN, WWPN]}
        host_list = []  # [host name, host name, ...]
        pattern = '^[0-9A-Fa-f]{16}$'
        with open(conf_file_name) as f:
            line = f.readline()
            while line:
                if ('=' in line) and (not line.strip().startswith('#')):
                    # ['host','WWPN,WWPN']
                    infos = [x.strip() for x in line.split('=', 1)]
                    host_name = infos[0]
                    wwpns = [x.strip().lower() for x in infos[1].split(',')]
                    wwpns = list(set(wwpns))
                    avai_wwpns = [x for x in wwpns if re.match(pattern, x)]

                    if avai_wwpns:
                        if host_name not in host_list:
                            host_list.append(host_name)
                        host_ag_dict[host_name] = avai_wwpns

                    if len(wwpns) != len(avai_wwpns):
                        LOG.warning('_get_host_wwpn_from_cfg_file, '
                                    'some WWPNs format is not right. '
                                    'WWPN must be hexadecimal number '
                                    'with 16 characters. '
                                    'Cfg info: %(line)s, '
                                    'all WWPNs: %(WWPNs)s, '
                                    'available WWPNs: %(avai_wwpns)s.',
                                    {'line': line,
                                     'WWPNs': wwpns,
                                     'avai_wwpns': avai_wwpns})
                line = f.readline()

        if not host_list:
            msg = (_('_get_host_wwpn_from_cfg_file, '
                     'cfg file: %(path)s, '
                     'no host WWPN info in this file. '
                     'Please check the content of this file!')
                   % {'path': conf_file_name})
            LOG.error(msg)
            raise Exception(msg)

        LOG.debug('_get_host_wwpn_from_cfg_file, exit method. '
                  'cfg file name: %(path)s, '
                  'FC host name list: %(hosts)s, '
                  'FC host WWPNs dict: %(dict)s.',
                  {'path': conf_file_name,
                   'hosts': host_list,
                   'dict': host_ag_dict})
        return host_list, host_ag_dict

    def _check_wwpn_registered(self, wwpn):
        LOG.debug('_check_wwpn_registered, enter method. '
                  'WWPN: %s.', wwpn)

        wwpn_num = None
        rc, emsg, clidata = self._exec_eternus_cli('show_host_wwn_names')
        if rc != 0:
            msg = (_('_check_wwpn_registered, '
                     'show_host_wwn_names failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{Host Num:xxx,Host Name:xxx},...]
        for hostdata in clidata:
            if hostdata['Host Name'].lower() == wwpn.lower():
                wwpn_num = hostdata['Host Num']
                break

        LOG.debug('_check_wwpn_registered, exit method. '
                  'WWPN: %(WWPN)s, WWPN number on ETERNUS: %(number)s.',
                  {'WWPN': wwpn, 'number': wwpn_num})
        return wwpn_num

    def _register_wwpn(self, wwpn):
        LOG.debug('_register_wwpn, enter method. '
                  'WWPN: %s.', wwpn)

        wwpn_num = None
        option = {'wwn': wwpn}
        rc, emsg, clidata = (
            self._exec_eternus_cli('create_host_wwn_name', **option))
        if rc != 0:
            msg = (_('_register_wwpn, '
                     'create_host_wwn_name failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        wwpn_num = int(clidata[0], 16)

        LOG.debug('_register_wwpn, exit method. '
                  'WWPN: %(WWPN)s, WWPN number on ETERNUS: %(number)s.',
                  {'WWPN': wwpn, 'number': wwpn_num})
        return wwpn_num

    def _get_affinity_groups_by_host_number(self, host_num):
        LOG.debug('_get_affinity_groups_by_host_number, enter method. '
                  'WWPN number on ETERNUS: %s.', host_num)

        ag_list = []
        option = {'host-number': host_num,
                  'ce_support': self.ce_support}
        rc, emsg, clidata = (
            self._exec_eternus_cli('show_host_affinity_list', **option))
        if rc != 0:
            msg = (_('_get_affinity_groups_by_host_number, '
                     'show_host_affinity_list failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{Port:x,Host Num:x,Host Name:x,AG Num:x,AG Name:x},...]
        ag_list = list(set([x.get('AG Name') for x in clidata
                            if x.get('Port') in self.fc_ports]))

        LOG.debug('_get_affinity_groups_by_host_number, exit method. '
                  'WWPN number on ETERNUS: %(number)s, '
                  'affinity group name list: %(list)s.',
                  {'number': host_num, 'list': ag_list})
        return ag_list

    def _create_affinity_group(self, ag_name=None):
        LOG.debug('_create_affinity_group, enter method. '
                  'Affinity group name: %s.', ag_name)

        if ag_name:
            option = {'name': ag_name,
                      'volume-name': self.temp_vol_name,
                      'lun': 0}
        else:
            option = {'volume-name': self.temp_vol_name,
                      'lun': 0}

        rc, emsg, clidata = (
            self._exec_eternus_cli('create_affinity_group', **option))
        if rc != 0:
            msg = (_('_create_affinity_group, '
                     'create_affinity_group failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        if not ag_name:
            ag_num = int(clidata[0], 16)
            option = {'ag-number': ag_num}
            rc, emsg, clidata = (
                self._exec_eternus_cli('get_ag_name_by_num', **option))
            if rc != 0:
                msg = (_('_create_affinity_group, '
                         'get_ag_name_by_num failed. '
                         'Return code: %(rc)lu, '
                         'Error code:%(clidata)s, '
                         'Message:%(emsg)s.')
                       % {'rc': rc,
                          'clidata': clidata,
                          'emsg': emsg})
                LOG.error(msg)
                raise Exception(msg)

            ag_name = clidata

        LOG.debug('_create_affinity_group, exit method. '
                  'Affinity group name: %s.', ag_name)
        return ag_name

    def _get_host_affinity_by_ag_name(self, ag_name):
        LOG.debug('_get_host_affinity_by_ag_name, enter method. '
                  'Affinity group name: %s.', ag_name)

        option = {'ag-name': ag_name,
                  'ce_support': self.ce_support}
        rc, emsg, clidata = (
            self._exec_eternus_cli('show_host_affinity_list', **option))
        if rc != 0:
            msg = (_('_get_host_affinity_by_ag_name, '
                     'show_host_affinity_list failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{Port:x,Host Num:x,Host Name:x,AG Num:x,AG Name:x},...]
        wwpn_num_list = []  # [WWPN number,...]
        wwpn_num_ports_dict = {}  # {WWPN number: [port1,...],...}
        for line in clidata:
            wwpn_num = line.get('Host Num')
            port = line.get('Port')
            wwpn_num_list.append(wwpn_num)

            if port in self.fc_ports:
                temp_list = wwpn_num_ports_dict.get(wwpn_num, [])
                temp_list.append(port)
                wwpn_num_ports_dict[wwpn_num] = temp_list

        wwpn_num_list = list(set(wwpn_num_list))

        LOG.debug('_get_host_affinity_by_ag_name, exit method. '
                  'Affinity group name: %(ag_name)s, '
                  'WWPN number list: %(list)s, '
                  'WWPN port dict: %(dict)s.',
                  {'ag_name': ag_name,
                   'list': wwpn_num_list,
                   'dict': wwpn_num_ports_dict})
        return wwpn_num_list, wwpn_num_ports_dict

    def _write_configuration(self, path, objects):
        type = 'a' if objects else 'w'
        with open(path, type) as f:
            if objects:
                for obj in objects:
                    if obj.is_effective:
                        cfg = (','.join(obj.port_list) +
                               '=' + obj.wwpn +
                               '=' + obj.ag_name)
                        f.write(cfg)
                        f.write('\n')

    def _check_host_affinity(self, hostname, fc_zone_dict):
        LOG.debug('_check_host_affinity, enter method. '
                  'FC host name: %s.', hostname)

        try:
            # 1. Get WWPN list of target host.
            wwpn_list = self.host_wwpns_dict.get(hostname)

            LOG.info('_check_host_affinity, '
                     'FC host name: %(host)s, '
                     'WWPN list: %(list)s.',
                     {'host': hostname,
                      'list': wwpn_list})

            # 2. Check FC zone info for every WWPN.
            #    At least exist one WWPN setting in FC zone.
            host_ag_list = []
            is_exist = False
            for wwpn in wwpn_list:
                port_list = fc_zone_dict.get(wwpn)
                if not port_list:
                    LOG.warning('_check_host_affinity, '
                                'WWPN(%(WWPN)s) of host(%(host)s) '
                                'is not set in FC zone.',
                                {'host': hostname, 'WWPN': wwpn})
                else:
                    is_exist = True
                    host_ag = FCHostAffinityGroup()
                    host_ag.wwpn = wwpn
                    host_ag.port_list = port_list
                    host_ag.port_list_for_create = port_list
                    host_ag_list.append(host_ag)

            if not is_exist:
                msg = (_('_check_host_affinity, '
                         'all WWPNs(%(list)s) of host(%(host)s) '
                         'are not set in FC zone. '
                         'Please check the FC zone setting.')
                       % {'host': hostname, 'list': wwpn_list})
                LOG.error(msg)
                raise Exception(msg)

            # 3. Check WWPN if has registered on ETERNUS.
            exist_ag_list = []
            for host_ag in host_ag_list:
                wwpn = host_ag.wwpn

                # 3.1 Get WWPN number registered on ETERNUS.
                wwpn_num = self._check_wwpn_registered(wwpn)

                # If not exist, register WWPN on ETERNUS.
                if wwpn_num is None:
                    is_new_wwpn = True
                    wwpn_num = self._register_wwpn(wwpn)
                else:
                    is_new_wwpn = False

                host_ag.wwpn_num = wwpn_num
                host_ag.is_new_wwpn = is_new_wwpn

                if not is_new_wwpn:
                    # 3.2 Get exist ag name list by registered WWPN number.
                    exist_ag_list += (
                        self._get_affinity_groups_by_host_number(wwpn_num))

            # 4. Check exist affinity group list, make sure only has one exist.
            #    If no affinity group, create a new one.
            target_ag_name = None
            check_ag = False
            exist_ag_list = list(set(exist_ag_list))

            LOG.debug('_check_host_affinity, '
                      'FC host name: %(host)s, '
                      'WWPN list: %(WWPNS)s, '
                      'exist affinity group list on ETERNUS: %(list)s.',
                      {'host': hostname,
                       'WWPNS': wwpn_list,
                       'list': exist_ag_list})

            if len(exist_ag_list) > 1:
                msg = (_('_check_host_affinity, '
                         'the host(host name: %(host)s, WWPN list: %(WWPNS)s) '
                         'has more than one affinity group(%(ag_list)s). '
                         'Please check the ETERNUS configuration.')
                       % {'host': hostname,
                          'WWPNS': wwpn_list,
                          'ag_list': exist_ag_list})
                LOG.error(msg)
                raise Exception(msg)
            elif len(exist_ag_list) == 1:
                target_ag_name = exist_ag_list[0]
                check_ag = True
            else:
                # Create a new affinity group with TEMP_VOL_NAME.
                target_ag_name = self._create_affinity_group()

            LOG.debug('_check_host_affinity, '
                      'FC host name: %(host)s, '
                      'WWPN list: %(WWPNS)s, '
                      'target affinity group name: %(ag_name)s.',
                      {'host': hostname,
                       'WWPNS': wwpn_list,
                       'ag_name': target_ag_name})

            # 5. Check host affinity group.
            if check_ag:
                wwpn_num_list, wwpn_num_ports_dict = (
                    self._get_host_affinity_by_ag_name(target_ag_name))

                # WWPN numbers of target host.
                temp_list = [x.wwpn_num for x in host_ag_list]
                # Get WWPN numbers that are not in the target host.
                warn_wwpn_num_list = [x for x in wwpn_num_list
                                      if x not in temp_list]

                if warn_wwpn_num_list:
                    LOG.warning('_check_host_affinity, '
                                'Other hosts '
                                '(WWPN number list on ETERNUS: %(list)s) '
                                'share the same affinity '
                                'group with the target host(%(host)s). '
                                'Affinity group name: %(ag_name)s, '
                                'please check the ETERNUS configuration.',
                                {'host': hostname,
                                 'ag_name': target_ag_name,
                                 'list': warn_wwpn_num_list})

                # Get port list for setting host affinity.
                for idx, wwpn_num in enumerate(temp_list):
                    used_ports = wwpn_num_ports_dict.get(wwpn_num)
                    if used_ports:
                        # Get port list for target WWPN.
                        fc_zone_ports = host_ag_list[idx].port_list
                        # Get unused port list(create_ports) on ETERNUS.
                        create_ports = [x for x in fc_zone_ports
                                        if x not in used_ports]
                        host_ag_list[idx].port_list_for_create = create_ports

                # Add TEMP volume to the exist affinity group.
                flag = self._check_affinity_group(
                    self.temp_vol_name, target_ag_name)
                if not flag:
                    msg = (_('_check_host_affinity, '
                             'add TEMP volume '
                             'to the exist affinity group failed. '
                             'host name: %(host)s, '
                             'TEMP volume: %(volume)s, '
                             'affinity group name: %(ag_name)s.')
                           % {'host': hostname,
                              'volume': self.temp_vol_name,
                              'ag_name': target_ag_name})
                    LOG.error(msg)
                    raise Exception(msg)

            # 6. Set host affinity.
            for host_ag in host_ag_list:
                wwpn = host_ag.wwpn
                wwpn_num = host_ag.wwpn_num
                ports = host_ag.port_list_for_create
                if ports:
                    flag = self._set_host_affinity(
                        target_ag_name, wwpn_num, ports)
                    if not flag:
                        msg = (_('_check_host_affinity, '
                                 'set host affinity failed. '
                                 'host name: %(host)s, '
                                 'WWPN: %(WWPN)s, '
                                 'affinity group name: %(ag_name)s, '
                                 'setting port list: %(create_list)s.')
                               % {'host': hostname,
                                  'WWPN': wwpn,
                                  'ag_name': target_ag_name,
                                  'create_list': ports})
                        LOG.error(msg)
                        continue

                host_ag.ag_name = target_ag_name
                host_ag.is_effective = True
                temp_ports = host_ag.port_list
                LOG.debug('_check_host_affinity, set host affinity. '
                          'FC host name: %(host)s, '
                          'WWPN: %(WWPN)s, '
                          'affinity group name: %(ag_name)s, '
                          'port list: %(temp_list)s, '
                          'setting port list: %(create_list)s.',
                          {'WWPN': wwpn,
                           'host': hostname,
                           'ag_name': target_ag_name,
                           'temp_list': temp_ports,
                           'create_list': ports})

            # 7. Write configuration to the HOST_AG_MAP_FILE.
            map_file = self.cfg_dic.get(HOST_AG_MAP_FILE, None)
            self._write_configuration(map_file, host_ag_list)

            LOG.debug('_check_host_affinity, exit method. '
                      'FC host name: %s.', hostname)
        except Exception as e:
            msg = (_('_check_host_affinity, exit method. '
                     'FC host name: %(host)s, '
                     'exception or Error is occurred: %(error)s.')
                   % {'host': hostname, 'error': e})
            LOG.error(msg)

    def _get_host_affinity(self):
        LOG.debug('_get_host_affinity, enter method.')

        param_dict = {'ce_support': self.ce_support}
        rc, emsg, clidata = self._exec_eternus_cli('show_host_affinity_list',
                                                   **param_dict)
        if rc != 0:
            msg = (_('_get_host_affinity, '
                     'show_host_affinity_list failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{Port:x,Host Num:x,Host Name:x,AG Num:x,AG Name:x},...]
        # return: {WWPN number:{portlist:[port1,...],lungrouplist:[xxx,...]}}
        fc_host_ag_dict = {}
        for line in clidata:
            wwpn_num = line.get('Host Num')
            port = line.get('Port')
            ag_name = line.get('AG Name')

            if port in self.fc_ports:
                wwpn_info = fc_host_ag_dict.get(wwpn_num)
                if wwpn_info:
                    wwpn_info.get('portlist', []).append(port)
                    wwpn_info.get('lungrouplist', []).append(ag_name)
                else:
                    fc_host_ag_dict[wwpn_num] = {'portlist': [port],
                                                 'lungrouplist': [ag_name]}

        LOG.debug('_get_host_affinity, exit method. '
                  'FC host WWPN dict: %s.', fc_host_ag_dict)
        return fc_host_ag_dict

    def _get_wwpn_names(self):
        LOG.debug('_get_wwpn_names, enter method.')

        rc, emsg, clidata = self._exec_eternus_cli('show_host_wwn_names')
        if rc != 0:
            msg = (_('_get_wwpn_names, '
                     'show_host_wwn_names failed. '
                     'Return code: %(rc)lu, '
                     'Error code:%(clidata)s, '
                     'Message:%(emsg)s.')
                   % {'rc': rc,
                      'clidata': clidata,
                      'emsg': emsg})
            LOG.error(msg)
            raise Exception(msg)

        # clidata: [{Host Num:xxx,Host Name:xxx},...]
        # return: {WWPN:WWPN number}
        fc_wwpn_name_dict = {}
        for line in clidata:
            wwpn_num = line.get('Host Num')
            wwpn = line.get('Host Name').lower()
            fc_wwpn_name_dict[wwpn] = wwpn_num

        LOG.debug('_get_wwpn_names, exit method. '
                  'FC host WWPN dict: %s.', fc_wwpn_name_dict)
        return fc_wwpn_name_dict

    def _period_task(self):
        LOG.info('_period_task, enter method.')

        try:
            # 1. Read FC map file.(port1,port2,...=WWPN=affinity group)
            cfg_file = self.cfg_dic.get(HOST_AG_MAP_FILE, None)

            # wwpn_list: [WWPN,WWPN,...]
            # fc_map_dict: {WWPN:{portlist:[port1,port2,...],lungroup:xxx}}
            wwpn_list, fc_map_dict = self._get_host_ag_from_cfg_file(cfg_file)

            # 2. Get FC zone info and check with FC map file.
            # fc_zone_dict: {host WWPN: [port1, port2]}
            fc_zone_dict = self._get_enternus_fc_zone_info()
            for wwpn in wwpn_list:
                cfg_map_info = fc_map_dict.get(wwpn)
                cfg_ports = cfg_map_info.get('portlist')
                zone_ports = fc_zone_dict.get(wwpn)
                miss_ports = [x for x in cfg_ports if x not in zone_ports]
                if miss_ports:
                    msg = (_('_period_task, '
                             'some ports in FC map file are not in FC zone. '
                             'WWPN: %(WWPN)s, '
                             'FC map file: %(file)s, '
                             'ports in FC map file: %(cfg_ports)s, '
                             'ports in FC zone: %(zone_ports)s, '
                             'miss ports: %(miss_ports)s.')
                           % {'WWPN': wwpn,
                              'file': cfg_file,
                              'cfg_ports': cfg_ports,
                              'miss_ports': miss_ports,
                              'zone_ports': zone_ports})
                    LOG.error(msg)

            # 2. Use CLI: show host-affinity to get host affinity on ETERNUS.
            # fc_host_ag_dict:
            #    {WWPN number:{portlist:[port1,...],lungrouplist:[xxx,...]}}
            fc_host_ag_dict = self._get_host_affinity()

            # 3. Use CLI: show host-wwn-names to get all FC hosts on ETERNUS.
            # fc_wwpn_name_dict: {WWPN:WWPN number}
            fc_wwpn_name_dict = self._get_wwpn_names()

            # Three case will log error:
            #    1. Map relationship not exist:
            #       WWPN not registered or no host affinity.
            #    2. Affinity group error:
            #       more than one affinity group
            #       or not the right affinity group.
            #    3. Miss ports.
            for wwpn in wwpn_list:
                cfg_map_info = fc_map_dict.get(wwpn)
                cfg_ports = cfg_map_info.get('portlist')
                cfg_ag_name = cfg_map_info.get('lungroup')

                wwpn_num = fc_wwpn_name_dict.get(wwpn)
                if wwpn_num is not None:
                    eternus_map_info = fc_host_ag_dict.get(wwpn_num)
                    if eternus_map_info:
                        eternus_ports = eternus_map_info.get('portlist')
                        eternus_ags = eternus_map_info.get('lungrouplist')
                        temp_ags = list(set(eternus_ags))

                        if len(temp_ags) != 1:
                            # More than one affinity group on ETERNUS.
                            msg = (_('_period_task, '
                                     'more than one affinity group. '
                                     'Host WWPN: %(WWPN)s, '
                                     'ports configured in map file: '
                                     '%(cfg_ports)s, '
                                     'ports used on ETERNUS: %(used_ports)s, '
                                     'affinity group in map file: %(cfg_ag)s, '
                                     'affinity groups on ETERNUS: '
                                     '%(used_ags)s.')
                                   % {'WWPN': wwpn,
                                      'cfg_ports': cfg_ports,
                                      'used_ports': eternus_ports,
                                      'cfg_ag': cfg_ag_name,
                                      'used_ags': temp_ags})
                            LOG.error(msg)
                        else:
                            if eternus_ags[0] != cfg_ag_name:
                                # Not the right affinity group on ETERNUS.
                                msg = (_('_period_task, '
                                         'affinity group on ETERNUS is '
                                         'not the same with map file. '
                                         'Host WWPN: %(WWPN)s, '
                                         'ports configured in map file: '
                                         '%(cfg_ports)s, '
                                         'ports used on ETERNUS: '
                                         '%(used_ports)s, '
                                         'affinity group in map file: '
                                         '%(cfg_ag)s, '
                                         'affinity groups on ETERNUS: '
                                         '%(used_ags)s.')
                                       % {'WWPN': wwpn,
                                          'cfg_ports': cfg_ports,
                                          'used_ports': eternus_ports,
                                          'cfg_ag': cfg_ag_name,
                                          'used_ags': temp_ags})
                                LOG.error(msg)
                            else:
                                miss_ports = [x for x in cfg_ports
                                              if x not in eternus_ports]
                                if miss_ports:
                                    # Miss ports on ETERNUS.
                                    msg = (_('_period_task, '
                                             'some ports are '
                                             'not configured on ETERNUS. '
                                             'Host WWPN: %(WWPN)s, '
                                             'ports configured in map file: '
                                             '%(cfg_ports)s, '
                                             'ports used on ETERNUS: '
                                             '%(used_ports)s, '
                                             'ports not configured: '
                                             '%(miss_ports)s, '
                                             'affinity group in map file: '
                                             '%(cfg_ag)s, '
                                             'affinity groups on ETERNUS: '
                                             '%(used_ags)s.')
                                           % {'WWPN': wwpn,
                                              'cfg_ports': cfg_ports,
                                              'used_ports': eternus_ports,
                                              'miss_ports': miss_ports,
                                              'cfg_ag': cfg_ag_name,
                                              'used_ags': temp_ags})
                                    LOG.error(msg)
                    else:
                        # No host affinity on ETERNUS.
                        msg = (_('_period_task, '
                                 'no host affinity group on ETERNUS. '
                                 'Host WWPN: %(WWPN)s, '
                                 'WWPN number on ETERNUS: %(wwpn_num)s, '
                                 'ports configured in map file: %(ports)s, '
                                 'affinity group in map file: %(cfg_ag)s.')
                               % {'WWPN': wwpn,
                                  'wwpn_num': wwpn_num,
                                  'ports': cfg_ports,
                                  'cfg_ag': cfg_ag_name})
                        LOG.error(msg)
                else:
                    # WWPN not registered on ETERNUS.
                    msg = (_('_period_task, '
                             'no host affinity group on ETERNUS, '
                             'WWPN is not registered on ETERNUS. '
                             'Host WWPN: %(WWPN)s, '
                             'ports configured in map file: %(cfg_ports)s, '
                             'affinity group in map file: %(cfg_ag)s.')
                           % {'WWPN': wwpn,
                              'cfg_ports': cfg_ports,
                              'cfg_ag': cfg_ag_name})
                    LOG.error(msg)

            for hostname in self.f_hosts:
                wwpn_list = self.host_wwpns_dict.get(hostname)
                ag_name_list = []
                for wwpn in wwpn_list:
                    cfg_map_info = fc_map_dict.get(wwpn)
                    if cfg_map_info:
                        ag_name_list.append(cfg_map_info.get('lungroup'))

                ag_name_list = list(set(ag_name_list))
                if len(ag_name_list) > 1:
                    msg = (_('_period_task, '
                             'the host(host name: %(host)s, '
                             'WWPN list: %(WWPNS)s) '
                             'has more than one affinity group(%(ag_list)s). '
                             'Please check the ETERNUS configuration.')
                           % {'host': hostname,
                              'WWPNS': wwpn_list,
                              'ag_list': ag_name_list})
                    LOG.error(msg)

            LOG.info('_period_task, exit method.')
        except Exception as e:
            msg = (_('_period_task, exit method. '
                     'exception or Error is occurred: %(error)s.')
                   % {'error': e})
            LOG.error(msg)


def get_value_from_cfg_file(cfg_file):
    cfg_dic = {}
    with open(cfg_file) as f:
        line = f.readline()
        while line:
            if ('=' in line) and (not line.startswith('#')):
                infos = [x.strip() for x in line.split('=', 1)]
                cfg_dic[infos[0].lower()] = infos[1] if infos[1] else None
            line = f.readline()
    return cfg_dic


def check_monitor_cfg_file():
    if not os.path.exists(MONITOR_CFG_FILE):
        msg = (_('Monitor configuration file(%(path)s) is not exist!')
               % {'path': MONITOR_CFG_FILE})
        LOG.error(msg)
        raise Exception(msg)

    cfg_dic = get_value_from_cfg_file(MONITOR_CFG_FILE)

    driver_type = cfg_dic.get(DRIVER_TYPE, None)
    if not driver_type:
        msg = (_('Driver type(%(key)s) must be configured '
                 'in monitor configuration file(%(path)s)!')
               % {'path': MONITOR_CFG_FILE, 'key': DRIVER_TYPE})
        LOG.error(msg)
        raise Exception(msg)

    pool_name = cfg_dic.get(POOL_NAME, None)
    if not pool_name:
        msg = (_('Pool name for creating TEMP volume(%(key)s) '
                 'must be configured in monitor configuration file(%(path)s)!')
               % {'path': MONITOR_CFG_FILE, 'key': POOL_NAME})
        LOG.error(msg)
        raise Exception(msg)

    map_file = cfg_dic.get(HOST_AG_MAP_FILE, None)
    if not map_file:
        msg = (_('Host affinity map file(%(key)s) '
                 'must be configured in monitor configuration file(%(path)s)!')
               % {'path': MONITOR_CFG_FILE, 'key': HOST_AG_MAP_FILE})
        LOG.error(msg)
        raise Exception(msg)

    cfg_file = cfg_dic.get(OSVD_CONF, None)
    if not cfg_file:
        msg = (_('Volume driver cfg file(%(key)s) must be configured '
                 'in monitor configuration file(%(path)s)!')
               % {'path': MONITOR_CFG_FILE, 'key': OSVD_CONF})
        LOG.error(msg)
        raise Exception(msg)
    return cfg_dic

if __name__ == '__main__':
    cfg_dic = check_monitor_cfg_file()

    cfg_polling_time = cfg_dic.get(POLLING_TIME, None)
    if cfg_polling_time is None:
        polling_time = MIN_POLLING_TIME
    else:
        polling_time = int(cfg_polling_time)

    driver_type = cfg_dic.get(DRIVER_TYPE, None)
    LOG.debug('Monitor(type: %(type)s) start, polling time is %(time)s.',
              {'type': driver_type, 'time': polling_time})

    if polling_time < MIN_POLLING_TIME or polling_time > MAX_POLLING_TIME:
        msg = (_('Polling time should be in '
                 '[%(min_h)sh - %(max_h)sh (%(min)ss - %(max)ss)].')
               % {'min': MIN_POLLING_TIME,
                  'max': MAX_POLLING_TIME,
                  'min_h': (MIN_POLLING_TIME / 3600),
                  'max_h': (MAX_POLLING_TIME / 3600)})
        LOG.error(msg)
        raise Exception(msg)

    if driver_type.lower() == DRIVER_TYPE_ISCSI.lower():
        monitor = FJDXiSCSIMonitor()
    elif driver_type.lower() == DRIVER_TYPE_FC.lower():
        host_info = cfg_dic.get(HOST_INFO_FILE, None)
        if not host_info:
            msg = (_('FC host WWPNs cfg file(%(key)s) must be configured '
                     'in monitor configuration file(%(path)s)!')
                   % {'path': MONITOR_CFG_FILE, 'key': HOST_INFO_FILE})
            LOG.error(msg)
            raise Exception(msg)

        monitor = FJDXFCMonitor()
        monitor.create_fc_map_file()
    else:
        msg = (_('Monitor type(%(type)s) do not supported!!! '
                 'Supported driver list: %(list)s.')
               % {'type': driver_type,
                  'list': SUPPORT_DRIVER_TYPE_LIST})
        LOG.error(msg)
        raise Exception(msg)

    monitor.polling_check(polling_time)
