# Copyright (c) 2019 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""iSCSI Cinder Volume driver for Fujitsu ETERNUS DX S3 series."""

from oslo_log import log as logging

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.fujitsu.eternus_dx import constants as CONSTANTS
from cinder.volume.drivers.fujitsu.eternus_dx import eternus_dx_common
from cinder.volume.drivers.fujitsu.eternus_dx \
    import eternus_dx_utils as dx_utils

LOG = logging.getLogger(__name__)


@interface.volumedriver
class FJDXISCSIDriver(driver.ISCSIDriver):
    """ETERNUS iSCSI Cinder Volume Driver for Fujitsu ETERNUS DX S3 series."""

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Fujitsu_ETERNUS_CI"
    VERSION = eternus_dx_common.FJDXCommon.VERSION

    def __init__(self, *args, **kwargs):

        super(FJDXISCSIDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(
            CONSTANTS.FJ_ETERNUS_DX_OPT_list)
        self.common = (
            eternus_dx_common.FJDXCommon(
                'iSCSI', configuration=self.configuration,
                active_backend_id=kwargs.get('active_backend_id')))
        self.common.host = self.host
        self.common.filter_function = self.get_filter_function()
        self.VERSION = self.common.VERSION
        self.semaphore_num = self.common.semaphore_num
        self.semaphore = self.common.semaphore
        self._stats = None

    def check_for_setup_error(self):
        """Returns an error if prerequisites aren't met."""
        if not self.configuration.cinder_eternus_config_file:
            raise exception.InvalidInput(
                reason=_('Cinder_eternus_config_file must be set.'))

    def create_volume(self, volume):
        """Create volume."""
        LOG.debug('create_volume, '
                  'volume id: %s, Enter method.', volume['id'])

        d_metadata = dx_utils.get_metadata(volume)

        # Choice: standard create volume or CCM create volume.
        if CONSTANTS.FJ_SRC_VOL_ID not in d_metadata:
            # Standard create volume.
            if self.semaphore:
                # In the case of using semaphore, if semaphore is not acquired,
                # the Create queue has reached the upper limit.
                if not self.semaphore.acquire(blocking=False):
                    msg = (_('create_volume, volume id: %(vid)s '
                             'semaphore is not acquired. There are already '
                             '%(num)s create-requests in the queue.')
                           % {'vid': volume['id'],
                              'num': self.semaphore_num})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                remain = self.semaphore.counter - len(self.semaphore._waiters)
                LOG.debug('create_volume, volume id: %(vid)s acquire '
                          'semaphore, remain %(remain)s.',
                          {'vid': volume['id'],
                           'remain': remain})
            try:
                model_update = self.common.create_volume(
                    volume, timeout=self.common.set_sec)
            finally:
                if self.semaphore:
                    self.semaphore.release()
                    remain = (self.semaphore.counter -
                              len(self.semaphore._waiters))
                    LOG.debug('create_volume, volume id: %(vid)s release '
                              'semaphore, remain %(remain)s.',
                              {'vid': volume['id'],
                               'remain': remain})
        else:
            # CCM create volume.
            # Create cloned volume for backup solution.
            # Get source and target volume information.
            change_metadata = {}
            ctxt = context.get_admin_context()
            volume_size = volume['size']
            src_volid = d_metadata[CONSTANTS.FJ_SRC_VOL_ID]
            if src_volid is None:
                msg = _('create_volume, Source volume id is None.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Modify metadata value Mirror to 'mirror'.
            if CONSTANTS.FJ_REMOTE_DEST_META in d_metadata:
                change_metadata[CONSTANTS.FJ_REMOTE_DEST_META] = (
                    (d_metadata[CONSTANTS.FJ_REMOTE_DEST_META]).lower())

            # Get info of source volume.
            src_vref = self.db.volume_get(ctxt, src_volid)

            # Check target volume information with reference to source volume.
            src_vref_size = src_vref['size']
            src_vref_status = src_vref['status']

            if 'error' in src_vref_status or 'deleting' in src_vref_status:
                msg = (_('Invalid volume status: %(status)s.')
                       % {'status': src_vref_status})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            elif volume_size < src_vref_size:
                msg = (_('Volume size %(volume_size)sGB '
                         'cannot be smaller than '
                         'original volume size %(src_vref_size)sGB. '
                         'They must be >= original volume size.')
                       % {'volume_size': volume_size,
                          'src_vref_size': src_vref_size})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('create_volume, Call create cloned volume.')
            model_update = self.common.create_cloned_volume(
                volume,
                src_vref,
                use_service_name=True,
                timeout=self.common.set_sec)

            # Post processing.
            try:
                if src_vref.bootable:
                    self.db.volume_update(ctxt,
                                          volume['id'],
                                          {'bootable': True})
                    self.db.volume_glance_metadata_copy_from_volume_to_volume(
                        ctxt, src_volid, volume['id'])

                self.db.volume_update(ctxt,
                                      volume['id'],
                                      {'source_volid': src_volid})
            except Exception as ex:
                msg = (_('create volume, Failed updating volume metadata, '
                         'reason: %(reason)s, volume_id: %(volume_id)s, '
                         'src_volid: %(src_volid)s.')
                       % {'reason': ex,
                          'volume_id': volume['id'],
                          'src_volid': src_volid})
                raise exception.MetadataCopyFailure(reason=msg)

            model_update['metadata'].update(change_metadata)

        LOG.debug('create_volume, '
                  'info: %s, Exit method.', model_update['metadata'])
        return model_update

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot."""
        LOG.debug('create_volume_from_snapshot, '
                  'volume id: %(vid)s, snap id: %(sid)s, Enter method.',
                  {'vid': volume['id'], 'sid': snapshot['id']})

        model_update = (
            self.common.create_volume_from_snapshot(
                volume, snapshot, timeout=self.common.set_sec))

        LOG.debug('create_volume_from_snapshot, '
                  'info: %s, Exit method.', model_update['metadata'])
        return model_update

    def create_cloned_volume(self, volume, src_vref):
        """Create cloned volume."""
        LOG.debug('create_cloned_volume, '
                  'target volume id: %(tid)s, '
                  'source volume id: %(sid)s, Enter method.',
                  {'tid': volume['id'], 'sid': src_vref['id']})

        model_update = self.common.create_cloned_volume(
            volume, src_vref, timeout=self.common.set_sec)

        LOG.debug('create_cloned_volume, '
                  'info: %s, Exit method.', model_update['metadata'])
        return model_update

    def delete_volume(self, volume):
        """Delete volume on ETERNUS."""
        LOG.debug('delete_volume, '
                  'volume id: %s, Enter method.', volume['id'])

        vol_exist = self.common.delete_volume(
            volume, timeout=self.common.set_sec)

        LOG.debug('delete_volume, '
                  'delete: %s, Exit method.', vol_exist)

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        LOG.debug('create_snapshot, '
                  'snap id: %(sid)s, volume id: %(vid)s, Enter method.',
                  {'sid': snapshot['id'], 'vid': snapshot['volume_id']})

        model_update = self.common.create_snapshot(
            snapshot, timeout=self.common.set_sec)

        LOG.debug('create_snapshot, '
                  'info: %s, Exit method.', model_update['metadata'])
        return model_update

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        LOG.debug('delete_snapshot, '
                  'snap id: %(sid)s, volume id: %(vid)s, Enter method.',
                  {'sid': snapshot['id'], 'vid': snapshot['volume_id']})

        vol_exist = self.common.delete_snapshot(snapshot)

        LOG.debug('delete_snapshot, '
                  'delete: %s, Exit method.', vol_exist)

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume."""
        pass

    def create_export(self, context, volume, connector):
        """Driver entry point to get the export info for a new volume."""
        pass

    def remove_export(self, context, volume):
        """Driver entry point to remove an export for a volume."""
        pass

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""
        initiator = connector.get('initiator') if connector else None

        LOG.debug('initialize_connection, volume id: %(vid)s, '
                  'initiator: %(initiator)s, Enter method.',
                  {'vid': volume['id'], 'initiator': initiator})

        info = self.common.initialize_connection(
            volume, connector, timeout=self.common.set_sec)

        LOG.debug('initialize_connection, '
                  'info: %s, Exit method.', info)
        return info

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        initiator = connector.get('initiator') if connector else None

        LOG.debug('terminate_connection, volume id: %(vid)s, '
                  'initiator: %(initiator)s, Enter method.',
                  {'vid': volume['id'], 'initiator': initiator})

        map_info = self.common.terminate_connection(
            volume, connector, timeout=self.common.set_sec)

        LOG.debug('terminate_connection, '
                  'map_num: %s, Exit method.', map_info.get('map_num'))

    def get_volume_stats(self, refresh=False):
        """Get volume stats."""
        LOG.debug('get_volume_stats, refresh: %s, Enter method.', refresh)

        pool_name = None
        if refresh:
            data, pool_name = self.common.update_volume_stats(
                timeout=self.common.query_sec)
            backend_name = self.configuration.safe_get('volume_backend_name')
            data['volume_backend_name'] = backend_name or 'FJDXISCSIDriver'
            data['storage_protocol'] = 'iSCSI'
            if self.common.remote_backend_name:
                data['replication_enabled'] = True
                data['replication_targets'] = self.common.remote_backend_name

            self._stats = data

        LOG.debug('get_volume_stats, '
                  'pool name: %s, Exit method.', pool_name)
        return self._stats

    def extend_volume(self, volume, new_size):
        """Extend volume."""
        LOG.debug('extend_volume, '
                  'volume id: %s, Enter method.', volume['id'])

        used_pool_name = self.common.extend_volume(
            volume, new_size, timeout=self.common.set_sec)

        LOG.debug('extend_volume, '
                  'used pool name: %s, Exit method.', used_pool_name)

    def manage_existing(self, volume, existing_ref):
        """Bring an existing volume under Cinder management."""
        LOG.debug('manage_existing, volume id: %(vid)s, '
                  'id or name: %(ref)s, Enter method.',
                  {'vid': volume['id'], 'ref': existing_ref})

        info = self.common.manage_existing(
            'volume', volume, existing_ref, timeout=self.common.set_sec)

        LOG.debug('manage_existing, '
                  'info: %s, Exit method.', info['metadata'])
        return info

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage existing."""
        LOG.debug('manage_existing_get_size, volume id: %(vid)s, '
                  'id or name: %(ref)s, Enter method.',
                  {'vid': volume['id'], 'ref': existing_ref})

        vol_size = self.common.manage_existing_get_size(
            'volume', volume, existing_ref, timeout=self.common.set_sec)

        LOG.debug('manage_existing_get_size, '
                  'volume size: %s, Exit method.', vol_size)
        return vol_size

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management."""
        pass

    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Bring an existing snapshot under Cinder management."""
        LOG.debug('manage_existing_snapshot, snapshot id: %(sid)s, '
                  'id or name: %(ref)s, Enter method.',
                  {'sid': snapshot['id'], 'ref': existing_ref})

        info = self.common.manage_existing(
            'snapshot', snapshot, existing_ref, timeout=self.common.set_sec)

        LOG.debug('manage_existing_snapshot, '
                  'info: %s, Exit method.', info['metadata'])
        return info

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return size of snapshot to be managed by manage existing."""
        LOG.debug('manage_existing_snapshot_get_size, snapshot id: %(sid)s, '
                  'id or name: %(ref)s, Enter method.',
                  {'sid': snapshot['id'], 'ref': existing_ref})

        snap_size = self.common.manage_existing_get_size(
            'snapshot', snapshot, existing_ref, timeout=self.common.set_sec)

        LOG.debug('manage_existing_snapshot_get_size, '
                  'volume size: %s, Exit method.', snap_size)
        return snap_size

    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Cinder management."""
        pass

    def update_migrated_volume(self, ctxt, volume, new_volume,
                               original_volume_status):
        """Update migrated volume."""
        LOG.debug('update_migrated_volume, '
                  'source volume id: %(s_id)s, '
                  'target volume id: %(t_id)s, Enter method.',
                  {'s_id': volume['id'], 't_id': new_volume['id']})

        model_update = self.common.update_migrated_volume(
            ctxt, volume, new_volume)

        LOG.debug('update_migrated_volume, '
                  'target volume meta: %s, Exit method.', model_update)

        return model_update

    def get_manageable_volumes(self, cinder_volumes,
                               marker, limit, offset, sort_keys, sort_dirs):
        """List volumes on the backend available for management by Cinder."""

        return self.common._get_manageable_resource_info(
            cinder_volumes, 'volume', marker, limit, offset,
            sort_keys, sort_dirs, timeout=self.common.query_sec)

    def get_manageable_snapshots(self, cinder_snapshots,
                                 marker, limit, offset, sort_keys, sort_dirs):
        """List snapshots on the backend available for management by Cinder."""

        return self.common._get_manageable_resource_info(
            cinder_snapshots, 'snapshot', marker, limit,
            offset, sort_keys, sort_dirs, timeout=self.common.query_sec)

    def failover_host(self, context, volumes, secondary_id=None, groups=None):
        """Failover volumes to a secondary host/ backend."""
        LOG.debug('failover_host, secondary_id: %s, Enter method.',
                  secondary_id)
        secondary_id, volume_update_list, group_update_list = (
            self.common.failover_host(volumes,
                                      secondary_id,
                                      groups,
                                      timeout=self.common.set_sec))
        LOG.debug('failover_host, volume_update_list: %s, Exit method.',
                  str(volume_update_list))
        return secondary_id, volume_update_list, group_update_list

    def revert_to_snapshot(self, context, volume, snapshot):
        """Revert volume to snapshot."""
        return self.common._revert_to_snapshot(volume,
                                               snapshot,
                                               timeout=self.common.set_sec)
