# Copyright (c) 2019 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cinder Volume driver for Fujitsu ETERNUS DX S3 series."""

import six

from cinder.i18n import _
from cinder import ssh_utils
import time


class FJDXCLI(object):
    """ETERNUS CLI Code."""

    def __init__(self, user, storage_ip, password=None, keyfile=None):
        """Constructor."""
        self.user = user
        self.storage_ip = storage_ip
        if password and keyfile:
            raise Exception(_('can not specify both password and keyfile'))

        self.use_ipv6 = False
        if password:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              password=password, max_size=2)
            if storage_ip.find(':') != -1:
                self.use_ipv6 = True

        if keyfile:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              privatekey=keyfile, max_size=2)
            if storage_ip.find(':') != -1:
                self.use_ipv6 = True

        self.ce_support = False

        self.CMD_dic = {'check_user_role': self._check_user_role,
                        'expand_volume': self._expand_volume,
                        'format_volume': self._format_volume,
                        'format_tpv': self._format_tpv,
                        'set_volume_qos': self._set_volume_qos,
                        'show_volume_progress': self._show_volume_progress,
                        'show_tpv_progress': self._show_tpv_progress,
                        'show_host_wwn_names': self._show_host_wwn_names,
                        'show_host_iscsi_names': self._show_host_iscsi_names,
                        'create_host_wwn_name': self._create_host_wwn_name,
                        'create_host_iscsi_name': self._create_host_iscsi_name,
                        'create_affinity_group': self._create_affinity_group,
                        'delete_affinity_group': self._delete_affinity_group,
                        'show_affinity_group': self._show_affinity_group,
                        'show_iscsi_parameters': self._show_iscsi_parameters,
                        'show_fc_parameters': self._show_fc_parameters,
                        'set_affinity_group': self._set_affinity_group,
                        'show_host_affinity': self._show_host_affinity,
                        'show_host_affinity_list': self._show_host_ag_list,
                        'set_host_affinity': self._set_host_affinity,
                        'show_pool_provision': self._show_pool_provision,
                        'discover_host_wwn_names': self._discover_wwn_names,
                        'show_affinity_group_list': self._show_ag_list,
                        'create_volume': self._create_volume,
                        'get_ag_name_by_num': self._get_ag_name_by_num,
                        'show_raid_groups': self._show_raid_groups,
                        'show_thin_pro_pools': self._show_thin_pro_pools,
                        'show_model_supplement_info': (
                            self._show_model_supplement_info)}

        self.SMIS_dic = {'0000': '0',  # Success
                         '0060': '32787',  # The device is in busy state
                         '0100': '4097'}  # Size not supported

    def done(self, command, **option):
        """Done ETERNUS CLI."""
        func = self.CMD_dic.get(command, self._default_func)
        return func(**option)

    def _exec_cli(self, cmd, StrictHostKeyChecking=True, **option):
        """Execute CLI command with arguments."""
        exec_cmdline = cmd + self._get_option(**option)
        try:
            ssh = self.ssh_pool.get()
            chan = ssh.invoke_shell()
            chan.send(exec_cmdline + '\n')
            while not chan.recv_ready():
                time.sleep(0.1)
            stdoutdata = ''
            start_time = time.time()
            while True:
                stdoutdata += str(chan.recv(65535))
                # CLI command end with 'CLI>'.
                if stdoutdata == '\r\nCLI> ':
                    continue
                if (stdoutdata[len(stdoutdata) - 5: len(stdoutdata) - 1] ==
                        'CLI>'):
                    break
                # Timeout: 300s.
                duration = time.time() - start_time
                if 300 <= duration:
                    raise Exception("Execute command: %s,"
                                    "timeout!" % exec_cmdline)
        except Exception as e:
            raise Exception(_("Execute CLI "
                              "command error. Error: %s") % six.text_type(e))
        finally:
            self.ssh_pool.put(ssh)
            self.ssh_pool.remove(ssh)

        output = []
        message = []
        stdoutlist = stdoutdata.split('\r\n')
        output_header = ""

        for no, outline in enumerate(stdoutlist):
            if len(outline) <= 0 or outline is None:
                continue

            if not output_header.endswith(exec_cmdline):
                output_header += outline
                continue

            if 0 <= outline.find('Error'):
                raise Exception(_("Output: %(outline)s: "
                                  "Command: %(cmdline)s")
                                % {'outline': outline,
                                   'cmdline': exec_cmdline})

            if not self._is_status(outline):
                continue

            status = int(outline, 16)
            lineno = no + 1
            break
        else:
            raise Exception(_(
                "Invalid CLI output: %(exec_cmdline)s, %(stdoutlist)s")
                % {'exec_cmdline': exec_cmdline,
                   'stdoutlist': stdoutlist})

        if status == 0:
            rc = '0'
            for outline in stdoutlist[lineno:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                message.append(outline)
        else:
            code = stdoutlist[lineno]
            for outline in stdoutlist[lineno + 1:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                output.append(outline)

            rc, message = self._create_error_message(code, output)

        return {'result': 0, 'rc': rc, 'message': message}

    def _create_error_message(self, code, msg):
        """Create error code and message using arguements."""
        message = None
        if code in self.SMIS_dic:
            rc = self.SMIS_dic[code]
        else:
            rc = 'E' + code

            if rc == "E0001":
                message = "Bad value: %s" % msg
            elif rc == "ED184":
                message = "Because OPC is being executed, "
                "the processing was discontinued."
            else:
                message = msg

        return rc, message

    def _is_status(self, value):
        """Check whether input value is status value or not."""
        try:
            if len(value) != 2:
                return False

            int(value, 16)
            int(value[0], 16)
            int(value[1], 16)

            return True
        except ValueError:
            return False

    def _get_option(self, **option):
        """Create option strings from dictionary."""
        ret = ""
        for key, value in six.iteritems(option):
            ret += " -%(key)s %(value)s" % {'key': key, 'value': value}
        return ret

    def _default_func(self, **option):
        """Default function."""
        raise Exception(_("Invalid function is specified"))

    def _check_user_role(self, **option):
        """Check user role."""
        try:
            output = self._exec_cli("show users",
                                    StrictHostKeyChecking=False,
                                    **option)
            userlist = output.get('message')
            role = None
            for userinfo in userlist:
                username = userinfo.split('\t')[0]
                if username == self.user:
                    role = userinfo.split('\t')[1]
                    break

            output['message'] = role
        except Exception as ex:
            if 'show users' in six.text_type(ex):
                msg = ("Specified user(%s) does not have Software role"
                       % self.user)
            elif 'Error connecting' in six.text_type(ex):
                msg = (six.text_type(ex)[34:] +
                       ', Please check private_key_path')
            else:
                msg = six.text_type(ex)
            output = {
                'result': 0,
                'rc': '4',
                'message': msg
            }
        return output

    def _expand_volume(self, **option):
        """Exec expand volume."""
        return self._exec_cli("expand volume", **option)

    def _format_volume(self, **option):
        """Exec format volume."""
        return self._exec_cli("format volume", **option)

    def _format_tpv(self, **option):
        """Exec format tpv."""
        thick_option = {
            'volume-name': option['volume-name'],
            'allocation': 'thick'
        }

        output = self._exec_cli("set volume-parameters", **thick_option)
        rc = output['rc']
        if rc != "0":
            return output

        return self._exec_cli("format volume", **option)

    def _set_volume_qos(self, **option):
        """Exec set volume-qos."""
        return self._exec_cli("set volume-qos", **option)

    def _show_volume_progress(self, **option):
        """Exec show vollume-progress."""
        output = self._exec_cli("show volume-progress", **option)

        # Return error
        rc = output['rc']
        if rc != "0":
            return output

        # Get remain time.
        message = output['message']
        volume_count = message[0]
        if volume_count == "00":
            # Format completion.
            output['message'] = '0'
            return output
        else:
            remain_minute = 1
            remain_time_str = message[1].split("\t")[9]

            if remain_time_str != "FFFFFFFF":
                remain_minute = int(remain_time_str, 16) + 1

            output['message'] = str(remain_minute * 60)

        return output

    def _show_tpv_progress(self, **option):
        """Exec show tpv-progress."""
        output = self._exec_cli("show thin-pro-pools", **option)
        rc = output['rc']
        if rc != "0":
            return output

        message = output['message']
        pool_count = int(message[1], 16)
        raid_count_lineno = 2 + pool_count
        raid_count = int(message[raid_count_lineno], 16)

        for num in range(raid_count):
            raid_info = message[raid_count_lineno + num + 1]
            raid_status = int(raid_info.split('\t')[4], 16)
            CNFG_LU_STATUS_BIT_QF = raid_status & 0x0100

            if CNFG_LU_STATUS_BIT_QF != 0:
                output['message'] = '600'
                break
        else:
            output['message'] = '0'

        return output

    def _show_host_wwn_names(self, **option):
        """Get host wwn names."""
        try:
            output = self._exec_cli("show host-wwn-names", **option)

            # Return error.
            rc = output['rc']

            if rc != "0":
                return output

            hostdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                hostdatalist.append({'Host Num': int(clidata[0], 16),
                                     'Host Name': clidata[2]})

            output['message'] = hostdatalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host wwn names error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_host_iscsi_names(self, **option):
        """Get host iscsi names."""
        try:
            output = self._exec_cli("show host-iscsi-names", **option)

            # Return error.
            rc = output['rc']

            if rc != "0":
                return output

            hostdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                hostdatalist.append({'Host Num': int(clidata[0], 16),
                                     'Host Name': clidata[5]})

            output['message'] = hostdatalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host iscsi names error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_affinity_group(self, **option):
        """Get affinity group information."""
        try:
            output = self._exec_cli("show affinity-group", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[3:]:
                clidata = clidataline.split('\t')
                datalist.append({'Lun Num': int(clidata[0], 16),
                                 'Lun Name': clidata[2]})

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host iscsi names error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_pool_provision(self, **option):
        """Get TPP provision capacity information."""
        try:
            output = self._exec_cli("show volumes", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')

            data = 0
            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                if clidata[0] == 'FFFF':
                    break
                data += int(clidata[7], 16)
            provision = data / 2097152

            output['message'] = provision
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show pool provision capacity error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_iscsi_parameters(self, **option):
        """Get iscsi port information."""

        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show iscsi-parameters", **option)

            rc = output['rc']

            if rc != "0":
                return output

            ipv4_data_list = []
            ipv6_data_list = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                tmpport = clidata[0] + clidata[1]
                port_str = self._get_port(clidata[0], clidata[1])
                if self.use_ipv6:
                    if clidata[31] != '::':
                        ipv6_data_list.append({'port num': tmpport,
                                               'port iqn': clidata[4],
                                               'port ip': '[' + clidata[31] +
                                                          ']',
                                               'port str': port_str})
                else:
                    if clidata[10] != '0.0.0.0':
                        ipv4_data_list.append({'port num': tmpport,
                                               'port iqn': clidata[4],
                                               'port ip': clidata[10],
                                               'port str': port_str})

            if self.use_ipv6:
                output['message'] = ipv6_data_list
            else:
                output['message'] = ipv4_data_list
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show iscsi-parameters error: %s"
                           % six.text_type(ex)
            }

        return output

    def _show_host_affinity(self, **option):
        """Get host affinity group information."""
        try:
            output = self._exec_cli("show host-affinity", **option)

            rc = output['rc']

            if rc != "0":
                return output

            host_ag_dict = {}
            ag_dict = {}
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                if len(clidata) == 2:
                    tmplist = list(clidata[0])
                    tmp_port = tmplist[5] + (
                        str(int(tmplist[4]) - 4) + tmplist[7])
                    continue
                elif len(clidata) > 2:
                    host_num = int(clidata[0], 16)
                    lun_group = clidata[3]
                    ag_dict = host_ag_dict.get(host_num, {})
                    ag_dict[tmp_port] = lun_group
                    host_ag_dict[host_num] = ag_dict

            output['message'] = host_ag_dict
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "show iscsi-parameters error: %s"
                                 % six.text_type(ex)}

        return output

    def _show_host_ag_list(self, **option):

        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show host-affinity", **option)

            rc = output['rc']

            if rc != "0":
                return output

            # [{Port:xxx,Host Num:xxx,Host Name:xxx,AG Num:xxx,AG Name:xxx}]
            datalist = []
            clidatalist = output.get('message')

            for line in clidatalist:
                infos = line.split()
                if len(infos) == 2:
                    port_str = self._get_port(
                        infos[0][-4:-2], infos[0][-2:])
                if len(infos) == 7 or len(infos) == 6:
                    host_ag = {'Port': port_str,
                               'Host Num': int(infos[0], 16),
                               'Host Name': infos[1],
                               'AG Num': int(infos[2], 16),
                               'AG Name': infos[3]}
                    datalist.append(host_ag)

            output['message'] = datalist
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "discover host-wwn-names: %s"
                                 % six.text_type(ex)}

        return output

    def _show_ag_list(self, **option):
        """Get affinity group list."""
        try:
            output = self._exec_cli("show affinity-group", **option)
            rc = output['rc']
            if rc != "0":
                return output

            datalist = []
            clidatalist = output.get('message')
            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                datalist.append(clidata[1])
            output['message'] = datalist
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "show host iscsi names error: %s"
                                 % six.text_type(ex)}

        return output

    def _show_model_supplement_info(self, **option):
        """Show model supplement info."""
        try:
            output = self._exec_cli("show model-supplement-info", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')
            infos = clidatalist[0].split()
            output['message'] = {'machine_info': infos[0],
                                 'model_info': infos[1]}

        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "show model-supplement-info error: %s"
                                 % six.text_type(ex)}

        return output

    def _get_port(self, ca_module_id, port_number):
        if self.ce_support:
            return self._get_port_str_v2(ca_module_id, port_number)
        else:
            return self._get_port_str(ca_module_id, port_number)

    def _get_port_str(self, ca_module_id, port_number):
        cm = int(ca_module_id[1], 16) % 8
        ca = int(ca_module_id[0], 16) % 4
        port = ((int(ca_module_id[1], 16) / 4) / 2) * 2 + (
            int(port_number[1], 16) % 2)
        temp_str = str(hex(cm)) + str(hex(ca)) + str(hex(port))
        return temp_str.replace('0x', '').upper()

    def _get_port_str_v2(self, ca_module_id, port_number):
        # port_str format: CE + CM + CA + PORT
        # CE: 0,1,2,3,4,5,6,7,8,9,10,11
        # CM: 0,1
        # CA: 0,1,2,3
        # PORT: 0,1,2,3
        base = 4
        ce = ((int(ca_module_id[1], 16) % 8) // 2) + (
            ((int(ca_module_id[0], 16) - base) // 4) * 4)
        cm = int(ca_module_id[1], 16) % 2
        ca = int(ca_module_id[0], 16) % 4
        port = ((int(ca_module_id[1], 16) // 4) // 2) * 2 + (
                int(port_number[1], 16) % 2)
        temp_str = str(hex(ce)) + str(hex(cm)) + str(hex(ca)) + str(hex(port))
        return temp_str.replace('0x', '').upper()

    def _create_host_wwn_name(self, **option):
        """Create host wwn name."""
        return self._exec_cli("create host-wwn-name", **option)

    def _create_host_iscsi_name(self, **option):
        """Create host iscsi name."""
        return self._exec_cli("create host-iscsi-name", **option)

    def _create_affinity_group(self, **option):
        """Create affinity group."""
        return self._exec_cli("create affinity-group", **option)

    def _delete_affinity_group(self, **option):
        """Delete affinity group."""
        return self._exec_cli("delete affinity-group", **option)

    def _set_host_affinity(self, **option):
        """Set host affinity."""
        return self._exec_cli("set host-affinity", **option)

    def _set_affinity_group(self, **option):
        """Set affinity group."""
        return self._exec_cli("set affinity-group", **option)

    def _create_volume(self, **option):
        """Create volume."""
        return self._exec_cli("create volume", **option)

    def _show_fc_parameters(self, **option):
        """Get fc port information."""

        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show fc-parameters", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                tmpport = clidata[0] + clidata[1]
                port_str = self._get_port(clidata[0], clidata[1])
                datalist.append({'port num': tmpport,
                                 # 'port wwpn': clidata[17],
                                 'port str': port_str})

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "discover host-wwn-names: %s"
                           % six.text_type(ex)
            }

        return output

    def _discover_wwn_names(self, **option):

        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("discover host-wwn-names", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = {}  # {host WWPN: [port1, port2]}
            clidatalist = output.get('message')

            port_str = None
            for idx, line in enumerate(clidatalist):
                infos = line.split()
                if len(infos) == 3:
                    port_str = self._get_port(infos[0], infos[1])
                    port_idx = idx
                    continue
                if port_str and idx != port_idx + 1:
                    wwpn = infos[0]
                    temp_list = datalist.get(wwpn, [])
                    temp_list.append(port_str)
                    datalist[wwpn] = temp_list

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "discover host-wwn-names: %s"
                           % six.text_type(ex)
            }

        return output

    def _get_ag_name_by_num(self, **option):
        """Get affinity group list."""
        try:
            output = self._exec_cli("show affinity-group", **option)
            rc = output['rc']
            if rc != "0":
                return output

            output['message'] = output.get('message')[1].split()[1]
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "show host iscsi names error: %s"
                                 % six.text_type(ex)}

        return output

    def _show_raid_groups(self, **option):
        """show raid-groups"""
        try:
            output = self._exec_cli("show raid-groups", **option)
            rc = output['rc']
            if rc != "0":
                return output

            datalist = output.get('message')
            out = []
            for line in datalist:
                infos = line.split()
                if len(infos) > 1:
                    out.append({'Name': infos[1]})

            output['message'] = out
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "show raid-groups error: %s"
                                 % six.text_type(ex)}

        return output

    def _show_thin_pro_pools(self, **option):
        """show thin-pro-pools"""
        try:
            output = self._exec_cli("show thin-pro-pools", **option)
            rc = output['rc']
            if rc != "0":
                return output

            datalist = output.get('message')
            out = []
            for line in datalist:
                infos = line.split()
                if len(infos) > 1:
                    out.append({'Name': infos[1]})

            output['message'] = out
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "show thin-pro-pools error: %s"
                                 % six.text_type(ex)}

        return output
