# Copyright (c) 2019 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cinder Volume driver for Fujitsu ETERNUS DX S3 series."""

import base64
import hashlib
import os
import re
import time

from eventlet.semaphore import BoundedSemaphore
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import excutils
from oslo_utils import units
import six
from xml.etree.ElementTree import parse as tree_parse

from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume.drivers.fujitsu.eternus_dx import constants as CONSTANTS
from cinder.volume.drivers.fujitsu.eternus_dx import eternus_dx_cli
from cinder.volume.drivers.fujitsu.eternus_dx \
    import eternus_dx_utils as dx_utils
from cinder.volume.drivers.fujitsu.eternus_dx import https
from cinder.volume import manager as volume_manager
from cinder.volume import qos_specs
from cinder.volume import utils as volume_utils
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)

try:
    import pywbem
except ImportError:
    msg = ('import pywbem failed!! '
           'pywbem is necessary for this volume driver.')
    LOG.error(msg)


class FJDXCommon(object):
    """Common code that does not depend on protocol."""

    VERSION = "1.6.1.2"
    stats = {
        'driver_version': VERSION,
        'reserved_percentage': 0,
        'storage_protocol': None,
        'vendor_name': 'FUJITSU',
        'QoS_support': True,
        'volume_backend_name': None,
    }

    def __init__(self, prtcl, configuration=None, active_backend_id=None):
        self.protocol = prtcl
        self.configuration = configuration
        (self.remote_host, self.remote_backend_name,
         self.remote_eternus_config, self.local_boxid, self.tfog_name,
         self.remote_pool) = (self._prepare_replica())
        self.active_backend_id = active_backend_id
        if self.active_backend_id:
            self.filename = self.remote_eternus_config
            self.remote_filename = (
                self.configuration.cinder_eternus_config_file)
            self.failover = True
        else:
            self.filename = self.configuration.cinder_eternus_config_file
            self.remote_filename = self.remote_eternus_config
            self.failover = False

        self.storage_name = self._get_drvcfg('EternusIP')
        self.set_sec = self.configuration.fujitsu_set_func_timeout
        self.query_sec = self.configuration.fujitsu_query_func_timeout
        self.req_sec = self.configuration.fujitsu_eternus_req_timeout
        self.retry_times = self.configuration.fujitsu_timeout_retry_times
        self.create_retry_wait_time = \
            self.configuration.fujitsu_create_retry_wait_time

        self.semaphore_num = self.configuration.fujitsu_concurrent_create_num
        if self.semaphore_num > 0 and self.semaphore_num <= 500:
            self.semaphore = BoundedSemaphore(self.semaphore_num)
        else:
            self.semaphore = None

        self.passwordless = self.configuration.fujitsu_passwordless
        self.private_key_path = self.configuration.fujitsu_private_key_path
        self.copy_wait_time = self.configuration.fujitsu_copy_wait_time

        self.conn = None
        self.remote_conn = None
        self.fjdxcli = {}
        self.invalid_migration_list = []
        self.use_ipv6 = False
        self.model_name = self._get_eternus_model()
        self._check_user()
        self.system_name = self._get_eternus_systemname()
        local_model_info = self._get_model_supplement_info()
        self.local_ce_support = False
        if (local_model_info.get('machine_info') == CONSTANTS.DOLPHIN and
                local_model_info.get('model_info') in CONSTANTS.BELUGA):
            self.local_ce_support = True
        if (local_model_info.get('machine_info') == CONSTANTS.PENGUIN and
                local_model_info.get('model_info') in CONSTANTS.ADELIE):
            self.local_ce_support = True

        self.remote_ce_support = False
        if self.remote_eternus_config:
            remote_model_info = self._get_model_supplement_info(remote=True)
            if (remote_model_info.get('machine_info') == CONSTANTS.DOLPHIN and
                    remote_model_info.get('model_info') in CONSTANTS.BELUGA):
                self.remote_ce_support = True
            if (remote_model_info.get('machine_info') == CONSTANTS.PENGUIN and
                    remote_model_info.get('model_info') in CONSTANTS.ADELIE):
                self.remote_ce_support = True

        self.port_info_dict = self._get_port_info()
        self.need_port_num = self.configuration.fujitsu_target_port_use_number
        map_file = self.configuration.fujitsu_port_host_mapping_file
        if not map_file:
            LOG.info('Starting Fujitsu volume driver '
                     'without mapping file.')
            self.host_ag_dict = None
        else:
            LOG.info('Starting Fujitsu volume driver '
                     'with mapping file: %s.', map_file)
            tfo_ports = self._show_all_tfo_port()
            self.host_ag_dict = self._get_host_ag_from_cfg_file(
                map_file, tfo_ports)

        if self.protocol == 'iSCSI':
            if self.need_port_num < CONSTANTS.MIN_NEED_PORT_NUM:
                msg = (_('FJDXCommonExt __init__, '
                         'fujitsu target port use number value: %(need_num)d '
                         'is not available, the available value should be '
                         'greater than or equal to %(min_num)d.')
                       % {'need_num': self.need_port_num,
                          'min_num': CONSTANTS.MIN_NEED_PORT_NUM})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

    @dx_utils.FJDXTimeout()
    def create_volume(self, volume, timeout=None):
        """Create volume on ETERNUS."""
        LOG.debug('create_volume, '
                  'volume id: %(vid)s, volume size: %(vsize)s, '
                  'timeout: %(timeout)s.',
                  {'vid': volume['id'], 'vsize': volume['size'],
                   'timeout': timeout})

        if self.failover:
            msg = (_('create_volume, '
                     'volume id: %(vid)s, '
                     'can not create volume on failed over backend!')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self.conn = self._get_eternus_connection()

        d_metadata = dx_utils.get_metadata(volume)

        element_path, metadata = self._create_volume(volume,
                                                     timeout=self.req_sec)

        d_metadata.update(metadata)

        model_update = {
            'provider_location': six.text_type(element_path),
            'metadata': d_metadata
        }

        # Set qos to created volume.
        self._set_qos(volume, use_id=True, timeout=self.req_sec)

        repl_value = dx_utils.get_extra_specs(volume,
                                              CONSTANTS.FJ_VOL_REPL_KEY)

        if repl_value == CONSTANTS.FJ_VOL_REPL_VALUE:
            try:
                repl_data = None
                repl_status = 'error'
                # Do replication.
                src_host = volume_utils.extract_host(volume['host'], 'host')
                if self.remote_host != src_host:
                    LOG.warning('Do replication, '
                                'remote host: %(remote)s '
                                'and local host: %(local)s '
                                'must be the same one.',
                                {'remote': self.remote_host,
                                 'local': src_host})
                    model_update['replication_status'] = repl_status
                    model_update['replication_driver_data'] = (
                        six.text_type(repl_data))
                    model_update['metadata']['Replication'] = repl_status
                    return model_update

                # Create remote volume.
                replication_driver_data, repl_data, repl_status, error_msg = (
                    self._create_replica_volume(volume, d_metadata))

                if repl_status == 'error':
                    # Create replica failed.
                    LOG.warning(error_msg)
                    raise exception.VolumeBackendAPIException(data=error_msg)

                model_update['replication_status'] = repl_status
                model_update['replication_driver_data'] = (
                    six.text_type(replication_driver_data))
                if repl_data:
                    # Add remote volume information to metadata.
                    model_update['metadata'].update(repl_data)
                # Add the result of replication to metadata.
                model_update['metadata']['Replication'] = repl_status

            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error('create_volume, '
                              'create remote volume, '
                              'undefined error was occurred.')

            # Set qos to remote volume.
            self._set_qos(volume, remote=True, use_id=True,
                          timeout=self.req_sec)

        return model_update

    @dx_utils.FJDXLockutils('consume_pool', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _create_volume(self, volume, remote=False, timeout=None):
        LOG.debug('_create_volume, '
                  'volume id: %(vid)s, volume size: %(vsize)s, '
                  'remote: %(remote)s, timeout: %(timeout)s.',
                  {'vid': volume['id'], 'vsize': volume['size'],
                   'remote': remote, 'timeout': timeout})

        if not remote:
            filename = self.filename
        else:
            filename = self.remote_filename

        conn = self._get_eternus_connection(filename)

        volumesize = int(volume['size']) * units.Gi
        volumename = self._get_volume_name(volume, use_id=True)

        LOG.debug('_create_volume, volumename: %(volumename)s, '
                  'volumesize: %(volumesize)u.',
                  {'volumename': volumename,
                   'volumesize': volumesize})

        configservice = self._find_eternus_service(CONSTANTS.STOR_CONF,
                                                   conn=conn)
        if not configservice:
            msg = (_('_create_volume, volume: %(volume)s, '
                     'volumename: %(volumename)s, '
                     'Storage Configuration Service not found.')
                   % {'volume': volume,
                      'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Get all pools information on ETERNUS.
        pools_instance_list = self._find_all_pools_instances(conn)

        if not remote:
            # First get the pool in the host.
            if 'host' in volume:
                poolname = volume_utils.extract_host(volume['host'], 'pool')

                for pool, ptype in pools_instance_list:
                    if poolname == pool['ElementName']:
                        pool_instance = pool
                        if ptype == 'RAID':
                            pooltype = CONSTANTS.RAIDGROUP
                        else:
                            pooltype = CONSTANTS.TPPOOL
                        break
                else:
                    msg = (_('_create_volume, volume: %(volume)s, '
                             'volumename: %(volumename)s, '
                             'poolname: %(poolname)s, '
                             'Cannot find this pool on ETERNUS.')
                           % {'volume': volume,
                              'volumename': volumename,
                              'poolname': poolname})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                eternus_pool = poolname
                LOG.debug('_create_volume, '
                          'CreateOrModifyElementFromStoragePool, '
                          'ConfigService: %(service)s, '
                          'ElementName: %(volumename)s, '
                          'InPool: %(eternus_pool)s, '
                          'ElementType: %(pooltype)u, '
                          'Size: %(volumesize)u.',
                          {'service': configservice,
                           'volumename': volumename,
                           'eternus_pool': eternus_pool,
                           'pooltype': pooltype,
                           'volumesize': volumesize})

                # Invoke method for create volume.
                rc, errordesc, job = self._exec_eternus_service(
                    'CreateOrModifyElementFromStoragePool',
                    configservice,
                    conn=conn,
                    timeout=self.req_sec,
                    ElementName=volumename,
                    InPool=pool_instance.path,
                    ElementType=self._pywbem_uint(pooltype, '16'),
                    Size=self._pywbem_uint(volumesize, '64'))

            else:
                msg = (_('create_volume, volume id: %(vid)s, '
                         'volume size: %(vsize)s, remote: %(remote)s, '
                         'Cannot find volume host.')
                       % {'vid': volume['id'], 'vsize': volume['size'],
                          'remote': remote})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        else:
            # Get the user specified pool name.
            pool_name_list = self._get_drvcfg('EternusPool', filename=filename,
                                              multiple=True)
            poollen = len(pool_name_list)
            for i in range(poollen):
                # Traverse the user specified pool one by one.
                pool_instances, notfound_poolnames = self._find_pools(
                    [pool_name_list[i]], conn, remote=remote,
                    poolinstances_list=pools_instance_list)

                if pool_instances['pools']:
                    pool_instance = pool_instances['pools'][0]
                    useable = pool_instance['useable_capacity_gb']
                    eternus_pool = pool_instance['pool_name']
                    pool = pool_instance['path']
                    if 'RSP' in pool['InstanceID']:
                        pooltype = CONSTANTS.RAIDGROUP
                    else:
                        pooltype = CONSTANTS.TPPOOL
                    if volume['size'] > useable:
                        continue
                    if pooltype == CONSTANTS.RAIDGROUP:
                        # If it is a raid pool, we need to determine
                        # the number of volumes and fragmentation capacity.
                        # The number of raid pool volumes cannot exceed 128.
                        fragment = pool_instance['fragment_capacity_mb']
                        volcnt = pool_instance['total_volumes']
                        if volcnt >= 128 or volume['size'] > fragment / 1024:
                            LOG.debug('_create_volume, The pool: %(poolname)s '
                                      'can not create volume. '
                                      'Volume Count: %(volcnt)s, '
                                      'Maximum fragment capacity: %(frag)s.',
                                      {'poolname': eternus_pool,
                                       'volcnt': volcnt, 'frag': fragment})
                            continue

                    LOG.debug('_create_volume, '
                              'CreateOrModifyElementFromStoragePool, '
                              'ConfigService: %(service)s, '
                              'ElementName: %(volumename)s, '
                              'InPool: %(eternus_pool)s, '
                              'ElementType: %(pooltype)u, '
                              'Size: %(volumesize)u.',
                              {'service': configservice,
                               'volumename': volumename,
                               'eternus_pool': eternus_pool,
                               'pooltype': pooltype,
                               'volumesize': volumesize})

                    # Invoke method for create volume.
                    rc, errordesc, job = self._exec_eternus_service(
                        'CreateOrModifyElementFromStoragePool',
                        configservice,
                        conn=conn,
                        timeout=self.req_sec,
                        ElementName=volumename,
                        InPool=pool,
                        ElementType=self._pywbem_uint(pooltype, '16'),
                        Size=self._pywbem_uint(volumesize, '64'))

                    if rc == 32769:
                        LOG.warning('_create_volume, raid pool: %s. Maximum '
                                    'number of Logical Volume in a RAID group '
                                    'has been reached. Try other pool.',
                                    eternus_pool)
                        continue
                    else:
                        break
                else:
                    if notfound_poolnames:
                        LOG.warning('_create_volume, '
                                    'pool names: %(notfound_poolnames)s '
                                    'are not found.',
                                    {'notfound_poolnames': notfound_poolnames})
            else:
                # It means that all raid pools do not meet
                # the volume limit (<128), and the creation request of
                # this volume will be rejected.
                # If there is a thin pool available,
                # it will not enter this branch.
                msg = (_('create_volume, volume id: %(vid)s, '
                         'volume size: %(vsize)s, remote: %(remote)s, '
                         'All pools cannot create this volume.')
                       % {'vid': volume['id'], 'vsize': volume['size'],
                          'remote': remote})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        if rc == CONSTANTS.VOLUMENAME_IN_USE:  # Element Name is in use.
            if not remote:
                LOG.warning('_create_volume, '
                            'volumename: %(volumename)s, '
                            'Element Name is in use.',
                            {'volumename': volumename})
                vol_instance = self._find_lun(volume)
            else:
                LOG.warning('_create_volume, '
                            'volumename: %(volumename)s, '
                            'Remote Element Name is in use.',
                            {'volumename': volumename})
                vol_name = {
                    'source-name': volumename
                }
                vol_instance = self._find_lun_with_listup(
                    conn=conn, **vol_name)
            element = vol_instance
        elif rc == CONSTANTS.MAXIMUN_NUMBER_LUN or \
                rc == CONSTANTS.SIZE_NOT_SUPPORTED:
            # Volume size is bigger than fragment_capacity_mb
            # Or volume count reaches 128
            # Waiting for re-scheduling.
            msg = (_('_create_volume, '
                     'volumename: %(volumename)s, '
                     'poolname: %(eternus_pool)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'volumename': volumename,
                      'eternus_pool': eternus_pool,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            if not remote:
                time.sleep(self.create_retry_wait_time)
            raise exception.VolumeBackendAPIException(data=msg)
        elif rc != 0:
            msg = (_('_create_volume, '
                     'volumename: %(volumename)s, '
                     'poolname: %(eternus_pool)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'volumename': volumename,
                      'eternus_pool': eternus_pool,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:
            element = job['TheElement']

        # Get eternus model name.
        try:
            systemnamelist = self._enum_eternus_instances(
                'FUJITSU_StorageProduct',
                conn=conn,
                retry=self.retry_times)
        except Exception:
            msg = (_('_create_volume, '
                     'volume: %(volume)s, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.')
                   % {'volume': volume})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_create_volume, '
                  'volumename: %(volumename)s, '
                  'Backend: %(backend)s, '
                  'Pool Name: %(eternus_pool)s, '
                  'Pool Type: %(pooltype)s.',
                  {'volumename': volumename,
                   'backend': systemnamelist[0]['IdentifyingNumber'],
                   'eternus_pool': eternus_pool,
                   'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype]})

        # Create return value.
        element_path = {
            'classname': element.classname,
            'keybindings': {
                'SystemName': element['SystemName'],
                'DeviceID': element['DeviceID'],
            },
            'vol_name': volumename,
        }

        volume_no = self._get_volume_no(element)
        metadata = {
            'FJ_Backend': systemnamelist[0]['IdentifyingNumber'],
            'FJ_Volume_Name': volumename,
            'FJ_Volume_No': volume_no,
            'FJ_Pool_Name': eternus_pool,
            'FJ_Pool_Type': CONSTANTS.POOL_TYPE_dic[pooltype],
        }

        if not remote:
            return element_path, metadata
        else:
            repl_data = {
                'Remote_Backend': systemnamelist[0]['IdentifyingNumber'],
                'Remote_Volume_No': volume_no,
                'Remote_Pool_Name': eternus_pool,
                'Remote_Device_ID': element['DeviceID'],
            }
            return element_path, metadata, repl_data, 'enabled'

    def create_pool_info(self, pool_instance, volume_count, pool_type):
        """Create pool information from pool instance."""
        LOG.debug('create_pool_info, pool_instance: %(pool)s, '
                  'volume_count: %(volcount)s, pool_type: %(ptype)s.',
                  {'pool': pool_instance,
                   'volcount': volume_count, 'ptype': pool_type})

        if pool_type not in CONSTANTS.POOL_TYPE_list:
            msg = (_('Invalid pool type was specified : %s.') % pool_type)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        total_mb = pool_instance['TotalManagedSpace'] * 1.0 / units.Mi
        free_mb = pool_instance['RemainingManagedSpace'] * 1.0 / units.Mi
        fragment_mb = free_mb

        if hasattr(pool_instance, 'provisioned_capacity_mb'):
            prov_mb = pool_instance.provisioned_capacity_mb
        else:
            prov_mb = total_mb - free_mb

        if pool_type == 'RAID':
            useable_mb = free_mb
            if hasattr(pool_instance, 'fragment_size'):
                if pool_instance.fragment_size != -1:
                    fragment_mb = pool_instance.fragment_size / (2 * 1024)
                else:
                    fragment_mb = useable_mb
        else:
            max_capacity_mb = total_mb * float(
                self.configuration.max_over_subscription_ratio)
            useable_mb = max_capacity_mb - prov_mb

        pool = {
            'name': pool_instance['ElementName'],
            'path': pool_instance.path,
            'total_capacity_gb': int(total_mb / 1024),
            'free_capacity_gb': int(free_mb / 1024),
            'type': pool_type,
            'volume_count': volume_count,
            'provisioned_capacity_gb': int(prov_mb / 1024),
            'useable_capacity_gb': int(useable_mb / 1024),
            'useable_capacity_mb': useable_mb,
            'fragment_capacity_mb': fragment_mb,
        }

        LOG.debug('create_pool_info, pool: %s.', pool)
        return pool

    @dx_utils.FJDXTimeout()
    def create_volume_from_snapshot(self, volume, snapshot,
                                    timeout=None):
        """Create a volume from a snapshot."""
        LOG.debug('create_volume_from_snapshot, '
                  'volume id: %(vid)s, volume size: %(vsize)s, '
                  'snapshot id: %(sid)s, timeout: %(timeout)s.',
                  {'vid': volume['id'], 'vsize': volume['size'],
                   'sid': snapshot['id'], 'timeout': timeout})

        self.conn = self._get_eternus_connection()
        source_volume_instance = self._find_lun(snapshot)

        # Check the existence of source volume.
        if not source_volume_instance:
            msg = _('create_volume_from_snapshot, '
                    'Source Volume does not exist in ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        checkstart_time = time.time()
        while self._check_if_cascade_copy_volume(snapshot):
            checkend_time = time.time()
            remaining_time = (checkstart_time +
                              self.copy_wait_time) - checkend_time
            if remaining_time <= 0:
                msg = (_('create_volume_from_snapshot, '
                         'The source snapshot is cascade copy '
                         'destination volume. '
                         'check start time: %(st)s, '
                         'check end time: %(et)s, '
                         'remaining time: %(rt)s, '
                         'wait time:%(wt)s.')
                       % {'st': checkstart_time,
                          'et': checkend_time,
                          'rt': remaining_time,
                          'wt': self.copy_wait_time})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                wait_time = remaining_time if (
                            remaining_time < CONSTANTS.CHECK_COPY_WAIT_TIME
                ) else CONSTANTS.CHECK_COPY_WAIT_TIME
                LOG.debug('create_cloned_volume, '
                          'check start time: %(st)s, '
                          'check end time: %(et)s, '
                          'remaining time: %(rt)s, '
                          'wait time: %(wt)s.',
                          {'st': checkstart_time,
                           'et': checkend_time,
                           'rt': remaining_time,
                           'wt': self.copy_wait_time})
                time.sleep(wait_time)

        # Create volume for destination of cloned volume.
        model_update = self.create_volume(volume, timeout=self.set_sec)
        element_path = eval(model_update.get('provider_location'))
        target_volume_instancename = self._create_eternus_instance_name(
            element_path['classname'], element_path['keybindings'])

        try:
            target_volume_instance = self._get_eternus_instance(
                target_volume_instancename, retry=self.retry_times)
        except Exception:
            msg = (_('create_volume_from_snapshot, '
                     'target volume instancename: %(volume_instancename)s, '
                     'Get Instance Failed.')
                   % {'volume_instancename': target_volume_instancename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self._create_local_cloned_volume(target_volume_instance,
                                         source_volume_instance,
                                         timeout=self.req_sec)

        return model_update

    @dx_utils.FJDXTimeout()
    def create_cloned_volume(self, volume, src_vref,
                             clone_only=False,
                             use_service_name=False,
                             timeout=None):
        """Create clone of the specified volume."""
        LOG.debug('create_cloned_volume, '
                  'tgt: (%(tid)s, %(tsize)s), src: (%(sid)s, %(ssize)s), '
                  'timeout: %(timeout)s.',
                  {'tid': volume['id'], 'tsize': volume['size'],
                   'sid': src_vref['id'], 'ssize': src_vref['size'],
                   'timeout': timeout})

        if self.failover:
            msg = (_('create_cloned_volume, '
                     'volume id: %(vid)s, '
                     'can not create cloned volume on failed over backend!')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        model_update = None

        self.conn = self._get_eternus_connection()
        source_volume_instance = self._find_lun(src_vref, use_service_name)

        if not source_volume_instance:
            msg = _('create_cloned_volume, '
                    'Source Volume does not exist in ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # When there is cascade copy, wait or report error.
        checkstart_time = time.time()
        while self._check_if_cascade_copy_volume(src_vref):
            checkend_time = time.time()
            remaining_time = (checkstart_time +
                              self.copy_wait_time) - checkend_time
            if remaining_time <= 0:
                msg = (_('create_cloned_volume, '
                         'The source volume is cascade copy '
                         'destination volume. '
                         'check start time: %(st)s, '
                         'check end time: %(et)s, '
                         'remaining time: %(rt)s, '
                         'wait time: %(wt)s.')
                       % {'st': checkstart_time,
                          'et': checkend_time,
                          'rt': remaining_time,
                          'wt': self.copy_wait_time})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                wait_time = remaining_time if (
                            remaining_time < CONSTANTS.CHECK_COPY_WAIT_TIME
                ) else CONSTANTS.CHECK_COPY_WAIT_TIME
                LOG.debug('create_cloned_volume, '
                          'check start time: %(st)s, '
                          'check end time: %(et)s, '
                          'remaining time: %(rt)s, '
                          'wait time: %(wt)s.',
                          {'st': checkstart_time,
                           'et': checkend_time,
                           'rt': remaining_time,
                           'wt': self.copy_wait_time})
                time.sleep(wait_time)

        if not clone_only:
            model_update = self.create_volume(volume, timeout=self.set_sec)
            element_path = eval(model_update.get('provider_location'))
            metadata = model_update.get('metadata')
            target_volume_instancename = self._create_eternus_instance_name(
                element_path['classname'], element_path['keybindings'])
            target_volume_instance = self._get_eternus_instance(
                target_volume_instancename, retry=self.retry_times)
        else:
            target_volume_instance = self._find_lun(volume)
            try:
                element_path = eval(volume['provider_location'])
                metadata = volume['volume_metadata']
            except Exception:
                element_path = None
                metadata = None

        remote_copy_type = (
            dx_utils.get_metadata(volume).get(CONSTANTS.FJ_REMOTE_DEST_META,
                                              None))

        if not remote_copy_type:
            self._create_local_cloned_volume(target_volume_instance,
                                             source_volume_instance,
                                             timeout=self.req_sec)
        else:
            source_volume_metadata = dx_utils.get_metadata(src_vref)
            self._create_remote_cloned_volume(volume,
                                              metadata,
                                              src_vref,
                                              source_volume_metadata,
                                              remote_copy_type)

        if not model_update:
            model_update = {
                'provider_location': six.text_type(element_path),
                'metadata': metadata,
            }

        return model_update

    def _create_remote_cloned_volume(self, volume, volume_metadata, src_vref,
                                     src_vref_metadata, remote_copy_type):
        """Create remote clone of the specified volume."""
        LOG.debug('_create_remote_cloned_volume, '
                  'tgt: %(tid)s, src: %(sid)s.',
                  {'tid': volume['id'], 'sid': src_vref['id']})

        ctxt = context.get_admin_context()
        copy_type = six.text_type(remote_copy_type).lower()

        # Check if source volume has already get a mirror target.
        source_volume_copy_dict = eval(
            src_vref_metadata.get(CONSTANTS.FJ_REMOTE_SRC_META, "{}"))

        # When remote_copy_type is mirror, check mirror copy already exist.
        if copy_type == CONSTANTS.FJ_REC_MIRROR:
            for value in source_volume_copy_dict.values():
                if six.text_type(value).lower() == CONSTANTS.FJ_REC_MIRROR:
                    s_volumename = self._get_volume_name(src_vref)
                    msg = (_('_create_remote_cloned_volume, '
                             'source volume: %(src)s already has a mirror!')
                           % {'src': s_volumename})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        # Execute CCM script.
        self._exec_ccm_script("start",
                              source=src_vref_metadata,
                              source_volume=src_vref,
                              target=volume_metadata,
                              copy_type=copy_type)

        # Update metadata of source volume.
        lockname = 'db-update-' + src_vref['id']

        @dx_utils.FJDXLockutils(lockname, 'cinder-', True)
        @dx_utils.FJDXTimeout()
        def __update_metadata(self, timeout=None):
            """Update volume metadata."""
            LOG.debug('__update_metadata, copy_type: %(copy_type)s, '
                      'timeout: %(timeout)s.',
                      {'copy_type': copy_type, 'timeout': timeout})
            source_volume_copy_dict = {volume['id']: copy_type}
            metadata = {
                CONSTANTS.FJ_REMOTE_SRC_META: six.text_type(
                    source_volume_copy_dict)
            }
            db.volume_metadata_update(ctxt.elevated(), src_vref['id'],
                                      metadata, False)

        if copy_type == CONSTANTS.FJ_REC_MIRROR:
            # If remote copy type is mirror, update metadata.
            __update_metadata(self, timeout=self.req_sec)
        elif copy_type == CONSTANTS.FJ_REC_CLONE:
            # If remote copy type and copy_list is {}, update metadata.
            if not source_volume_copy_dict:
                __update_metadata(self, timeout=self.req_sec)

    def _create_local_cloned_volume(self, tgt_vol_instance, src_vol_instance,
                                    timeout=None):
        """Create local clone of the specified volume."""
        s_volumename = src_vol_instance['ElementName']
        t_volumename = tgt_vol_instance['ElementName']

        LOG.debug('_create_local_cloned_volume, '
                  'tgt volume name: %(t_volumename)s, '
                  'src volume name: %(s_volumename)s, '
                  'timeout: %(timeout)s.',
                  {'t_volumename': t_volumename,
                   's_volumename': s_volumename,
                   'timeout': timeout})

        # Get replication service for CreateElementReplica.
        repservice = self._find_eternus_service(CONSTANTS.REPL)

        if not repservice:
            msg = _('_create_local_cloned_volume, '
                    'Replication Service not found.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Invoke method for create cloned volume from volume.
        rc, errordesc, job = self._exec_eternus_service(
            'CreateElementReplica',
            repservice,
            timeout=self.req_sec,
            SyncType=self._pywbem_uint(8, '16'),
            SourceElement=src_vol_instance.path,
            TargetElement=tgt_vol_instance.path)

        if rc != 0:
            msg = (_('_create_local_cloned_volume, '
                     'volumename: %(volumename)s, '
                     'sourcevolumename: %(sourcevolumename)s, '
                     'source volume instance: %(source_volume)s, '
                     'target volume instance: %(target_volume)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'volumename': t_volumename,
                      'sourcevolumename': s_volumename,
                      'source_volume': src_vol_instance.path,
                      'target_volume': tgt_vol_instance.path,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)

            if rc == 5 and self.model_name == CONSTANTS.DX_S2:
                msg = (_('_create_local_cloned_volume, '
                         'volume name: %(t_name)s, '
                         'source volume name: %(s_name)s, '
                         'not supported on DX S2.')
                       % {'t_name': t_volumename,
                          's_name': s_volumename})
                LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_create_local_cloned_volume, out: %(rc)s, %(job)s.',
                  {'rc': rc, 'job': job})

    @dx_utils.FJDXTimeout()
    def delete_volume(self, volume, timeout=None):
        """Delete volume on ETERNUS."""
        LOG.debug('delete_volume, '
                  'volume id: %(vid)s, timeout: %(timeout)s.',
                  {'vid': volume['id'], 'timeout': timeout})

        if self.failover:
            msg = (_('delete_volume, '
                     'volume id: %(vid)s, '
                     'can not delete a failed over volume!')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        deleted = False

        self.conn = self._get_eternus_connection()
        vol_instance = self._find_lun(volume)
        if not vol_instance:
            return True

        # CCM processing
        metadata = dx_utils.get_metadata(volume)
        copy_type = metadata.get(CONSTANTS.FJ_REMOTE_DEST_META, None)

        if copy_type and ("FJ_Backend" in metadata):
            self._exec_ccm_script("stop", target=metadata)

            # If copy_type is mirror, update FJ_Remote_Copy_Source info.
            tmp_flag1 = CONSTANTS.FJ_SRC_VOL_ID in metadata
            tmp_flag2 = six.text_type(copy_type).lower() == (
                CONSTANTS.FJ_REC_MIRROR)
            if tmp_flag1 and tmp_flag2:
                source_volume_id = metadata.get(CONSTANTS.FJ_SRC_VOL_ID)
                ctxt = context.get_admin_context()
                src_vref = db.volume_get(ctxt, source_volume_id)
                source_volume_metadata = dx_utils.get_metadata(src_vref)
                source_volume_copy_list = (
                    eval(source_volume_metadata.get(
                        CONSTANTS.FJ_REMOTE_SRC_META, "{}")))

                # If source_volume_copy_list has this volume, delete it.
                volume_id = volume['id']
                if volume_id in source_volume_copy_list:
                    del source_volume_copy_list[volume_id]

                update_metadata = {
                    CONSTANTS.FJ_REMOTE_SRC_META: six.text_type(
                        source_volume_copy_list)}

                db.volume_metadata_update(ctxt.elevated(),
                                          source_volume_id,
                                          update_metadata,
                                          False)

        if metadata.get(CONSTANTS.FJ_REMOTE_SRC_META, None):
            self._exec_ccm_script(
                "stop", source=metadata, source_volume=volume)

            # Update volume metadata.
            copy_list = eval(metadata.get(CONSTANTS.FJ_REMOTE_SRC_META))
            ctxt = context.get_admin_context()
            for target_volume_id, copy_type in copy_list.items():
                try:
                    if six.text_type(copy_type).lower() == (
                            CONSTANTS.FJ_REC_MIRROR):
                        db.volume_metadata_update(
                            ctxt.elevated(), target_volume_id,
                            {CONSTANTS.FJ_REMOTE_DEST_META: None}, False)
                except Exception:
                    pass

        with_format = dx_utils.get_format_property(volume,
                                                   CONSTANTS.FJ_VOL_FORMAT_KEY)

        repl_status = volume.get('replication_status', None)
        if repl_status and repl_status != 'disabled':
            remote_vol = None
            try:
                remote_vol = self._find_replica_lun(volume)
            except Exception:
                LOG.warning('delete_volume, '
                            'remote volume does not found.')

            if remote_vol:
                repl_deleted = False
                src_host = volume_utils.extract_host(volume['host'], 'host')
                if self.remote_host != src_host:
                    msg = (_('Do replication, '
                             'remote host:%(remote)s '
                             'and local host:%(local)s '
                             'must be the same one.')
                           % {'remote': self.remote_host,
                              'local': src_host})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                self.remote_conn = self._get_eternus_connection(
                    self.remote_filename)

                repl_vol_exist = self._delete_volume_setting(
                    volume, remote=True, timeout=self.req_sec)

                if not repl_vol_exist:
                    LOG.debug('delete_repl_volume, '
                              'volume not found in 1st check.')
                else:
                    try:
                        repl_deleted = self._delete_volume_base(
                            volume, with_format, remote=True)
                    except Exception:
                        msg = (_('delete_volume, '
                                 'delete replica volume failed, '
                                 'deleted status is: %s.')
                               % repl_deleted)
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

        vol_exist = self._delete_volume_setting(volume, timeout=self.req_sec)

        if not vol_exist:
            LOG.debug('delete_volume, volume not found in 1st check.')
            return True

        try:
            deleted = self._delete_volume_base(volume, with_format)
        except Exception:
            msg = (_('delete_volume, '
                     'delete volume failed, '
                     'deleted status is: %s.')
                   % deleted)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('delete_volume, deleted status is: %s.', deleted)

        return deleted

    @dx_utils.FJDXLockutils('vol', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _delete_volume_setting(self, volume, remote=False, timeout=None):
        """Delete volume setting (HostAffinity, CopySession) on ETERNUS."""
        LOG.debug('_delete_volume_setting, '
                  'volume id: %(vid)s, '
                  'remote: %(remote)s, '
                  'timeout: %(timeout)s.',
                  {'vid': volume['id'], 'remote': remote, 'timeout': timeout})

        # Check the existence of volume.
        volumename = self._get_volume_name(volume)
        if not remote:
            vol_instance = self._find_lun(volume)
            conn = self.conn
        else:
            vol_instance = self._find_replica_lun(volume)
            conn = self.remote_conn

        if not vol_instance:
            LOG.info('_delete_volume_setting, volumename:%(volumename)s, '
                     'volume not found on ETERNUS.',
                     {'volumename': volumename})
            return False

        # Delete host-affinity setting remained by unexpected error.
        self._unmap_lun(volume, None,
                        force=True, remote=remote, timeout=self.req_sec)

        # Check copy session relating to target volume.
        cpsessionlist = self._find_copysession(vol_instance, conn=conn)
        delete_copysession_list = []
        wait_copysession_list = []

        for cpsession in cpsessionlist:
            LOG.debug('_delete_volume_setting, '
                      'volumename: %(volumename)s, '
                      'cpsession: %(cpsession)s.',
                      {'volumename': volumename,
                       'cpsession': cpsession})

            if cpsession['SyncedElement'] == vol_instance.path:
                # Copy target : other_volume --(copy)--> vol_instance.
                delete_copysession_list.append(cpsession)
            elif cpsession['SystemElement'] == vol_instance.path:
                # Copy source : vol_instance --(copy)--> other volume.
                wait_copysession_list.append(cpsession)

        LOG.debug('_delete_volume_setting, '
                  'wait_cpsession: %(wait_cpsession)s, '
                  'delete_cpsession: %(delete_cpsession)s.',
                  {'wait_cpsession': wait_copysession_list,
                   'delete_cpsession': delete_copysession_list})

        for cpsession in wait_copysession_list:
            complete = self._wait_for_copy_complete(cpsession, conn=conn)
            if not complete:
                # Cpsession is EC/REC and it is needed to be deleted.
                delete_copysession_list.append(cpsession)

        for cpsession in delete_copysession_list:
            self._delete_copysession(cpsession, conn=conn)

        LOG.debug('_delete_volume_setting, '
                  'wait_cpsession: %(wait_cpsession)s, '
                  'delete_cpsession: %(delete_cpsession)s, complete.',
                  {'wait_cpsession': wait_copysession_list,
                   'delete_cpsession': delete_copysession_list})
        return True

    @dx_utils.FJDXLockutils('vol', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _delete_volume(self, volume, conn=None, remote=False, timeout=None):
        """Delete volume on ETERNUS."""
        LOG.debug('_delete_volume, volume id: %(vid)s, '
                  'timeout: %(timeout)s.',
                  {'vid': volume['id'],
                   'timeout': timeout})

        if remote:
            vol_instance = self._find_replica_lun(volume)
        else:
            vol_instance = self._find_lun(volume)

        if not vol_instance:
            LOG.debug('_delete_volume, volume not found in 2nd check, '
                      'but no problem.')
            return

        volumename = vol_instance['ElementName']

        configservice = self._find_eternus_service(CONSTANTS.STOR_CONF,
                                                   conn=conn)
        if not configservice:
            msg = (_('_delete_volume, volumename: %(volumename)s, '
                     'Storage Configuration Service not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_delete_volume, volumename: %(volumename)s, '
                  'vol_instance: %(vol_instance)s, '
                  'Method: ReturnToStoragePool.',
                  {'volumename': volumename,
                   'vol_instance': vol_instance.path})

        # Invoke method for delete volume.
        rc, errordesc, _x = self._exec_eternus_service(
            'ReturnToStoragePool',
            configservice,
            conn=conn,
            timeout=self.req_sec,
            TheElement=vol_instance.path)

        if rc != 0:
            msg = (_('_delete_volume, volumename: %(volumename)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_delete_volume, volumename: %(volumename)s, '
                  'Return code: %(rc)lu, '
                  'Error: %(errordesc)s.',
                  {'volumename': volumename,
                   'rc': rc,
                   'errordesc': errordesc})

    @dx_utils.FJDXTimeout()
    def create_snapshot(self, snapshot, timeout=None):
        """Create snapshot using SnapOPC."""
        LOG.debug('create_snapshot, '
                  'snapshot id: %(sid)s, volume id: %(vid)s, '
                  'timeout: %(timeout)s.',
                  {'sid': snapshot['id'], 'vid': snapshot['volume_id'],
                   'timeout': timeout})

        if self.failover:
            msg = (_('create_snapshot, '
                     'snapshot id: %(sid)s, '
                     'can not create snapshot on failed over backend!')
                   % {'sid': snapshot['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self.conn = self._get_eternus_connection()
        volume = snapshot['volume']
        s_volumename = self._get_volume_name(volume)
        vol_instance = self._find_lun(volume)

        # Check the existence of volume.
        if not vol_instance:
            # Volume not found on ETERNUS.
            msg = (_('create_snapshot, '
                     'volumename: %(s_volumename)s, '
                     'source volume not found on ETERNUS.')
                   % {'s_volumename': s_volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        checkstart_time = time.time()
        while self._check_if_cascade_copy_volume(volume):
            checkend_time = time.time()
            remaining_time = (checkstart_time +
                              self.copy_wait_time) - checkend_time
            if remaining_time <= 0:
                msg = (_('create_snapshot, '
                         'The source volume is cascade copy '
                         'destination volume. '
                         'check start time: %(st)s, '
                         'check end time: %(et)s, '
                         'remaining time: %(rt)s, '
                         'wait time: %(wt)s.')
                       % {'st': checkstart_time,
                          'et': checkend_time,
                          'rt': remaining_time,
                          'wt': self.copy_wait_time})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                wait_time = remaining_time if (
                            remaining_time < CONSTANTS.CHECK_COPY_WAIT_TIME
                ) else CONSTANTS.CHECK_COPY_WAIT_TIME
                LOG.debug('create_cloned_volume, '
                          'check start time: %(st)s, '
                          'check end time: %(et)s, '
                          'remaining time: %(rt)s, '
                          'wait time: %(wt)s.',
                          {'st': checkstart_time,
                           'et': checkend_time,
                           'rt': remaining_time,
                           'wt': self.copy_wait_time})
                time.sleep(wait_time)

        model_update = self._create_snapshot(snapshot, timeout)

        return model_update

    @dx_utils.FJDXLockutils('vol', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _create_snapshot(self, snapshot, timeout=None):

        LOG.debug('_create_snapshot, '
                  'snapshot id: %(sid)s, volume id: %(vid)s, '
                  'timeout: %(timeout)s.',
                  {'sid': snapshot['id'], 'vid': snapshot['volume_id'],
                   'timeout': timeout})

        snapshotname = snapshot['name']
        volume = snapshot['volume']
        volumename = snapshot['volume_name']
        d_volumename = self._get_volume_name(snapshot, use_id=True)
        vol_instance = self._find_lun(volume)
        service_name = (CONSTANTS.REPL
                        if self.model_name != CONSTANTS.DX_S2
                        else CONSTANTS.STOR_CONF)

        volume_size = int(float(snapshot['volume']['size'])) * 1024

        smis_service = self._find_eternus_service(service_name)

        if not smis_service:
            msg = (_('_create_snapshot, '
                     'volumename: %(volumename)s, '
                     '%(servicename)s not found.')
                   % {'volumename': volumename,
                      'servicename': service_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Get all pools information on ETERNUS.
        pools_instance_list = self._find_all_pools_instances(self.conn)

        # Get the user specified pool name.
        pool_name_list = self._get_drvcfg('EternusSnapPool', multiple=True)

        poollen = len(pool_name_list)
        for i in range(poollen):
            # Traverse the user specified pool one by one.
            pool_instances, notfound_poolnames = self._find_pools(
                [pool_name_list[i]], self.conn,
                poolinstances_list=pools_instance_list)

            if pool_instances['pools']:
                useable = pool_instances['pools'][0]['useable_capacity_mb']
                poolname = pool_instances['pools'][0]['pool_name']
                istpp = pool_instances['pools'][0]['thin_provisioning_support']
                if useable < 24 + volume_size * 0.001:
                    continue
                if not istpp:
                    # If it is a raid pool, we need to determine
                    # the number of volumes and fragmentation capacity.
                    # The number of raid pool volumes cannot exceed 128.
                    # The minimum space required for snapshot is 24m.
                    fragment = pool_instances['pools'][0][
                        'fragment_capacity_mb']
                    volcnt = pool_instances['pools'][0]['total_volumes']
                    if volcnt >= 128 or fragment < 24 + volume_size * 0.001:
                        LOG.debug('_create_volume, The pool: %(poolname)s '
                                  'can not create volume. '
                                  'Volume Count: %(volcnt)s, '
                                  'Maximum fragment capacity: %(frag)s.',
                                  {'poolname': poolname,
                                   'volcnt': volcnt, 'frag': fragment})
                        continue

                pool_instance = pool_instances['pools'][0]
                eternus_pool = pool_instance['pool_name']
                pool = pool_instance['path']
                if 'RSP' in pool['InstanceID']:
                    pooltype = CONSTANTS.RAIDGROUP
                else:
                    pooltype = CONSTANTS.TPPOOL

                LOG.debug('_create_snapshot, '
                          'snapshotname: %(snapshotname)s, '
                          'source volume name: %(volumename)s, '
                          'vol_instance.path: %(vol_instance)s, '
                          'dest_volumename: %(d_volumename)s, '
                          'pool: %(pool)s, '
                          'Invoke CreateElementReplica.',
                          {'snapshotname': snapshotname,
                           'volumename': volumename,
                           'vol_instance': vol_instance.path,
                           'd_volumename': d_volumename,
                           'pool': eternus_pool})

                if self.model_name != CONSTANTS.DX_S2:
                    smis_method = 'CreateElementReplica'
                    params = {
                        'ElementName': d_volumename,
                        'TargetPool': pool,
                        'SyncType': self._pywbem_uint(7, '16'),
                        'SourceElement': vol_instance.path
                    }
                else:
                    smis_method = 'CreateReplica'
                    params = {
                        'ElementName': d_volumename,
                        'TargetPool': pool,
                        'CopyType': self._pywbem_uint(4, '16'),
                        'SourceElement': vol_instance.path
                    }
                # Invoke method for create snapshot.
                rc, errordesc, job = self._exec_eternus_service(
                    smis_method, smis_service, timeout=self.req_sec, **params)

                if rc != 0:
                    LOG.warning('_create_snapshot, '
                                'snapshotname: %(snapshotname)s, '
                                'source volume name: %(volumename)s, '
                                'vol_instance.path: %(vol_instance)s, '
                                'dest volume name: %(d_volumename)s, '
                                'pool: %(pool)s, Return code: %(rc)lu, '
                                'Error: %(errordesc)s.',
                                {'snapshotname': snapshotname,
                                 'volumename': volumename,
                                 'vol_instance': vol_instance.path,
                                 'd_volumename': d_volumename,
                                 'pool': eternus_pool,
                                 'rc': rc,
                                 'errordesc': errordesc})
                    continue
                else:
                    break
            else:
                if notfound_poolnames:
                    LOG.warning('_create_snapshot, '
                                'pool names: %(notfound_poolnames)s '
                                'are not found.',
                                {'notfound_poolnames': notfound_poolnames})
        else:
            # It means that all raid pools do not meet
            # the volume limit (<128), and the creation request of
            # this volume will be rejected.
            # If there is a thin pool available, it will not enter this branch.
            msg = (_('_create_snapshot, volume id: %(sid)s, '
                     'All pools cannot create this volume.')
                   % {'sid': snapshot['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        element = job['TargetElement']

        LOG.debug('_create_snapshot, '
                  'volumename: %(volumename)s, '
                  'Return code: %(rc)lu, '
                  'Error: %(errordesc)s.',
                  {'volumename': volumename,
                   'rc': rc,
                   'errordesc': errordesc})

        # Create return value.
        element_path = {
            'classname': element.classname,
            'keybindings': {
                'SystemName': element['SystemName'],
                'DeviceID': element['DeviceID'],
            },
            'vol_name': d_volumename,
        }
        sdv_no = self._get_volume_no(element)

        metadata = {
            'FJ_SDV_Name': d_volumename,
            'FJ_SDV_No': sdv_no,
            'FJ_Pool_Name': eternus_pool,
            'FJ_Pool_Type': pooltype
        }
        d_metadata = dx_utils.get_metadata(snapshot)
        d_metadata.update(metadata)

        model_update = {
            'provider_location': six.text_type(element_path),
            'metadata': d_metadata,
        }
        return model_update

    def delete_snapshot(self, snapshot):
        """Delete snapshot."""
        LOG.debug('delete_snapshot, '
                  'snapshot id: %(sid)s, volume id: %(vid)s.',
                  {'sid': snapshot['id'], 'vid': snapshot['volume_id']})

        if self.failover:
            msg = (_('delete_snapshot, '
                     'snapshot id: %(sid)s, '
                     'can not delete a snapshot on failed over backend!')
                   % {'sid': snapshot['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        vol_exist = self.delete_volume(snapshot, timeout=self.set_sec)
        LOG.debug('delete_snapshot, vol_exist: %s.', vol_exist)
        return vol_exist

    @dx_utils.FJDXTimeout()
    def initialize_connection(self, volume, connector, timeout=None):
        """Allow connection to connector and return connection info."""
        LOG.debug('initialize_connection, '
                  'volume id: %(vid)s, protocol: %(prtcl)s, '
                  'timeout: %(timeout)s.',
                  {'vid': volume['id'], 'prtcl': self.protocol,
                   'timeout': timeout})

        if self.failover:
            msg = (_('initialize_connection, '
                     'volume id: %(vid)s, '
                     'can not attach a failed over volume!')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self.conn = self._get_eternus_connection()

        if not self.port_info_dict:
            msg = (_('initialize_connection, '
                     'no available ports on ETERNUS for %s.')
                   % self.protocol)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # CCM processing.
        metadata = dx_utils.get_metadata(volume)
        dest_meta = metadata.get(CONSTANTS.FJ_REMOTE_DEST_META, None)
        ccm_flag = (six.text_type(dest_meta).lower() ==
                    CONSTANTS.FJ_REC_MIRROR)
        if ccm_flag:
            if volume['status'] == "in-use":
                msg = _('Live migration for backup volume is not allowed.')
                self.invalid_migration_list.append(volume['id'])
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            self._exec_ccm_script("suspend", target=metadata)

        try:
            replica_volume = self._is_replicated(volume)
            if replica_volume:
                mapdata, ports = self._map_replication_volume(
                    volume, connector)
            else:
                mapdata, ports = self._map_ordinary_volume(
                    volume, connector)

            if self.protocol == 'FC':
                device_info = {
                    'driver_volume_type': 'fibre_channel',
                    'data': mapdata
                }
            elif self.protocol == 'iSCSI':
                device_info = {
                    'driver_volume_type': 'iscsi',
                    'data': mapdata
                }

        except Exception:
            with excutils.save_and_reraise_exception():
                # When volume is set to REC Mirror, resume the session.
                if ccm_flag:
                    self._exec_ccm_script("resume", target=metadata)

        if replica_volume:
            ctxt = context.get_admin_context()
            db.volume_update(
                ctxt, volume['id'], {
                    'replication_extended_status': six.text_type(ports)})
        LOG.debug('initialize_connection, device_info: %s.', device_info)
        return device_info

    def _find_tfo_port_pairs(self, own_ports, pair_ports, connector, volume):
        LOG.debug('_find_tfo_port_pairs, '
                  'own_ports: %(own_ports)s, '
                  'pair_ports: %(pair_ports)s, '
                  'connector: %(connector)s, '
                  'volume id: %(id)s.',
                  {'own_ports': own_ports,
                   'pair_ports': pair_ports,
                   'connector': connector,
                   'id': volume['id']})

        ports_dict = {}  # {own_port: pair_port}.
        index = 0
        for own_port in own_ports:
            ports_dict[own_port] = pair_ports[index]
            index += 1

        # Get FC zone info on ETERNUS for FC.
        host_tfo_ports_dict = {}
        # Get hosts info from connector.
        initiatorlist = self._find_initiator_names(connector)
        if self.protocol == 'FC':
            host_info_list = [x.lower() for x in initiatorlist]
            fc_zone_dict = self._get_enternus_fc_zone_info()
            if not fc_zone_dict:
                msg = (_('_find_tfo_port_pairs, '
                         'volume id: %(vol_id)s, connector(%(connector)s), '
                         'FC zone info is empty on ETERNUS. '
                         'Please check the FC connection and '
                         'the setting on FC switch.')
                       % {'vol_id': volume['id'],
                          'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                for host in host_info_list:
                    # Get available ports for host affinity group.
                    tmp = fc_zone_dict.get(host, [])
                    avai_own_ports = [x for x in tmp if x in own_ports]
                    avai_pair_ports = []
                    for avai_own_port in avai_own_ports:
                        if avai_own_port in ports_dict:
                            avai_pair_ports.append(
                                ports_dict.get(avai_own_port))
                    host_tfo_ports_dict[host] = {
                        'avai_own_ports': avai_own_ports,
                        'avai_pair_ports': avai_pair_ports
                    }
        elif self.protocol == 'iSCSI':
            host_info_list = [initiatorlist[0].lower()]
            for host in host_info_list:
                host_tfo_ports_dict[host] = {
                    'avai_own_ports': own_ports,
                    'avai_pair_ports': pair_ports
                }

        LOG.debug('_find_tfo_port_pairs, '
                  'host_tfo_ports_dict: %(host_tfo_ports_dict)s.',
                  {'host_tfo_ports_dict': host_tfo_ports_dict})
        return host_tfo_ports_dict

    @dx_utils.FJDXTimeout()
    def terminate_connection(self, volume, connector,
                             timeout=None, **kwargs):
        """Disallow connection from connector."""
        LOG.debug('terminate_connection, '
                  'volume id: %(vid)s, protocol: %(prtcl)s, '
                  'timeout: %(timeout)s, options: %(options)s.',
                  {'vid': volume['id'], 'prtcl': self.protocol,
                   'timeout': timeout, 'options': kwargs})

        if self.failover:
            msg = (_('terminate_connection, '
                     'volume id: %(vid)s, '
                     'can not detach a failed over volume!')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        initiator = connector.get('initiator') if connector else None
        if initiator is None:
            # Delete host-affinity setting remained by unexpected error.
            self._unmap_lun(volume, None, force=True, timeout=self.req_sec)
            return {}

        map_num = None
        if volume['id'] in self.invalid_migration_list:
            msg = ('Live migration for backup volume is not allowed, '
                   'Terminate Connection Skip.')
            self.invalid_migration_list.remove(volume['id'])
            LOG.info(msg)
            self.conn = self._get_eternus_connection()

            map_num = self._get_map_number(connector)
            return map_num

        metadata = dx_utils.get_metadata(volume)
        dest_meta = metadata.get(CONSTANTS.FJ_REMOTE_DEST_META, None)
        ccm_flag = (six.text_type(dest_meta).lower() ==
                    CONSTANTS.FJ_REC_MIRROR)

        self.conn = self._get_eternus_connection()

        fc_zone_dict = {}
        try:
            replica_volume = self._is_replicated(volume)
            if replica_volume:
                map_num, fc_zone_dict = (
                    self._unmap_replication_volume(volume, connector))
            else:
                map_num, fc_zone_dict = (
                    self._unmap_ordinary_volume(volume, connector))
        except Exception:
            with excutils.save_and_reraise_exception():
                # When volume is set to REC Mirror, resume the session.
                if ccm_flag:
                    self._exec_ccm_script("resume", target=metadata)

        # CCM processing.
        if ccm_flag:
            self._exec_ccm_script("resume", target=metadata)

            ctxt = context.get_admin_context()
            db.volume_update(ctxt, volume['id'],
                             {'replication_extended_status': ""})

        return_val = {
            'map_num': map_num,
            'initiator_target_map': fc_zone_dict
        }
        LOG.debug('terminate_connection, Exit method. Return value: %s.',
                  return_val)
        return return_val

    @dx_utils.FJDXTimeout()
    def build_fc_init_tgt_map(self, connector, timeout=None):
        """Build parameter for Zone Manager."""
        LOG.debug('build_fc_init_tgt_map, '
                  'connector: %(connector)s, timeout: %(timeout)s.',
                  {'connector': connector, 'timeout': timeout})

        initiatorlist = self._find_initiator_names(connector)

        init_tgt_map = {}
        if not self.host_ag_dict:
            fc_zone_dict = self._get_enternus_fc_zone_info()
            for initiator in initiatorlist:
                host_wwpn = initiator.lower()
                ports = fc_zone_dict.get(host_wwpn, [])
                port_wwpns = []
                for port in ports:
                    if port in self.port_info_dict:
                        port_wwpn = self.port_info_dict[port]['port_wwpn']
                        port_wwpns.append(port_wwpn)
                init_tgt_map[host_wwpn] = port_wwpns
        else:
            for initiator in initiatorlist:
                host_wwpn = initiator.lower()
                if host_wwpn in self.host_ag_dict:
                    temp_ports = self.host_ag_dict[host_wwpn]['portlist']
                    port_wwpns = []
                    for port in temp_ports:
                        if port in self.port_info_dict:
                            port_wwpn = self.port_info_dict[port]['port_wwpn']
                            port_wwpns.append(port_wwpn)
                    init_tgt_map[host_wwpn] = port_wwpns

        LOG.debug('build_fc_init_tgt_map, '
                  'initiator target mapping: %s.', init_tgt_map)
        return init_tgt_map

    @dx_utils.FJDXLockutils('vol', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def extend_volume(self, volume, new_size, timeout=None):
        """Extend volume on ETERNUS."""
        LOG.debug('extend_volume, volume id: %(vid)s, '
                  'size: %(size)s, new_size: %(nsize)s, '
                  'timeout: %(timeout)s.',
                  {'vid': volume['id'], 'size': volume['size'],
                   'nsize': new_size, 'timeout': timeout})

        if self._is_replicated(volume):
            msg = (_('extend_volume, '
                     'volume id: %(vid)s, '
                     'can not extend a replication volume!')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if self.failover:
            msg = (_('extend_volume, '
                     'volume id: %(vid)s, '
                     'can not extend a failed over volume!')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self.conn = self._get_eternus_connection()

        volumename = self._get_volume_name(volume)

        # Get source volume instance.
        volume_instance = self._find_lun(volume)
        if not volume_instance:
            msg = (_('extend_volume, '
                     'source_volume: %(volumename)s, '
                     'not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('extend_volume, volumename: %(volumename)s, '
                  'volumesize: %(volumesize)u, '
                  'source volume instance: %(volume_instance)s.',
                  {'volumename': volumename,
                   'volumesize': new_size,
                   'volume_instance': volume_instance.path})

        # Get poolname from driver configuration file.
        pool_name, pool = self._find_pool_from_volume(volume_instance)

        # Check the existence of pool.
        if not pool:
            msg = (_('extend_volume, '
                     'eternus_pool: %(eternus_pool)s, '
                     'not found.')
                   % {'eternus_pool': pool_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Set pooltype.
        if 'RSP' in pool['InstanceID']:
            pooltype = CONSTANTS.RAIDGROUP
        else:
            pooltype = CONSTANTS.TPPOOL

        if pooltype == CONSTANTS.RAIDGROUP:
            extend_size = str(new_size - volume['size']) + 'gb'
            param_dict = {
                'volume-name': volumename,
                'rg-name': pool_name,
                'size': extend_size
            }
            rc, errordesc, data = self._exec_eternus_cli(
                'expand_volume',
                retry=self.retry_times,
                timeout=self.req_sec,
                **param_dict)

            if rc != 0:
                msg = (_('extend_volume, '
                         'volumename: %(volumename)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Message: %(job)s, '
                         'PoolType: %(pooltype)s.')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype],
                          'job': data})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('extend_volume, '
                      'volumename: %(volumename)s, '
                      'Return code: %(rc)lu, '
                      'Error: %(errordesc)s, '
                      'Pool Name: %(eternus_pool)s, '
                      'Pool Type: %(pooltype)s.',
                      {'volumename': volumename,
                       'rc': rc,
                       'errordesc': errordesc,
                       'eternus_pool': pool_name,
                       'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype]})

        else:  # Pooltype is TPPOOL.
            volumesize = new_size * units.Gi
            configservice = self._find_eternus_service(CONSTANTS.STOR_CONF)
            if not configservice:
                msg = (_('extend_volume, volume: %(volume)s, '
                         'volumename: %(volumename)s, '
                         'eternus_pool: %(eternus_pool)s, '
                         'Storage Configuration Service not found.')
                       % {'volume': volume,
                          'volumename': volumename,
                          'eternus_pool': pool_name})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('extend_volume, '
                      'CreateOrModifyElementFromStoragePool, '
                      'ConfigService: %(service)s, '
                      'ElementName: %(volumename)s, '
                      'InPool: %(eternus_pool)s, '
                      'ElementType: %(pooltype)u, '
                      'Size: %(volumesize)u, '
                      'TheElement: %(vol_instance)s.',
                      {'service': configservice,
                       'volumename': volumename,
                       'eternus_pool': pool_name,
                       'pooltype': pooltype,
                       'volumesize': volumesize,
                       'vol_instance': volume_instance.path})

            # Invoke method for extend volume.
            rc, errordesc, _x = self._exec_eternus_service(
                'CreateOrModifyElementFromStoragePool',
                configservice,
                timeout=self.req_sec,
                ElementName=volumename,
                InPool=pool,
                ElementType=self._pywbem_uint(pooltype, '16'),
                Size=self._pywbem_uint(volumesize, '64'),
                TheElement=volume_instance.path)

            if rc != 0:
                msg = (_('extend_volume, '
                         'volumename: %(volumename)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'PoolType: %(pooltype)s.')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype]})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('extend_volume, '
                      'volumename: %(volumename)s, '
                      'Return code: %(rc)lu, '
                      'Error: %(errordesc)s, '
                      'Pool Name: %(eternus_pool)s, '
                      'Pool Type: %(pooltype)s.',
                      {'volumename': volumename,
                       'rc': rc,
                       'errordesc': errordesc,
                       'eternus_pool': pool_name,
                       'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype]})

    @dx_utils.FJDXTimeout()
    def update_volume_stats(self, remote=False, timeout=None):
        """Get pool capacity."""
        if remote:
            filename = self.remote_filename
        else:
            filename = self.filename

        self.conn = self._get_eternus_connection(filename)

        poolname_list = self._get_drvcfg('EternusPool',
                                         filename=filename,
                                         multiple=True)
        self._find_pools(poolname_list, self.conn)
        if remote:
            return
        return self.stats, poolname_list

    def _get_drvcfg(self, tagname, filename=None, multiple=False):
        """Read from driver configuration file."""
        if not filename:
            # Set default configuration file name.
            filename = self.filename

        LOG.debug('_get_drvcfg, input[%(filename)s][%(tagname)s].',
                  {'filename': filename, 'tagname': tagname})

        tree = tree_parse(filename)
        elem = tree.getroot()

        if not multiple:
            ret = elem.findtext(".//" + tagname)
        else:
            ret = []
            for e in elem.findall(".//" + tagname):
                if e.text and (e.text not in ret):
                    ret.append(e.text)

        if not ret:
            msg = (_('_get_drvcfg, '
                     'filename: %(filename)s, '
                     'tagname: %(tagname)s, '
                     'data is None!! '
                     'Please edit driver configuration file and correct.')
                   % {'filename': filename,
                      'tagname': tagname})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return ret

    def _get_eternus_connection(self, filename=None):
        """Return WBEM connection."""
        LOG.debug('_get_eternus_connection, filename: %s.', filename)

        ip = self._get_drvcfg('EternusIP', filename)
        port = self._get_drvcfg('EternusPort', filename)
        user = self._get_drvcfg('EternusUser', filename)
        password = self._get_drvcfg('EternusPassword', filename)

        # IPv6.
        if ip.find(':') != -1:
            ip = '[' + ip + ']'
            self.use_ipv6 = True

        if port == str(CONSTANTS.HTTPS_PORT):
            LOG.debug('_get_eternus_connection, https.')

            url = 'https://' + ip + ':' + port
            two_way_authentication = self.configuration.two_way_authentication
            if two_way_authentication:
                key_file = self.configuration.driver_client_cert_key
                cert_file = self.configuration.driver_client_cert
                key_file_exists = os.path.exists(key_file)
                cert_file_exists = os.path.exists(cert_file)

            ca_certs_file = self.configuration.fujitsu_ca_certs_path
            ca_certs_file_exists = os.path.exists(ca_certs_file)
            no_verification = not bool(
                self.configuration.fujitsu_ssl_verification)

            if not ca_certs_file_exists and not no_verification:
                msg = _('need verification and no ca_certs_file, Error.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if not two_way_authentication:
                pywbem.cim_http.wbem_request = https.wbem_request
                pywbem.cim_operations.wbem_request = https.wbem_request
                try:
                    conn = pywbem.WBEMConnection(
                        url,
                        (user, password),
                        default_namespace='root/eternus',
                        ca_certs=ca_certs_file,
                        no_verification=no_verification)
                except TypeError:
                    conn = pywbem.WBEMConnection(
                        url,
                        (user, password),
                        default_namespace='root/eternus')
                    msg = _('pywbem version has no no_verification parameter, '
                            'This version must verificate the certification, '
                            'and CA PATH is: /etc/pki/ca-trust/extracted/'
                            'openssl/ca-bundle.trust.crt.')
                    LOG.warning(msg)
            else:
                if not (key_file_exists and cert_file_exists):
                    msg = (_('_get_eternus_connection, '
                             'key file: %(key_file)s '
                             'or cert file: %(cert_file)s does not exist!!')
                           % {'key_file': key_file, 'cert_file': cert_file})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                try:
                    conn = pywbem.WBEMConnection(
                        url,
                        (user, password),
                        default_namespace='root/eternus',
                        x509={
                            "key_file": key_file,
                            "cert_file": cert_file
                        },
                        ca_certs=ca_certs_file,
                        no_verification=no_verification)
                except TypeError:
                    msg = _('pywbem version is not compatible, Error.')
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        elif port == str(CONSTANTS.HTTP_PORT):
            LOG.debug('_get_eternus_connection, http.')

            url = 'http://' + ip + ':' + port
            conn = pywbem.WBEMConnection(url,
                                         (user, password),
                                         default_namespace='root/eternus')
        else:
            msg = (_('_get_eternus_connection, '
                     'port: %(port)s '
                     'does not support!! '
                     'Please edit driver configuration file.')
                   % {'port': port})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not conn:
            msg = (_('_get_eternus_connection, '
                     'filename: %(filename)s, '
                     'ip: %(ip)s, '
                     'port: %(port)s, '
                     'user: %(user)s, '
                     'password: ****, '
                     'url: %(url)s, '
                     'FAILED!!')
                   % {'filename': filename,
                      'ip': ip,
                      'port': port,
                      'user': user,
                      'url': url})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_get_eternus_connection, conn: %s.', conn)
        return conn

    def _get_volume_name(self, volume, use_id=False):
        """Get volume_name on ETERNUS from volume on OpenStack."""
        LOG.debug('_get_volume_name, volume_id: %s.', volume['id'])

        if not use_id and volume['provider_location']:
            location = eval(volume['provider_location'])
            if 'vol_name' in location:
                LOG.debug('_get_volume_name, by provider_location, '
                          'vol_name: %s.', location['vol_name'])
                return location['vol_name']

        id_code = volume['id']

        m = hashlib.md5()
        m.update(id_code.encode('utf-8'))

        # Pylint: disable=E1121.
        volumename = base64.urlsafe_b64encode(m.digest()).decode()
        vol_name = CONSTANTS.VOL_PREFIX + six.text_type(volumename)

        if self.model_name == CONSTANTS.DX_S2:
            LOG.debug('_get_volume_name, volume name is 16 digit.')
            vol_name = vol_name[:16]

        LOG.debug('_get_volume_name, by volume id, '
                  'vol_name: %s.', vol_name)
        return vol_name

    def _find_snap_pool(self, snapshot):
        """Find pools for SDV and return the first avaiable pool instance."""
        sdv_poollist = self._get_drvcfg('EternusSnapPool', multiple=True)
        LOG.debug('_find_snap_pool, pool names: %s.', sdv_poollist)

        try:
            tppoollist = self._enum_eternus_instances(
                'FUJITSU_ThinProvisioningPool', retry=self.retry_times)
            rgpoollist = self._enum_eternus_instances(
                'FUJITSU_RAIDStoragePool', retry=self.retry_times)
        except Exception:
            msg = (_('_find_snap_pool, '
                     'pool names: %(names)s, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.')
                   % {'names': sdv_poollist})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        tppools = [(tppool, 'TPP') for tppool in tppoollist]
        rgpools = [(rgpool, 'RAID') for rgpool in rgpoollist]

        poollist = rgpools + tppools
        # Poolnamelist = [pool['ElementName'] for pool in poollist].
        snap_pool = None

        for pool, ptype in poollist:
            poolname = pool['ElementName']

            LOG.debug('_find_snap_pool, '
                      'pool: %(pool)s, ptype: %(ptype)s.',
                      {'pool': poolname, 'ptype': ptype})
            if poolname in sdv_poollist:
                if ptype == 'RAID':
                    pool_free_space = pool['RemainingManagedSpace']
                    if pool_free_space >= 24 * units.Mi:
                        try:
                            volume_list = self._assoc_eternus_names(
                                pool.path,
                                retry=self.retry_times,
                                AssocClass='FUJITSU_AllocatedFromStoragePool',
                                ResultClass='FUJITSU_StorageVolume')
                            volume_count = len(volume_list)
                        except Exception:
                            msg = (_('_find_snap_pool, '
                                     'pool name: %(poolname)s, '
                                     'AssociatorNames, '
                                     'cannot connect to ETERNUS.')
                                   % {'poolname': poolname})
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)

                        try:
                            sdpv_list = self._assoc_eternus_names(
                                pool.path,
                                retry=self.retry_times,
                                AssocClass='FUJITSU_AllocatedFromStoragePool',
                                ResultClass='FUJITSU_SDPVPool')
                            volume_count += len(sdpv_list)
                        except Exception:
                            msg = (_('_find_snap_pool, '
                                     'pool name: %(poolname)s, '
                                     'AssociatorNames, '
                                     'cannot connect to ETERNUS.')
                                   % {'poolname': poolname})
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)

                        if volume_count < CONSTANTS.STD_VOLUME_COUNT_MAX:
                            pool['pooltype'] = CONSTANTS.POOL_TYPE_dic[
                                CONSTANTS.RAIDGROUP]
                            snap_pool = pool
                            break
                # Ptype == 'TPP'
                else:
                    total_b = pool['TotalManagedSpace']
                    free_b = pool['RemainingManagedSpace']
                    if hasattr(pool, 'provisioned_capacity_gb'):
                        prov_b = pool.provisioned_capacity_gb * units.Gi
                    else:
                        prov_b = total_b - free_b
                    self.max_over_subscription_ratio = \
                        self.configuration.max_over_subscription_ratio
                    useable_b = (total_b * self.max_over_subscription_ratio -
                                 prov_b)
                    if useable_b >= 24 * units.Mi:
                        pool['pooltype'] = CONSTANTS.POOL_TYPE_dic[
                            CONSTANTS.TPPOOL]
                        snap_pool = pool
                        break

        LOG.debug('_find_snap_pool, snap: %(sid)s, target pool: %(pool)s.',
                  {'sid': snapshot['id'],
                   'pool': snap_pool['ElementName'] if snap_pool else None})
        return snap_pool

    def _find_all_pools_instances(self, conn):
        LOG.debug('_find_all_pools_instances, conn: %s', conn)

        try:
            tppoollist = self._enum_eternus_instances(
                'FUJITSU_ThinProvisioningPool',
                conn=conn,
                retry=self.retry_times)
            rgpoollist = self._enum_eternus_instances(
                'FUJITSU_RAIDStoragePool',
                conn=conn,
                retry=self.retry_times)
        except Exception:
            msg = _('_find_all_pools_instances, '
                    'EnumerateInstances cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Make total pools list.
        tppools = [(tppool, 'TPP') for tppool in tppoollist]
        rgpools = [(rgpool, 'RAID') for rgpool in rgpoollist]
        poollist = tppools + rgpools

        LOG.debug('_find_all_pools_instances, poollist: %s', len(poollist))
        return poollist

    def _find_pools(self, poolname_list, conn, remote=False,
                    poolinstances_list=None):
        """Find pool instances by using pool name on ETERNUS."""
        LOG.debug('_find_pools, pool names: %(poolname_list)s, '
                  'remote: %(remote)s.',
                  {'poolname_list': poolname_list,
                   'remote': remote})
        target_poolname = list(poolname_list)
        pools = []

        # Get pools info form CIM instance(include info about instance path).
        if not poolinstances_list:
            poollist = self._find_all_pools_instances(conn)
            is_create = False
        else:
            poollist = poolinstances_list
            is_create = True

        for pool, ptype in poollist:
            poolname = pool['ElementName']

            LOG.debug('_find_pools, '
                      'pool: %(pool)s, ptype: %(ptype)s.',
                      {'pool': poolname, 'ptype': ptype})
            volume_count = None
            if poolname in target_poolname:
                if ptype == 'TPP':
                    param_dict = {
                        'pool-name': poolname
                    }
                    rc, errordesc, data = self._exec_eternus_cli(
                        'show_pool_provision',
                        remote=remote,
                        retry=self.retry_times,
                        timeout=self.req_sec,
                        **param_dict)

                    if rc != 0:
                        msg = (_('_find_pools, show_pool_provision, '
                                 'pool name: %(pool_name)s, '
                                 'Return code: %(rc)lu, '
                                 'Error: %(errordesc)s, '
                                 'Message: %(job)s, '
                                 'Remote: %(remote)s.')
                               % {'pool_name': poolname,
                                  'rc': rc,
                                  'errordesc': errordesc,
                                  'job': data,
                                  'remote': remote})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

                    pool.provisioned_capacity_mb = data
                elif ptype == 'RAID':
                    # Get volume number and fragment capacity information
                    # only at creation time.
                    try:
                        volume_list = self._assoc_eternus_names(
                            pool.path,
                            conn=conn,
                            retry=self.retry_times,
                            AssocClass='FUJITSU_AllocatedFromStoragePool',
                            ResultClass='FUJITSU_StorageVolume')

                        volume_count = len(volume_list)
                    except Exception:
                        msg = (_('_find_pools, '
                                 'poolname: %(poolname)s, '
                                 'pooltype: %(ptype)s, '
                                 'Associator Names, '
                                 'cannot connect to ETERNUS.')
                               % {'ptype': ptype,
                                  'poolname': poolname})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

                    try:
                        sdpv_list = self._assoc_eternus_names(
                            pool.path,
                            retry=self.retry_times,
                            AssocClass='FUJITSU_AllocatedFromStoragePool',
                            ResultClass='FUJITSU_SDPVPool')
                        volume_count += len(sdpv_list)
                    except Exception:
                        msg = (_('_find_pools, '
                                 'pool name: %(poolname)s, '
                                 'Associator Names FUJITSU_SDPVPool, '
                                 'cannot connect to ETERNUS.')
                               % {'poolname': poolname})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

                    try:
                        fragment_list = self._assoc_eternus(
                            pool.path,
                            conn=conn,
                            retry=self.retry_times,
                            PropertyList=['NumberOfBlocks'],
                            ResultClass='FUJITSU_FreeExtent')

                        if fragment_list:
                            fragment_size = max(
                                fragment_list,
                                key=lambda x: x['NumberOfBlocks'])
                        else:
                            fragment_size = {'NumberOfBlocks': 0}
                    except Exception:
                        # S2 models do not support this query.
                        fragment_size = {'NumberOfBlocks': -1}

                    pool.fragment_size = fragment_size['NumberOfBlocks']

                poolinfo = self.create_pool_info(pool, volume_count, ptype)

                target_poolname.remove(poolname)
                pools.append((poolinfo, poolname))

            if not target_poolname:
                break

        if not pools:
            LOG.warning('_find_pools, all the EternusPools in driver '
                        'configuration file are not exist. '
                        'Please edit driver configuration file.')

        # Sort pools in the order defined in driver configuration file.
        sorted_pools = (
            [pool for name in poolname_list for pool, pname in pools
             if name == pname])

        LOG.debug('_find_pools, '
                  'pools: %(pools)s, '
                  'notfound_pools: %(notfound_pools)s.',
                  {'pools': pools,
                   'notfound_pools': target_poolname})
        pools_stats = {'pools': []}
        for pool in sorted_pools:
            single_pool = {}
            if pool['type'] == 'TPP':
                thin_enabled = True
                max_ratio = self.configuration.max_over_subscription_ratio
            else:
                thin_enabled = False
                max_ratio = 1
                single_pool['total_volumes'] = pool['volume_count']
                single_pool['fragment_capacity_mb'] = \
                    pool['fragment_capacity_mb']

            single_pool.update(dict(
                path=pool['path'],
                pool_name=pool['name'],
                total_capacity_gb=pool['total_capacity_gb'],
                free_capacity_gb=pool['free_capacity_gb'],
                provisioned_capacity_gb=pool['provisioned_capacity_gb'],
                useable_capacity_gb=pool['useable_capacity_gb'],
                thin_provisioning_support=thin_enabled,
                thick_provisioning_support=not thin_enabled,
                max_over_subscription_ratio=max_ratio,
            ))
            if self.filter_function:
                single_pool['filter_function'] = self.filter_function
            if is_create:
                single_pool['useable_capacity_mb'] = \
                    pool['useable_capacity_mb']
            single_pool['multiattach'] = True
            pools_stats['pools'].append(single_pool)

        if not remote:
            self.stats['pools'] = pools_stats['pools']
            return self.stats, target_poolname

        return pools_stats, target_poolname

    def _find_pools_resources(self, poolname_list, conn, filename):
        """Find volumes/snapshots by using pool name on ETERNUS."""
        LOG.debug('_find_pools_resources, pool names: %s.', poolname_list)
        target_poolname = list(poolname_list)
        resources = []

        # Get pools info form CIM instance(include info about instance path).
        try:
            tppoollist = self._enum_eternus_instances(
                'FUJITSU_ThinProvisioningPool', conn=conn)
            rgpoollist = self._enum_eternus_instances(
                'FUJITSU_RAIDStoragePool', conn=conn)
        except Exception:
            msg = (_('_find_pools_resources, '
                     'pool names: %(eternus_pool)s, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.')
                   % {'eternus_pool': target_poolname})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Make total pools list.
        tppools = [tppool for tppool in tppoollist]
        rgpools = [rgpool for rgpool in rgpoollist]
        poollist = tppools + rgpools

        # Get volumes from pools.
        for pool in poollist:
            poolname = pool['ElementName']

            if poolname in target_poolname:
                vol_instance_list = self._assoc_eternus(
                    pool.path,
                    conn=conn,
                    AssocClass='FUJITSU_AllocatedFromStoragePool',
                    ResultClass='FUJITSU_StorageVolume')

                LOG.debug('_find_pools_resources, '
                          'vollist: %(vollist)s.',
                          {'vollist': vol_instance_list})

                target_poolname.remove(poolname)
                resources.extend(vol_instance_list)

            if not target_poolname:
                break

        if not resources:
            LOG.warning('_find_pools_resources, '
                        'all the EternusPools in driver '
                        'configuration file are not exist. '
                        'Please edit driver configuration file.')

        LOG.debug('_find_pools_resources,'
                  'resources: %(resources)s,'
                  'notfound_pools: %(notfound_pools)s.',
                  {'resources': resources,
                   'notfound_pools': target_poolname})

        return resources

    def _find_eternus_service(self, classname, conn=None):
        """Find CIM instance about service information."""
        LOG.debug('_find_eternus_service, classname: %s.', classname)

        try:
            services = self._enum_eternus_instance_names(
                six.text_type(classname),
                conn=conn,
                retry=self.retry_times)
        except Exception:
            msg = (_('_find_eternus_service, '
                     'classname: %(classname)s, '
                     'EnumerateInstanceNames, '
                     'cannot connect to ETERNUS.')
                   % {'classname': classname})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        ret = services[0]
        LOG.debug('_find_eternus_service, '
                  'classname: %(classname)s, '
                  'ret: %(ret)s.',
                  {'classname': classname, 'ret': ret})
        return ret

    @dx_utils.FJDXLockutils('SMIS-exec', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _exec_eternus_service(self, classname, instanceNameList,
                              conn=None, timeout=None, **param_dict):
        """Execute SMI-S Method."""
        LOG.debug('_exec_eternus_service, '
                  'classname: %(a)s, '
                  'instanceNameList: %(b)s, '
                  'timeout: %(t)s, '
                  'parameters: %(c)s.',
                  {'a': classname,
                   'b': instanceNameList,
                   't': timeout,
                   'c': param_dict})
        rc = None

        if not conn:
            conn = self.conn

        # Use InvokeMethod.
        try:
            rc, retdata = self._exec_eternus_smis_InvokeMethod(
                classname,
                instanceNameList,
                conn=conn,
                timeout=self.req_sec,
                **param_dict)
        except Exception:
            if rc is None:
                msg = (_('_exec_eternus_service, '
                         'classname: %(classname)s, '
                         'InvokeMethod, '
                         'cannot connect to ETERNUS.')
                       % {'classname': classname})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        # If the result has job information, wait for completion.
        if "Job" in retdata:
            rc = self._wait_for_job_complete(conn, retdata)

        errordesc = CONSTANTS.RETCODE_dic.get(six.text_type(rc),
                                              CONSTANTS.UNDEF_MSG)

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_service, '
                  'classname: %(a)s, '
                  'instanceNameList: %(b)s, '
                  'parameters: %(c)s, '
                  'Return code: %(rc)s, '
                  'Error: %(errordesc)s, '
                  'Retrun data: %(retdata)s.',
                  {'a': classname,
                   'b': instanceNameList,
                   'c': param_dict,
                   'rc': rc,
                   'errordesc': errordesc,
                   'retdata': retdata})
        return ret

    def _enum_eternus_instances(self, classname,
                                conn=None, retry=CONSTANTS.TIMES_MIN,
                                retry_interval=CONSTANTS.RETRY_INTERVAL,
                                **param_dict):
        """Enumerate Instances."""
        LOG.debug('_enum_eternus_instances, classname: %s.', classname)

        if not conn:
            conn = self.conn

        for retry_num in range(retry):
            try:
                ret = self._exec_eternus_smis_EnumerateInstances(
                    classname,
                    conn=conn,
                    timeout=self.req_sec,
                    **param_dict)
                break
            except Exception as e:
                LOG.info('_enum_eternus_instances, '
                         'reason: %(reason)s, try (%(retrynum)d).',
                         {'reason': e.args,
                          'retrynum': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = _('_enum_eternus_instances, Error.')
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_enum_eternus_instances, enum %d instances.', len(ret))
        return ret

    def _enum_eternus_instance_names(self, classname,
                                     conn=None, retry=CONSTANTS.TIMES_MIN,
                                     retry_interval=CONSTANTS.RETRY_INTERVAL):
        """Enumerate Instance Names."""
        LOG.debug('_enum_eternus_instance_names, classname: %s.', classname)

        if not conn:
            conn = self.conn

        for retry_num in range(retry):
            try:
                ret = self._exec_eternus_smis_EnumerateInstanceNames(
                    classname,
                    conn=conn,
                    timeout=self.req_sec)
                break
            except Exception as e:
                LOG.info('_enum_eternus_instance_names, '
                         'reason: %(reason)s, try (%(retrynum)d).',
                         {'reason': e.args,
                          'retrynum': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = _('_enum_eternus_instance_names, Error.')
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_enum_eternus_instance_names, enum %d names.', len(ret))
        return ret

    def _get_eternus_instance(self, classname,
                              conn=None, AllowNone=False,
                              retry=CONSTANTS.TIMES_MIN,
                              retry_interval=CONSTANTS.RETRY_INTERVAL,
                              **param_dict):
        """Get Instance."""
        LOG.debug('_get_eternus_instance, '
                  'classname: %(cls)s, param: %(param)s.',
                  {'cls': classname, 'param': param_dict})

        if not conn:
            conn = self.conn

        ret = None
        for retry_num in range(retry):
            try:
                ret = self._exec_eternus_smis_GetInstance(
                    classname,
                    conn=conn,
                    timeout=self.req_sec,
                    **param_dict)
                break
            except Exception as e:
                if e.args[0] == 6 and AllowNone:
                    break
                else:
                    LOG.info('_get_eternus_instance, '
                             'reason: %(reason)s, try (%(retrynum)d).',
                             {'reason': e.args,
                              'retrynum': (retry_num + 1)})
                    time.sleep(retry_interval)
                    continue
        else:
            msg = _('_get_eternus_instance, Error.')
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_get_eternus_instance, ret: %s.', ret)
        return ret

    def _assoc_eternus(self, classname,
                       conn=None, retry=CONSTANTS.TIMES_MIN,
                       retry_interval=CONSTANTS.RETRY_INTERVAL, **param_dict):
        """Associator."""
        LOG.debug('_assoc_eternus, '
                  'classname: %(cls)s, param: %(param)s.',
                  {'cls': classname, 'param': param_dict})

        if not conn:
            conn = self.conn

        for retry_num in range(retry):
            try:
                ret = self._exec_eternus_smis_Associators(
                    classname,
                    conn=conn,
                    timeout=self.req_sec,
                    **param_dict)
                break
            except Exception as e:
                LOG.info('_assoc_eternus, '
                         'reason: %(reason)s, retry (%(retrynum)d).',
                         {'reason': (e.args),
                          'retrynum': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = _('_assoc_eternus, Error')
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_assoc_eternus, enum %d instances.', len(ret))
        return ret

    def _assoc_eternus_names(self, classname,
                             conn=None, retry=CONSTANTS.TIMES_MIN,
                             retry_interval=CONSTANTS.RETRY_INTERVAL,
                             **param_dict):
        """Associator Names."""
        LOG.debug('_assoc_eternus_names, '
                  'classname: %(cls)s, param: %(param)s.',
                  {'cls': classname, 'param': param_dict})

        if not conn:
            conn = self.conn

        for retry_num in range(retry):
            try:
                ret = self._exec_eternus_smis_AssociatorNames(
                    classname,
                    conn=conn,
                    timeout=self.req_sec,
                    **param_dict)
                break
            except Exception as e:
                LOG.info('_assoc_eternus_names, '
                         'reason: %(reason)s, try (%(retrynum)d).',
                         {'reason': e.args,
                          'retrynum': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = _('_assoc_eternus_names, Error.')
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_assoc_eternus_names, enum %d names.', len(ret))
        return ret

    def _create_eternus_instance_name(self, classname, bindings):
        """Create CIM InstanceName from classname and bindings."""
        LOG.debug('_create_eternus_instance_name, '
                  'classname: %(cls)s, bindings: %(bind)s.',
                  {'cls': classname, 'bind': bindings})

        bindings['CreationClassName'] = classname
        bindings['SystemCreationClassName'] = 'FUJITSU_StorageComputerSystem'

        try:
            instancename = self._exec_eternus_smis_CIMInstanceName(
                classname,
                namespace='root/eternus',
                keybindings=bindings,
                timeout=self.req_sec)
        except NameError:
            instancename = None

        LOG.debug('_create_eternus_instance_name, ret: %s.', instancename)
        return instancename

    def _find_lun(self, volume, use_service_name=False):
        """Find lun instance from volume class or volumename on ETERNUS."""
        LOG.debug('_find_lun, volume id: %s.', volume['id'])
        volumeinstance = None
        volumename = self._get_volume_name(volume)

        if not use_service_name:
            conn = self.conn
        else:
            conf_filename = self.filename
            conn = self._get_eternus_connection(conf_filename)

        try:
            location = eval(volume['provider_location'])
            classname = location['classname']
            bindings = location['keybindings']
            systemname = location['keybindings']['SystemName']
            issuccess = True
            if systemname != self.system_name:
                LOG.debug('_find_lun, The systemname of the volume '
                          'is inconsistent with eternus.')
            elif classname and bindings:
                LOG.debug('_find_lun, '
                          'classname: %(classname)s, '
                          'bindings: %(bindings)s.',
                          {'classname': classname,
                           'bindings': bindings})
                volume_instance_name = (
                    self._create_eternus_instance_name(classname, bindings))
                LOG.debug('_find_lun, '
                          'volume_insatnce_name: %(volume_instance_name)s.',
                          {'volume_instance_name': volume_instance_name})

                vol_instance = self._get_eternus_instance(
                    volume_instance_name,
                    conn=conn,
                    AllowNone=True,
                    retry=self.retry_times)

                if vol_instance and vol_instance['ElementName'] == volumename:
                    volumeinstance = vol_instance
        except Exception:
            issuccess = False
            LOG.debug('_find_lun, '
                      'Cannot get volume instance from provider location, '
                      'Search all volume using EnumerateInstanceNames.')

        if not issuccess and self.model_name == CONSTANTS.DX_S2:
            # For old version.
            LOG.debug('_find_lun, '
                      'volumename: %(volumename)s.',
                      {'volumename': volumename})

            vol_name = {
                'source-name': volumename
            }
            # Get volume instance from volumename on ETERNUS.
            volumeinstance = self._find_lun_with_listup(
                conn=conn, **vol_name)

        LOG.debug('_find_lun, ret: %s.', volumeinstance)
        return volumeinstance

    def _find_copysession(self, vol_instance, conn=None):
        """Find copysession from volumename on ETERNUS."""
        LOG.debug('_find_copysession, volume name: %s.',
                  vol_instance['ElementName'])

        if not conn:
            conn = self.conn

        try:
            cpsessionlist = self._exec_eternus_smis_ReferenceNames(
                vol_instance.path,
                conn=conn,
                timeout=self.req_sec,
                ResultClass='FUJITSU_StorageSynchronized')
        except Exception:
            msg = (_('_find_copysession, '
                     'ReferenceNames, '
                     'vol_instance: %(vol_instance_path)s, '
                     'Cannot connect to ETERNUS.')
                   % {'vol_instance_path': vol_instance.path})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_copysession, '
                  'cpsessionlist: %(cpsessionlist)s.',
                  {'cpsessionlist': cpsessionlist})

        LOG.debug('_find_copysession, ret: %s.', cpsessionlist)
        return cpsessionlist

    def _wait_for_copy_complete(self, cpsession, conn=None):
        """Wait for the completion of copy."""
        LOG.debug('_wait_for_copy_complete, cpsession: %s.', cpsession)

        complete = False

        if not conn:
            conn = self.conn

        while True:
            try:
                cpsession_instance = self._exec_eternus_smis_GetInstance(
                    cpsession,
                    conn=conn,
                    timeout=self.req_sec,
                    LocalOnly=False)
            except Exception:
                cpsession_instance = None

            # If copy session is none,
            # It means copy session was finished, break and return.
            if not cpsession_instance:
                complete = True
                break

            # There is no case that copy session is SnapOPC.
            # If copy session is OPC, loop is continued.
            if (cpsession_instance['CopyType'] == CONSTANTS.EC_REC and
                    cpsession_instance['CopyState'] not in [2, 3]):
                break

            LOG.debug('_wait_for_copy_complete, '
                      'find target copysession, '
                      'wait for end of copysession.')

            if cpsession_instance['CopyState'] == CONSTANTS.BROKEN:
                msg = (_('_wait_for_copy_complete, '
                         'cpsession: %(cpsession)s, '
                         'copysession state is BROKEN.')
                       % {'cpsession': cpsession})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            time.sleep(10)

        return complete

    def _delete_copysession(self, cpsession, conn=None):
        """Delete copysession."""
        LOG.debug('_delete_copysession: cpssession: %s.', cpsession)

        try:
            cpsession_instance = self._get_eternus_instance(
                cpsession,
                conn=conn,
                retry=self.retry_times,
                LocalOnly=False)
        except Exception:
            LOG.info('_delete_copysession, '
                     'The copysession was already completed.')
            return

        copytype = cpsession_instance['CopyType']

        # Set oparation code.
        # SnapOPC: 19 (Return To ResourcePool)
        # OPC:8 (Detach)
        # EC/REC:8 (Detach)
        operation = CONSTANTS.OPERATION_dic.get(copytype, None)
        if not operation:
            msg = (_('_delete_copysession, '
                     'copy session type is undefined! '
                     'copy session: %(cpsession)s, '
                     'copy type: %(copytype)s.')
                   % {'cpsession': cpsession,
                      'copytype': copytype})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        repservice = self._find_eternus_service(CONSTANTS.REPL, conn=conn)
        if not repservice:
            msg = (_('_delete_copysession, '
                     'Cannot find Replication Service.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Invoke method for delete copysession.
        rc, errordesc, _x = self._exec_eternus_service(
            'ModifyReplicaSynchronization',
            repservice,
            conn=conn,
            timeout=self.req_sec,
            Operation=self._pywbem_uint(operation, '16'),
            Synchronization=cpsession,
            Force=True,
            WaitForCopyState=self._pywbem_uint(15, '16'))

        LOG.debug('_delete_copysession, '
                  'copysession: %(cpsession)s, '
                  'operation: %(operation)s, '
                  'Return code: %(rc)lu, '
                  'Error: %(errordesc)s.',
                  {'cpsession': cpsession,
                   'operation': operation,
                   'rc': rc,
                   'errordesc': errordesc})

        if rc == CONSTANTS.COPYSESSION_NOT_EXIST:
            LOG.debug('_delete_copysession, '
                      'cpsession: %(cpsession)s, '
                      'copysession does not exist.',
                      {'cpsession': cpsession})
        elif rc != 0:
            msg = (_('_delete_copysession, '
                     'copysession: %(cpsession)s, '
                     'operation: %(operation)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'cpsession': cpsession,
                      'operation': operation,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def _get_target_port(self):
        """Return target portid."""
        LOG.debug('_get_target_port, protocol: %s.', self.protocol)

        target_portlist = []
        if self.protocol == 'FC':
            prtcl_endpoint = 'FUJITSU_SCSIProtocolEndpoint'
            connection_type = 2
        elif self.protocol == 'iSCSI':
            prtcl_endpoint = 'FUJITSU_iSCSIProtocolEndpoint'
            connection_type = 7

        try:
            tgtportlist = self._enum_eternus_instances(
                prtcl_endpoint, retry=self.retry_times)
        except Exception:
            msg = _('_get_target_port, '
                    'EnumerateInstances, '
                    'cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for tgtport in tgtportlist:
            # Check protocol of tgtport.
            if tgtport['ConnectionType'] != connection_type:
                continue

            # Check: if port is for remote copy, continue.
            if tgtport['RAMode'] & 0x7B != 0x00:
                continue

            # Check: if port is for StorageCluster, continue.
            if 'SCGroupNo' in tgtport:
                continue

            target_portlist.append(tgtport)

            LOG.debug('_get_target_port, '
                      'connection type: %(cont)s, '
                      'ramode: %(ramode)s.',
                      {'cont': tgtport['ConnectionType'],
                       'ramode': tgtport['RAMode']})

        LOG.debug('_get_target_port, target port: %s.', target_portlist)

        if not target_portlist:
            msg = (_('_get_target_port, '
                     'protcol: %(protocol)s, '
                     'target_port not found.')
                   % {'protocol': self.protocol})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_get_target_port, ret: %s.', target_portlist)
        return target_portlist

    @dx_utils.FJDXLockutils('connect', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _map_lun(self, volume, connector, ag_name, timeout=None):
        """Map volume to host."""
        vol_id = volume['id']
        LOG.debug('_map_lun, '
                  'volume id: %(vol_id)s, '
                  'connector: %(connector)s, '
                  'affinity group name: %(ag_name)s, '
                  'timeout: %(timeout)s.',
                  {'vol_id': vol_id,
                   'connector': connector,
                   'ag_name': ag_name,
                   'timeout': timeout})

        # Get volume No in affinity group.
        option = {
            'ag-name': ag_name
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'show_affinity_group',
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)

        if rc != 0:
            msg = (_('_map_lun, '
                     'show_affinity_group, '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not clidata:
            msg = (_('_map_lun, '
                     'show_affinity_group, '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s. '
                     'There is no lungroup: %(ag)s in ETERNUS .')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata,
                      'ag': ag_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        idx = 0
        used_nos = []
        for data in clidata:
            no = int(data['Lun Num'])
            used_nos.append(no)
            idx = no if idx < no else idx

        lun_no = idx + 1
        if CONSTANTS.MAX_AG_LUN_NO <= lun_no:
            temp = range(CONSTANTS.MAX_AG_LUN_NO)
            unused_num = None
            for t in temp:
                if t not in used_nos:
                    unused_num = t
                    break

            lun_no = unused_num
            if lun_no is None:
                msg = (_('_map_lun, '
                         'volume id: %(vol_id)s, '
                         'the volumes in affinity group(%(ag_name)s) '
                         'is full(MAX count: %(count)s).')
                       % {'vol_id': vol_id,
                          'ag_name': ag_name,
                          'count': CONSTANTS.MAX_AG_LUN_NO})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_map_lun, '
                  'volume id: %(vol_id)s, '
                  'affinity group name: %(ag_name)s, '
                  'volume no: %(lun_no)s.',
                  {'vol_id': vol_id,
                   'lun_no': lun_no,
                   'ag_name': ag_name})

        # Set affinity-group.
        volumename = self._get_volume_name(volume)
        option = {
            'ag-name': ag_name,
            'volume-name': volumename,
            'lun': lun_no
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'set_affinity_group',
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)

        if rc != 0:
            # If error, check if the volume has been mapped.
            option = {
                'ag-name': ag_name
            }
            rc_tmp, emsg_tmp, clidata_tmp = self._exec_eternus_cli(
                'show_affinity_group',
                retry=self.retry_times,
                timeout=self.req_sec,
                **option)

            if rc_tmp != 0:
                msg = (_('_map_lun, '
                         'show_affinity_group, '
                         'Return code: %(rc)lu, '
                         'Error: %(emsg)s, '
                         'Message: %(clidata)s.')
                       % {'rc': rc_tmp,
                          'emsg': emsg_tmp,
                          'clidata': clidata_tmp})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            temp_no = None
            for data in clidata_tmp:
                name = data['Lun Name']
                if volumename == name:
                    temp_no = int(data['Lun Num'])
                    break

            if temp_no is not None:
                lun_no = temp_no
            else:
                msg = (_('_map_lun, '
                         'set_affinity_group, '
                         'Return code: %(rc)lu, '
                         'Error: %(emsg)s, '
                         'Message: %(clidata)s.')
                       % {'rc': rc,
                          'emsg': emsg,
                          'clidata': clidata})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_map_lun, '
                  'volume id: %(vol_id)s, '
                  'affinity group name: %(ag_name)s, '
                  'volume No in affinity group: %(lun_no)s. Successful.',
                  {'vol_id': vol_id,
                   'lun_no': lun_no,
                   'ag_name': ag_name})
        return lun_no

    def _find_initiator_names(self, connector):
        """Return initiator names."""
        initiatornamelist = []

        if self.protocol == 'FC' and connector['wwpns']:
            LOG.debug('_find_initiator_names, wwpns: %s.',
                      connector['wwpns'])
            initiatornamelist = connector['wwpns']
        elif self.protocol == 'iSCSI' and connector['initiator']:
            LOG.debug('_find_initiator_names, initiator: %s.',
                      connector['initiator'])
            initiatornamelist.append(connector['initiator'])

        if not initiatornamelist:
            msg = (_('_find_initiator_names, '
                     'connector: %(connector)s, '
                     'initiator not found.')
                   % {'connector': connector})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_initiator_names, '
                  'initiator list: %s.', initiatornamelist)

        return initiatornamelist

    def _find_affinity_group(self, connector, vol_instance=None, conn=None):
        """Find affinity group from connector."""
        LOG.debug('_find_affinity_group, vol_instance: %s.', vol_instance)

        affinity_grouplist = []
        initiatorlist = self._find_initiator_names(connector)

        if not vol_instance:
            try:
                aglist = self._enum_eternus_instance_names(
                    'FUJITSU_AffinityGroupController',
                    conn=conn,
                    retry=self.retry_times)
            except Exception:
                msg = (_('_find_affinity_group, '
                         'connector: %(connector)s, '
                         'EnumerateInstanceNames, '
                         'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_find_affinity_group, '
                      'affinity_groups:%s.', aglist)
        else:
            try:
                aglist = self._assoc_eternus_names(
                    vol_instance.path,
                    conn=conn,
                    retry=self.retry_times,
                    AssocClass='CIM_ProtocolControllerForUnit',
                    ResultClass='FUJITSU_AffinityGroupController')
            except Exception:
                msg = (_('_find_affinity_group, '
                         'connector: %(connector)s, '
                         'AssocNames: CIM_ProtocolControllerForUnit, '
                         'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_find_affinity_group, '
                      'vol_instance.path: %(volume)s, '
                      'affinity_groups: %(aglist)s.',
                      {'volume': vol_instance.path,
                       'aglist': aglist})

        for ag in aglist:
            try:
                hostaglist = self._assoc_eternus(
                    ag,
                    conn=conn,
                    retry=self.retry_times,
                    AssocClass='CIM_AuthorizedTarget',
                    ResultClass='FUJITSU_AuthorizedPrivilege')
            except Exception:
                msg = (_('_find_affinity_group, '
                         'connector: %(connector)s, '
                         'Associators: CIM_AuthorizedTarget, '
                         'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            for hostag in hostaglist:
                for initiator in initiatorlist:
                    if initiator.lower() not in hostag['InstanceID'].lower():
                        continue

                    LOG.debug('_find_affinity_group, '
                              'AffinityGroup: %(ag)s.', {'ag': ag})
                    affinity_grouplist.append(ag)
                    break
                else:
                    continue
                break

        LOG.debug('_find_affinity_group, '
                  'initiators: %(initiator)s, '
                  'affinity_group: %(affinity_group)s.',
                  {'initiator': initiatorlist,
                   'affinity_group': affinity_grouplist})
        return affinity_grouplist

    @dx_utils.FJDXLockutils('connect', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _unmap_lun(self, volume, connector,
                   force=False, remote=False, timeout=None):
        """Unmap volume from host."""
        LOG.debug('_unmap_lun, volume id: %(vid)s, '
                  'connector: %(connector)s, force: %(frc)s, '
                  'remote: %(remote)s, timeout: %(timeout)s.',
                  {'vid': volume['id'], 'connector': connector,
                   'frc': force, 'remote': remote, 'timeout': timeout})

        volumename = self._get_volume_name(volume)
        if not remote:
            vol_instance = self._find_lun(volume)
            conn = self.conn
        else:
            vol_instance = self._find_replica_lun(volume)
            conn = self.remote_conn

        if not vol_instance:
            LOG.info('_unmap_lun, '
                     'volumename: %(volumename)s, '
                     'volume not found.',
                     {'volumename': volumename})
            return False

        volume_uid = vol_instance['Name']

        if not force:
            aglist = self._find_affinity_group(connector,
                                               vol_instance,
                                               conn=conn)
            if not aglist:
                LOG.info('_unmap_lun, '
                         'volumename: %(volumename)s, '
                         'volume is not mapped.',
                         {'volumename': volumename})
                return False
        else:
            try:
                aglist = self._assoc_eternus_names(
                    vol_instance.path,
                    conn=conn,
                    retry=self.retry_times,
                    AssocClass='CIM_ProtocolControllerForUnit',
                    ResultClass='FUJITSU_AffinityGroupController')
            except Exception:
                msg = (_('_unmap_lun,'
                         'vol_instance.path: %(volume)s, '
                         'AssociatorNames: CIM_ProtocolControllerForUnit, '
                         'cannot connect to ETERNUS.')
                       % {'volume': vol_instance.path})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_unmap_lun, '
                      'vol_instance.path: %(volume)s, '
                      'affinity_groups: %(aglist)s.',
                      {'volume': vol_instance.path,
                       'aglist': aglist})

        configservice = self._find_eternus_service(CONSTANTS.CTRL_CONF,
                                                   conn=conn)
        if not configservice:
            msg = (_('_unmap_lun, '
                     'vol_instance.path: %(volume)s, '
                     'volumename: %(volumename)s, '
                     'volume_uid: %(uid)s, '
                     'aglist: %(aglist)s, '
                     'Controller Configuration Service not found.')
                   % {'volume': vol_instance.path,
                      'volumename': [volumename],
                      'uid': [volume_uid],
                      'aglist': aglist})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for ag in aglist:
            LOG.debug('_unmap_lun, '
                      'volumename: %(volumename)s, '
                      'volume_uid: %(volume_uid)s, '
                      'AffinityGroup: %(ag)s.',
                      {'volumename': volumename,
                       'volume_uid': volume_uid,
                       'ag': ag})

            rc, errordesc, _x = self._exec_eternus_service(
                'HidePaths',
                configservice,
                conn=conn,
                timeout=self.req_sec,
                LUNames=[volume_uid],
                ProtocolControllers=[ag])

            LOG.debug('_unmap_lun, '
                      'Return code: %(rc)lu, '
                      'Error: %(errordesc)s.',
                      {'rc': rc,
                       'errordesc': errordesc})

            if rc == CONSTANTS.LUNAME_NOT_EXIST:
                LOG.debug('_unmap_lun, '
                          'volumename: %(volumename)s, '
                          'Invalid LUNames.',
                          {'volumename': volumename})
            elif rc != 0:
                msg = (_('_unmap_lun, '
                         'volumename: %(volumename)s, '
                         'volume_uid: %(volume_uid)s, '
                         'AffinityGroup: %(ag)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s.')
                       % {'volumename': volumename,
                          'volume_uid': volume_uid,
                          'ag': ag,
                          'rc': rc,
                          'errordesc': errordesc})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_unmap_lun, volumename: %s.', volumename)
        return True

    def _wait_for_job_complete(self, conn, job):
        """Given the job wait for it to complete."""
        self.retries = 0
        self.wait_for_job_called = False

        def _wait_for_job_complete():
            """Called at an interval until the job is finished."""
            if self._is_job_finished(conn, job):
                raise loopingcall.LoopingCallDone()
            if self.retries > CONSTANTS.JOB_RETRIES:
                LOG.error('_wait_for_job_complete, '
                          'failed after %(retries)d tries.',
                          {'retries': self.retries})
                raise loopingcall.LoopingCallDone()

            try:
                self.retries += 1
                if not self.wait_for_job_called:
                    if self._is_job_finished(conn, job):
                        self.wait_for_job_called = True
            except Exception as e:
                LOG.error('Exception: %s.', e)
                exceptionMessage = _('Issue encountered waiting for job.')
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(exceptionMessage)

        self.wait_for_job_called = False
        timer = loopingcall.FixedIntervalLoopingCall(_wait_for_job_complete)
        timer.start(interval=CONSTANTS.JOB_INTERVAL_SEC).wait()

        jobInstanceName = job['Job']
        jobinstance = self._exec_eternus_smis_GetInstance(
            jobInstanceName,
            conn=conn,
            timeout=self.req_sec,
            LocalOnly=False)

        rc = jobinstance['ErrorCode']

        LOG.debug('_wait_for_job_complete, rc: %s.', rc)
        return rc

    @staticmethod
    def _is_job_finished(conn, job):
        """Check if the job is finished."""
        jobInstanceName = job['Job']
        jobinstance = conn.GetInstance(jobInstanceName,
                                       LocalOnly=False)
        jobstate = jobinstance['JobState']
        LOG.debug('_is_job_finished, state: %(state)s.', {'state': jobstate})
        # From ValueMap of JobState in CIM_ConcreteJob
        # 2=New, 3=Starting, 4=Running, 32767=Queue Pending
        # ValueMap('2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13..32767,
        # 32768..65535'),
        # Values('New, Starting, Running, Suspended, Shutting Down,
        # Completed, Terminated, Killed, Exception, Service,
        # Query Pending, DMTF Reserved, Vendor Reserved')]
        # NOTE(deva): string matching based on
        #             http://ipmitool.cvs.sourceforge.net/
        #               viewvc/ipmitool/ipmitool/lib/ipmi_chassis.c

        if jobstate in [2, 3, 4]:
            job_finished = False
        else:
            job_finished = True

        LOG.debug('_is_job_finished, finish: %s.', job_finished)
        return job_finished

    @staticmethod
    def _pywbem_uint(num, datatype):
        try:
            if datatype == '8':
                result = pywbem.Uint8(num)
            elif datatype == '16':
                result = pywbem.Uint16(num)
            elif datatype == '32':
                result = pywbem.Uint32(num)
            elif datatype == '64':
                result = pywbem.Uint64(num)
        except NameError:
            result = num

        return result

    @dx_utils.FJDXTimeout()
    def manage_existing(self, manage_type, volume, existing_ref,
                        timeout=None):
        """Bring an existing storage object under Cinder management.

        existing_ref can contain source-id or source-name.
        source-id: lun uuid
        source-name: lun name
        """
        if self.failover:
            msg = (_('manage_existing, '
                     'volume id: %(vid)s, '
                     'can not manage volume on failed over backend! ')
                   % {'vid': volume['id']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        vol_instance = self._get_manage_volume_instance(manage_type, volume,
                                                        existing_ref)

        element_path = {
            'classname': vol_instance.classname,
            'vol_name': vol_instance['ElementName'],
            'keybindings': {
                'SystemName': vol_instance['SystemName'],
                'DeviceID': vol_instance['DeviceID'],
            }
        }

        # Get eternus model name.
        try:
            systemnamelist = (
                self._enum_eternus_instances(
                    'FUJITSU_StorageProduct', retry=self.retry_times))
        except Exception:
            msg = (_('manage_existing, '
                     'volume: %s, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.')
                   % volume)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        d_metadata = dx_utils.get_metadata(volume)
        volume_no = self._get_volume_no(vol_instance)
        if manage_type == 'volume':
            metadata = {
                'FJ_Backend': systemnamelist[0]['IdentifyingNumber'],
                'FJ_Volume_Name': vol_instance['ElementName'],
                'FJ_Volume_No': volume_no,
                'FJ_Pool_Name': vol_instance['poolname'],
                'FJ_Pool_Type': vol_instance['pooltype'],
            }
            d_metadata.update(metadata)
        elif manage_type == 'snapshot':
            metadata = {
                'FJ_SDV_Name': vol_instance['ElementName'],
                'FJ_SDV_No': volume_no,
                'FJ_Pool_Name': vol_instance['poolname'],
                'FJ_Pool_Type': vol_instance['pooltype'],
            }
            d_metadata.update(metadata)

        model_update = {
            'provider_location': six.text_type(element_path),
            'metadata': d_metadata,
        }
        volume['provider_location'] = model_update['provider_location']
        self._set_qos(volume, timeout=self.req_sec)

        return model_update

    @dx_utils.FJDXTimeout()
    def manage_existing_get_size(self, manage_type,
                                 volume, existing_ref, timeout=None):
        """Return size of volume to be managed by manage_existing.

        When calculating the size, round up to the next GB.
        """
        vol_instance = self._get_manage_volume_instance(
            manage_type, volume, existing_ref)

        size_byte = (vol_instance['BlockSize'] *
                     vol_instance['NumberOfBlocks'])
        vol_size = (size_byte + units.Gi - 1) / units.Gi

        LOG.debug('manage_existing_get_size, size: %s.', vol_size)
        return vol_size

    def _get_manage_volume_instance(self, manage_type, volume, existing_ref):
        """Get manage volume instance from existing_ref.

        The volume must be a standard volume or TPV,
        and in EternusPool of configuration file.
        """
        LOG.debug('_get_manage_volume_instance, '
                  'manage_type: %(manage_type)s. ',
                  {'manage_type': manage_type})

        self.conn = self._get_eternus_connection()
        vol_instance = self._find_lun_with_listup(self.conn, **existing_ref)

        if not vol_instance:
            msg = (_('_get_manage_volume_instance, '
                     'source name or source id: %s, '
                     'volume not found.')
                   % existing_ref)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_get_manage_volume_instance, '
                  'volume name: %(volumename)s, '
                  'Purpose: %(Purpose)s, '
                  'volume instance: %(vol_instance)s.',
                  {'volumename': vol_instance['ElementName'],
                   'Purpose': vol_instance['Purpose'],
                   'vol_instance': vol_instance.path})

        # Decide manageable.
        manageable_list = self._decide_manageable(manage_type,
                                                  [vol_instance],
                                                  volume_info=volume)
        if not manageable_list:
            msg = (_('_get_manage_volume_instance, '
                     'Volume type is incorrect or '
                     'Volume does not exist copy session.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not manageable_list[0]['safe_to_manage']:
            msg = (_('_get_manage_volume_instance, '
                     'reason: %s. ')
                   % manageable_list[0]['reason_not_safe'][:-1])
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        poolname, target_pool = self._find_pool_from_volume(vol_instance,
                                                            manage_type)
        # Set poolname and pooltype.
        vol_instance['poolname'] = poolname
        if 'RSP' in target_pool['InstanceID']:
            vol_instance['pooltype'] = \
                CONSTANTS.POOL_TYPE_dic[CONSTANTS.RAIDGROUP]
        else:
            vol_instance['pooltype'] = \
                CONSTANTS.POOL_TYPE_dic[CONSTANTS.TPPOOL]
        return vol_instance

    def _find_lun_with_listup(self, conn=None, **kwargs):
        """Find lun instance with source name or source id on ETERNUS."""
        LOG.debug('_find_lun_with_listup start.')

        volumeinstance = None
        src_id = kwargs.get('source-id', None)
        src_name = kwargs.get('source-name', None)

        if not src_id and not src_name:
            msg = (_('_find_lun_with_listup, '
                     'source-name or source-id: %s, '
                     'Must specify source-name or source-id.')
                   % kwargs)
            LOG.error(msg)
            raise exception.ManageExistingInvalidReference(data=msg)

        if src_id and src_name:
            msg = (_('_find_lun_with_listup, '
                     'source-name or source-id: %s, '
                     'Must only specify source-name or source-id.')
                   % kwargs)
            LOG.error(msg)
            raise exception.ManageExistingInvalidReference(data=msg)

        if src_id and not src_id.isdigit():
            msg = (_('_find_lun_with_listup, '
                     'the specified source-id(%s) must be a decimal number.')
                   % src_id)
            LOG.error(msg)
            raise exception.ManageExistingInvalidReference(data=msg)

        # Get volume instance by volumename or volumeno on ETERNUS.
        try:
            propertylist = [
                'SystemName',
                'DeviceID',
                'ElementName',
                'Purpose',
                'BlockSize',
                'NumberOfBlocks',
                'Name',
                'OtherUsageDescription',
                'IsCompressed',
                'IsDeduplicated'
            ]
            vollist = self._enum_eternus_instances(
                'FUJITSU_StorageVolume',
                conn=conn,
                retry=self.retry_times,
                PropertyList=propertylist)
        except Exception:
            msg = (_('_find_lun_with_listup, '
                     'source-name or source-id: %s, '
                     'EnumerateVolumeInstance.')
                   % kwargs)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for vol_instance in vollist:
            if src_id:
                volume_no = self._get_volume_no(vol_instance)
                try:
                    # Skip hidden tppv volumes.
                    if int(src_id) == int(volume_no, 16):
                        volumeinstance = vol_instance
                        break
                except ValueError:
                    continue
            if src_name:
                if vol_instance['ElementName'] == src_name:
                    volumeinstance = vol_instance
                    break
        else:
            LOG.debug('_find_lun_with_listup, '
                      'source-name or source-id: %s, '
                      'volume not found on ETERNUS.', kwargs)

        LOG.debug('_find_lun_with_listup end, '
                  'volume instance: %s.', volumeinstance)
        return volumeinstance

    def _find_pool_from_volume(self, vol_instance,
                               manage_type='volume', remote=False):
        """Find Instance or InstanceName of pool by volume instance."""
        LOG.debug('_find_pool_from_volume, remote: %(remote)s.',
                  {'remote': remote})
        poolname = None
        target_pool = None
        filename = None

        if not remote:
            conn = self.conn
        else:
            conn = self.remote_conn
            filename = self.remote_eternus_config

        # Get poolname of volume on Eternus.
        try:
            pools = self._assoc_eternus(
                vol_instance.path,
                conn=conn,
                retry=self.retry_times,
                AssocClass='FUJITSU_AllocatedFromStoragePool',
                ResultClass='CIM_StoragePool')
        except Exception:
            msg = (_('_find_pool_from_volume, '
                     'vol_instance: %s, '
                     'Associators: FUJITSU_AllocatedFromStoragePool, '
                     'cannot connect to ETERNUS.')
                   % vol_instance.path)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not pools:
            msg = (_('_find_pool_from_volume, '
                     'vol_instance: %s, '
                     'pool not found.')
                   % vol_instance.path)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Get poolname from driver configuration file.
        if manage_type == 'volume':
            cfgpool_list = list(self._get_drvcfg('EternusPool',
                                                 filename=filename,
                                                 multiple=True))
        elif manage_type == 'snapshot':
            cfgpool_list = list(self._get_drvcfg('EternusSnapPool',
                                                 filename=filename,
                                                 multiple=True))
        LOG.debug('_find_pool_from_volume, cfgpool_list: %(cfgpool_list)s.',
                  {'cfgpool_list': cfgpool_list})
        for pool in pools:
            if pool['ElementName'] in cfgpool_list:
                poolname = pool['ElementName']
                target_pool = pool.path
                break

        if not target_pool:
            msg = (_('_find_pool_from_volume, '
                     'vol_instance: %s, '
                     'the pool of volume not in driver configuration file.')
                   % vol_instance.path)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_pool_from_volume, poolname: %(poolname)s, '
                  'target_pool: %(target_pool)s.',
                  {'poolname': poolname, 'target_pool': target_pool})
        return poolname, target_pool

    def update_migrated_volume(self, ctxt, volume, new_volume):
        """Update migrated volume."""
        LOG.debug('update_migrated_volume, '
                  'source volume id: %(s_id)s, '
                  'target volume id: %(t_id)s.',
                  {'s_id': volume['id'], 't_id': new_volume['id']})

        dst_metadata = dx_utils.get_metadata(new_volume)
        src_metadata = dx_utils.get_metadata(volume)

        LOG.debug('source: (%(src_meta)s)(%(src_loc)s), '
                  'target: (%(dst_meta)s)(%(dst_loc)s).',
                  {'src_meta': src_metadata,
                   'src_loc': volume['provider_location'],
                   'dst_meta': dst_metadata,
                   'dst_loc': new_volume['provider_location']})

        if volume['provider_location']:
            src_location = eval(volume['provider_location'])

            if 'vol_name' not in src_location:
                src_name = self._get_volume_name(volume)

                if src_location['keybindings']['CreationClassName']:
                    del src_location['keybindings']['CreationClassName']

                if src_location['keybindings']['SystemCreationClassName']:
                    del src_location['keybindings']['SystemCreationClassName']

                src_location['vol_name'] = src_name

                db.volume_update(
                    ctxt, volume['id'],
                    {'provider_location': six.text_type(src_location)})

            db.volume_update(
                ctxt, new_volume['id'],
                {'provider_location': six.text_type(src_location)})

    def _prepare_replica(self):
        """Get replication target device informations from cinder.conf."""
        remote_host = None
        remote_eternus_config = None
        remote_backend = None
        local_boxid = None
        tfog_name = None
        remote_pool = None

        # Get replica_device in cinder.conf.
        replica_dev_list = self.configuration.replication_device

        if replica_dev_list:
            # Currently we only support one replication device.
            replica_device = replica_dev_list[0]
            LOG.debug('_prepare_replica, replica_device: %s.', replica_device)

            remote_eternus_config = replica_device.get(
                'cinder_eternus_config_file', None)

            if not remote_eternus_config:
                msg = _('_prepare_replica, '
                        'No cinder_eternus_config_file was '
                        'found in replica_device.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            local_boxid = replica_device.get(
                'local_eternus_boxid', None)

            if not local_boxid:
                msg = _('_prepare_replica, '
                        'No local_eternus_boxid was '
                        'found in replica_device.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            managed_backend = replica_device.get(
                'backend_id', None)

            if not managed_backend:
                msg = _('_prepare_replica, '
                        'No managed_backend_name was '
                        'found in replica_device.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                try:
                    # Ensure managed_backend_name must be the form of
                    # "host@backend#backend_name".
                    remote_host, temp = self._get_key_pair(
                        managed_backend, '@')
                    remote_backend, remote_pool = (
                        self._get_key_pair(temp, '#'))
                except Exception:
                    msg = _('_prepare_replica, '
                            'The form of managed_backend_name is not '
                            '"host@backend#backend_name".')
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

            tfog_name = replica_device.get(
                'tfog_name', None)
            if not tfog_name:
                msg = _('_prepare_replica, '
                        'No tfo_group name was '
                        'found in replica_device.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_prepare_replica, '
                      'remote host: %(remote_host)s, '
                      'remote backend: %(remote_backend)s, '
                      'local boxid: %(local_boxid)s, '
                      'tfog_name: %(tfog_name)s, '
                      'remote eternus config: %(remote_eternus_config)s.',
                      {'remote_host': remote_host,
                       'remote_backend': remote_backend,
                       'local_boxid': local_boxid,
                       'tfog_name': tfog_name,
                       'remote_eternus_config': remote_eternus_config})
        return (remote_host, remote_backend, remote_eternus_config,
                local_boxid, tfog_name, remote_pool)

    def _create_replica_volume(self, volume, src_metadata):
        """Create replica volume."""
        LOG.debug('_create_replica_volume, '
                  'src volume id: %(v_id)s, '
                  'src metadata: %(meta)s.',
                  {'v_id': volume['id'],
                   'meta': src_metadata})

        repl_data = None
        error_msg = None

        try:
            repl_element_path, _x, repl_data, repl_status = (
                self._create_volume(volume, remote=True, timeout=self.req_sec))
        except Exception:
            repl_status = 'error'
            error_msg = 'Create remote volume failed.'
            return None, repl_data, repl_status, error_msg

        replication_driver_data = (
            {'provider_location': six.text_type(repl_element_path)})

        return replication_driver_data, repl_data, repl_status, error_msg

    def _exec_ccm_script(self, command,
                         source=None,
                         source_volume=None,
                         target=None,
                         copy_type=None):
        """Execute CCM script."""
        LOG.debug('_exec_ccm_script, '
                  'command: %(cmd)s, src metadata: %(src_meta)s, '
                  'tgt metadata: %(tgt_meta)s, copy type: %(copy_type)s.',
                  {'cmd': command, 'src_meta': source,
                   'tgt_meta': target, 'copy_type': copy_type})

        s_option = ""
        d_option = ""
        t_option = ""

        if source:
            if 'FJ_Volume_No' in source:
                s_olu_no = source['FJ_Volume_No']
                if s_olu_no == '0x':
                    msg = (_('_exec_ccm_script, '
                             'command: %(command)s, '
                             'src metadata: %(src_meta)s, '
                             'not supported volume. It may be created on '
                             'DX S2 with old osvd (Prior to OSVD1.3.0).')
                           % {'command': command, 'src_meta': source})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
            else:
                vol_instance = self._find_lun(
                    source_volume, use_service_name=True)
                s_olu_no = self._get_volume_no(vol_instance)

            s_serial_no = source['FJ_Backend'][-10:]
            s_option = " -s \"%s/%s\"" % (s_serial_no, s_olu_no)
            command += s_option

        if target:
            d_olu_no = target['FJ_Volume_No']
            d_serial_no = target['FJ_Backend'][-10:]
            d_option = " -d \"%s/%s\"" % (d_serial_no, d_olu_no)
            command += d_option

        if s_option and d_option:
            if s_option[3:] == d_option[3:]:
                msg = (_('_exec_ccm_script, '
                         'command: %(command)s, '
                         'source_vol: %(s_option)s and '
                         'target_vol: %(d_option)s are the same one.')
                       % {'command': command, 's_option': s_option[3:],
                          'd_option': d_option[3:]})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        if copy_type:
            t_option = " -t %s" % ((copy_type[0]).lower() + copy_type[1:])
            command += t_option

        try:
            _x, _err = utils.execute('acrec', command, run_as_root=True)
            LOG.debug('_exec_ccm_script, '
                      'command: %(command)s, '
                      's_option: %(s_option)s, '
                      'd_option: %(d_option)s, '
                      't_option: %(t_option)s.',
                      {'command': command,
                       's_option': s_option,
                       'd_option': d_option,
                       't_option': t_option})

        except Exception as ex:
            LOG.error(ex)
            msg = (_('_exec_ccm_script, '
                     'command: %(command)s, '
                     's_option: %(s_option)s, '
                     'd_option: %(d_option)s, '
                     't_option: %(t_option)s.')
                   % {'command': command, 's_option': s_option,
                      'd_option': d_option, 't_option': t_option})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(ex)

    @staticmethod
    def _get_key_pair(info, delimiter):
        key_pair = info.split(delimiter, 1)
        if len(key_pair) != 2:
            msg = (_('Cannot use %(sep)s to split %(info)s, '
                     'please check the format.')
                   % {'sep': delimiter, 'info': info})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return key_pair[0], key_pair[1]

    def _find_replica_lun(self, volume):
        """Find replica volume by volume name."""
        LOG.debug('_find_replica_lun, '
                  'src volume id: %s.', volume['id'])

        volume_no = None
        volume_dic = {}

        if volume['replication_driver_data']:
            replica_data = eval(volume['replication_driver_data'])
        else:
            msg = _('_find_replica_lun, '
                    'replication_driver_data is None.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        try:
            replica_provider = eval(replica_data['provider_location'])
            volume_name = replica_provider['vol_name']
            volume_dic = {
                'source-name': six.text_type(volume_name)
            }
        except KeyError:
            vol_data = eval(volume['provider_location'])
            volume_name = vol_data['vol_name']
            volume_dic = {
                'source-name': six.text_type(volume_name)
            }

        try:
            volume_instance = self._find_lun_with_listup(self.remote_conn,
                                                         **volume_dic)
        except Exception:
            msg = (_('_find_replica_lun, '
                     'Cannot find replica volume, '
                     'volume no: %s.')
                   % volume_no)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_replica_lun, '
                  'volume instance: %s.', volume_instance)
        return volume_instance

    def _get_eternus_model(self):
        """Get ENTERNUS model."""
        self.conn = self._get_eternus_connection()
        ret = CONSTANTS.DX_S3
        try:
            systemnamelist = self._enum_eternus_instances(
                'FUJITSU_StorageProduct',
                conn=self.conn,
                retry=self.retry_times)
        except Exception:
            msg = _('_get_eternus_model, EnumerateInstances, '
                    'cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        systemname = systemnamelist[0]['IdentifyingNumber']

        LOG.debug('_get_eternus_model, '
                  'systemname: %(systemname)s, '
                  'storage is DX S%(model)s.',
                  {'systemname': systemname,
                   'model': systemname[4]})

        if six.text_type(systemname[4]) == '2':
            ret = CONSTANTS.DX_S2

        return ret

    def _get_volume_no(self, vol):
        """Get volume no(return a hex string)."""
        if self.model_name == CONSTANTS.DX_S2:
            volume_no = "0x%04X" % int(vol['DeviceID'][-5:])
        else:
            volume_no = "0x" + vol['DeviceID'][24:28]

        LOG.debug('_get_volume_no: %s.', volume_no)
        return volume_no

    @staticmethod
    def _is_replicated(volume):
        """check volume whether is replication volume or not."""
        repl_value = dx_utils.get_extra_specs(volume,
                                              CONSTANTS.FJ_VOL_REPL_KEY)
        return repl_value == CONSTANTS.FJ_VOL_REPL_VALUE

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_InvokeMethod(self, classname, instanceNameList,
                                        conn=None,
                                        timeout=None,
                                        **param_dict):
        rc, retdata = conn.InvokeMethod(
            classname,
            instanceNameList,
            **param_dict)
        return rc, retdata

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_EnumerateInstances(self, classname,
                                              conn=None,
                                              timeout=None,
                                              **param_dict):
        ret = conn.EnumerateInstances(classname, **param_dict)
        return ret

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_EnumerateInstanceNames(self, classname,
                                                  conn=None,
                                                  timeout=None):
        ret = conn.EnumerateInstanceNames(classname)
        return ret

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_GetInstance(self, classname,
                                       conn=None,
                                       timeout=None,
                                       **param_dict):
        ret = conn.GetInstance(classname, **param_dict)
        return ret

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_Associators(self, classname,
                                       conn=None,
                                       timeout=None,
                                       **param_dict):
        ret = conn.Associators(classname, **param_dict)
        return ret

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_AssociatorNames(self, classname,
                                           conn=None,
                                           timeout=None,
                                           **param_dict):
        ret = conn.AssociatorNames(classname, **param_dict)
        return ret

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_ReferenceNames(self, classname,
                                          conn=None,
                                          timeout=None,
                                          **param_dict):
        ret = conn.ReferenceNames(classname, **param_dict)
        return ret

    @dx_utils.FJDXTimeout()
    def _exec_eternus_smis_CIMInstanceName(self, classname,
                                           namespace=None,
                                           keybindings=None,
                                           timeout=None):
        instancename = pywbem.CIMInstanceName(
            classname,
            namespace=namespace,
            keybindings=keybindings)
        return instancename

    def _check_user(self):
        """Check whether user's role is accessible to ETERNUS and Software."""
        ret = True
        rc, errordesc, job = self._exec_eternus_cli(
            'check_user_role',
            retry=self.retry_times,
            timeout=self.req_sec)
        if rc != 0:
            msg = (_('_check_user, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(job)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if job != 'Software':
            msg = (_('_check_user, '
                     'Specified user(%(user)s) does not have '
                     'Software role: %(role)s.')
                   % {'user': self._get_drvcfg('EternusUser'),
                      'role': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return ret

    @dx_utils.FJDXLockutils('CLI-exec', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _exec_eternus_cli(self, command,
                          remote=False, retry=CONSTANTS.TIMES_MIN,
                          retry_interval=CONSTANTS.RETRY_INTERVAL,
                          retry_code=[32787], filename=None, timeout=None,
                          **param_dict):
        """Execute ETERNUS CLI."""
        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'filename: %(f)s, '
                  'timeout: %(t)s, '
                  'parameters: %(b)s, '
                  'remote: %(remote)s.',
                  {'a': command,
                   'f': filename,
                   't': timeout,
                   'b': param_dict,
                   'remote': remote})

        out = None
        _err = None
        result = None
        rc = None
        retdata = None
        errordesc = None
        filename = self.remote_filename if remote else self.filename
        storage_ip = self._get_drvcfg('EternusIP', filename)
        if not self.fjdxcli.get(filename):
            user = self._get_drvcfg('EternusUser', filename)
            if self.passwordless:
                self.fjdxcli[filename] = (
                    eternus_dx_cli.FJDXCLI(user,
                                           storage_ip,
                                           keyfile=self.private_key_path))
            else:
                password = self._get_drvcfg('EternusPassword', filename)
                self.fjdxcli[filename] = (
                    eternus_dx_cli.FJDXCLI(user, storage_ip,
                                           password=password))

        for retry_num in range(retry):
            # Execute ETERNUS CLI and get return value.
            try:
                out = self._exec_eternus_cli_with_timeout(
                    filename,
                    command,
                    timeout=self.req_sec,
                    **param_dict)
                out_dict = out
                result = out_dict.get('result')
                rc_str = out_dict.get('rc')
                retdata = out_dict.get('message')
            except Exception as ex:
                msg = (_('_exec_eternus_cli, '
                         'stdout: %(out)s, '
                         'stderr: %(err)s, '
                         'unexpected error: %(ex)s.')
                       % {'out': out,
                          'err': _err,
                          'ex': ex})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Check ssh result.
            if result == 255:
                LOG.info('_exec_eternus_cli, retry, '
                         'command: %(command)s, '
                         'option: %(option)s, '
                         'ip: %(ip)s, '
                         'SSH Result: %(result)s, '
                         'retdata: %(retdata)s, '
                         'TryNum: %(rn)s.',
                         {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata,
                          'rn': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
            elif result != 0:
                msg = (_('_exec_eternus_cli, '
                         'unexpected error, '
                         'command: %(command)s, '
                         'option: %(option)s, '
                         'ip: %(ip)s, '
                         'result: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Check CLI return code.
            if rc_str.isdigit():
                # SMI-S style return code.
                rc = int(rc_str)

                try:
                    errordesc = CONSTANTS.RETCODE_dic[str(rc)]
                except Exception:
                    errordesc = 'Undefined Error!!'

                if rc in retry_code:
                    LOG.info('_exec_eternus_cli, retry, '
                             'ip: %(ip)s, '
                             'RetryCode: %(rc)s, '
                             'TryNum: %(rn)s.',
                             {'ip': storage_ip,
                              'rc': rc,
                              'rn': (retry_num + 1)})
                    time.sleep(retry_interval)
                    continue
                if rc == 4:
                    if ('Authentication failed' in retdata and
                            retry_num + 1 < retry):
                        LOG.warning('_exec_eternus_cli, retry, ip: %(ip)s, '
                                    'Message: %(message)s, '
                                    'TryNum: %(rn)s.',
                                    {'ip': storage_ip,
                                     'message': retdata,
                                     'rn': (retry_num + 1)})
                        time.sleep(1)
                        continue

                break
            else:
                # CLI style return code.
                LOG.warning('_exec_eternus_cli, '
                            'WARNING!! '
                            'ip: %(ip)s, '
                            'ReturnCode: %(rc_str)s, '
                            'ReturnData: %(retdata)s.',
                            {'ip': storage_ip,
                             'rc_str': rc_str,
                             'retdata': retdata})

                errordesc = rc_str
                rc = 4  # Failed.
                break
        else:
            if 0 < result:
                msg = (_('_exec_eternus_cli, '
                         'cannot connect to ETERNUS. '
                         'SSH Result: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'result': result,
                          'retdata': retdata})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                LOG.warning('_exec_eternus_cli, Retry was exceeded.')

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'parameters: %(b)s, '
                  'ip: %(ip)s, '
                  'Return code: %(rc)s, '
                  'Error: %(errordesc)s.',
                  {'a': command,
                   'b': param_dict,
                   'ip': storage_ip,
                   'rc': rc,
                   'errordesc': errordesc})
        return ret

    @dx_utils.FJDXTimeout()
    def _exec_eternus_cli_with_timeout(self, filename, command,
                                       timeout=None, **param_dict):
        if param_dict:
            out = self.fjdxcli[filename].done(command, **param_dict)
        else:
            out = self.fjdxcli[filename].done(command)
        return out

    def _filter_available_tfo_port(self):
        """filter available tfo port."""
        LOG.debug('_filter_available_tfo_port, start! '
                  'tfog_name: %(tfog_name)s.',
                  {'tfog_name': self.tfog_name})
        own_ports_list = []
        pair_ports_list = []

        remote_boxid = self._show_boxid(remote=True)
        tfo_group_list = self._show_tfo_groups()
        for tfo_group in tfo_group_list:
            if (tfo_group['type'] == '01' and  # Primary.
                    tfo_group['status'] == '01' and  # Active.
                    tfo_group['pair_boxid'] == remote_boxid and
                    tfo_group['name'] == self.tfog_name):
                tfog_port_list = tfo_group['tfog_port_list']
                for tfog_port in tfog_port_list:
                    if tfog_port['port_prtcl'] == self.protocol:
                        own_ports_list.append(tfog_port['own_port'])
                        pair_ports_list.append(tfog_port['pair_port'])
                    else:
                        msg = (_("The port_prtcl of tfo group %(tfog_no)s is "
                                 "%(port_prtcl)s.")
                               % {'tfog_no': tfo_group['no'],
                                  'port_prtcl': tfog_port['port_prtcl']})
                        LOG.warning(msg)

        LOG.debug('_filter_available_tfo_port, '
                  'own tfo ports: %(own_ports_list)s, '
                  'pair tfo ports: %(pair_ports_list)s.',
                  {'own_ports_list': own_ports_list,
                   'pair_ports_list': pair_ports_list})
        return own_ports_list, pair_ports_list

    def _show_all_tfo_port(self):
        """filter all tfo own ports."""
        LOG.debug('_show_all_tfo_port, enter method.')

        own_ports_list = []
        try:
            tfo_group_list = self._show_tfo_groups()
            for tfo_group in tfo_group_list:
                tfog_port_list = tfo_group['tfog_port_list']
                for tfog_port in tfog_port_list:
                    own_ports_list.append(tfog_port['own_port'])
        except Exception as ex:
            msg = (_("_show_all_tfo_port, _show_tfo_groups failed."
                     "Error information: %(ex)s.")
                   % {'ex': ex})
            LOG.warning(msg)

        LOG.debug('_show_all_tfo_port, '
                  'own tfo ports: %(own_ports_list)s.',
                  {'own_ports_list': own_ports_list})
        return own_ports_list

    def _show_boxid(self, remote=False):
        """Show boxid."""
        LOG.debug('_show_boxid, remote: %s.', remote)

        rc, errordesc, clidata = self._exec_eternus_cli(
            'show_boxid',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec)

        if rc != 0:
            msg = (_('_show_boxid, '
                     'remote: %(remote)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(job)s.')
                   % {'remote': remote,
                      'rc': rc,
                      'errordesc': errordesc,
                      'job': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_show_boxid, boxid: %s, end!', clidata)
        return clidata

    def _show_tfo_groups(self, tfog_number=None, remote=False):
        """Show tfo groups."""
        LOG.debug('_show_tfo_groups, '
                  'tfo group number: %(number)s '
                  'remote: %(remote)s.',
                  {'number': tfog_number, 'remote': remote})

        local_ce_support = self.local_ce_support
        remote_ce_support = self.remote_ce_support
        if remote:
            local_ce_support = self.remote_ce_support
            remote_ce_support = self.local_ce_support
        if tfog_number:
            param_dict = {
                'tfog-number': tfog_number,
                'local_ce_support': local_ce_support,
                'remote_ce_support': remote_ce_support
            }
        else:
            param_dict = {
                'local_ce_support': local_ce_support,
                'remote_ce_support': remote_ce_support
            }
        rc, errordesc, clidata = self._exec_eternus_cli(
            'show_tfo_groups',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **param_dict)

        if rc != 0:
            msg = (_('_show_tfo_groups, '
                     'tfog number: %(tfog_number)s, '
                     'remote: %(remote)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(job)s.')
                   % {'tfog_number': tfog_number,
                      'remote': remote,
                      'rc': rc,
                      'errordesc': errordesc,
                      'job': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug("_show_tfo_groups, clidata=%s.", clidata)
        return clidata

    @staticmethod
    def _get_host_ag_from_cfg_file(conf_file_name, tfo_ports):
        if not conf_file_name:
            msg = _('_get_host_ag_from_cfg_file, '
                    'please configure fujitsu_port_host_mapping_file '
                    'in cinder.conf!')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not os.path.exists(conf_file_name):
            msg = (_('_get_host_ag_from_cfg_file, '
                     'cfg file(%(path)s) does not exist.')
                   % {'path': conf_file_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Every line in cfg_file likes follow:
        # port1,port2,...=iqn=lungroup
        # port: xyz,wxyz
        # xyz: Fixed 3 digits, CM number + CA number + Port number
        # wxyz: Fixed 4 digits,
        #       CE number(hex) + CM number + CA number + Port number
        # iqn: iqn.domaindate.reverse.domain.name:optional name
        # lungroup: lungroup name
        host_ag_dict = {}
        pattern = r"^(\s*[0-9A-Fa-f]{3,4}\s*,)*\s*[0-9A-Fa-f]{3,4}\s*=\s*\w"
        with open(conf_file_name) as f:
            line = f.readline()
            while line:
                result = re.match(pattern, line)
                if result:
                    infos = [x.strip() for x in line.split('=', 2)]
                    if (len(infos) != 3 or
                            not (infos[0] and infos[1] and infos[2])):
                        msg = _('_get_host_ag_from_cfg_file, '
                                'pattern of fujitsu_port_host_mapping_file '
                                'should be port1,port2,...=iqn=lungroup.')
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

                    port_list = [x.strip() for x in infos[0].split(',')]
                    if len(tfo_ports) > 0:
                        for port in port_list:
                            if port in tfo_ports:
                                msg = (_('_get_host_ag_from_cfg_file, '
                                         'the port(%(port)s) '
                                         'of fujitsu_port_host_mapping_file '
                                         'should not be tfo port.'),
                                       {'port': port})
                                LOG.error(msg)
                                raise exception.VolumeBackendAPIException(
                                    data=msg)
                    host_ag_dict[infos[1].lower()] = {
                        'portlist': port_list,
                        'lungroup': infos[2]
                    }
                line = f.readline()

        if not host_ag_dict:
            host_ag_dict = None
            LOG.warning('_get_host_ag_from_cfg_file, '
                        'no available map info in map file(%(file)s), '
                        'please check the map file!',
                        {'file': conf_file_name})

        LOG.debug('_get_host_ag_from_cfg_file, '
                  'host_ag_dict: %s.', host_ag_dict)
        return host_ag_dict

    def _get_port_info(self, remote=False):
        """Get iscsi/fc port iqn and ip information from ETERNUS."""

        emsg = None
        clidata = None
        self.ce_support = (
            self.local_ce_support if not remote else self.remote_ce_support)
        param_dict = {
            'ce_support': self.ce_support
        }
        if self.protocol == 'iSCSI':
            rc, emsg, clidata = self._exec_eternus_cli(
                'show_iscsi_parameters',
                remote=remote,
                retry=self.retry_times,
                timeout=self.req_sec,
                **param_dict)
        elif self.protocol == 'FC':
            if self.model_name == CONSTANTS.DX_S2:
                return self._get_fc_port_info_for_DXS2()
            rc, emsg, clidata = self._exec_eternus_cli(
                'show_fc_parameters',
                remote=remote,
                retry=self.retry_times,
                timeout=self.req_sec,
                **param_dict)

        if rc != 0:
            msg = (_('_get_port_info, '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        port_info_dict = {}
        for data in clidata:
            if self.protocol == 'iSCSI':
                port_info_dict[data['port str']] = (
                    {'port_iqn': data['port iqn'],
                     'port_ip': data['port ip'],
                     'port_num': data['port num']})
            elif self.protocol == 'FC':
                port_info_dict[data['port str']] = (
                    {'port_wwpn': data['port wwpn'],
                     'port_num': data['port num']})
        LOG.debug('_get_port_info, port_info_dict: %s.', port_info_dict)

        return port_info_dict

    def _get_fc_port_info_for_DXS2(self, remote=False):
        filename = self.remote_filename if remote else self.filename
        conn = self._get_eternus_connection(filename)
        try:
            tgtportlist = self._enum_eternus_instances(
                'FUJITSU_SCSIProtocolEndpoint',
                conn=conn,
                retry=self.retry_times)
        except Exception:
            msg = _('_get_fc_port_info_for_DXS2, '
                    'EnumerateInstances, '
                    'cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        port_info_dict = {}
        for tgtport in tgtportlist:
            # Check protocol of tgtport.
            if tgtport['ConnectionType'] != 2:  # 2 ("Fibre Channel")
                continue

            # Check: if port is for remote copy, continue.
            if tgtport['RAMode'] & 0x7B != 0x00:
                continue

            # Check: if port is for StorageCluster, continue.
            if 'SCGroupNo' in tgtport:
                continue

            wwpn = tgtport['Name']
            port_str = (str(tgtport['CMSlotNumber']) +
                        str(tgtport['CASlotNumber']) +
                        str(tgtport['PortNo']))
            port_info_dict[port_str] = {
                'port_wwpn': wwpn
            }

        LOG.debug('_get_fc_port_info_for_DXS2, FC port_info_dict: %s.',
                  port_info_dict)
        return port_info_dict

    def _map_replication_volume(self, volume, connector):
        """Map replication volume. """
        LOG.debug('_map_replication_volume, '
                  'volume id: %(id)s.',
                  {'id': volume['id']})

        own_ports, pair_ports = self._filter_available_tfo_port()
        if not own_ports or not pair_ports:
            msg = (_('_map_replication_volume, '
                     'no available tfo %(protocol)s '
                     'port(s) for volume to attach. '
                     'own port(s): %(own_ports)s, '
                     'pair port(s): %(pair_ports)s.')
                   % {'protocol': self.protocol,
                      'own_ports': own_ports,
                      'pair_ports': pair_ports})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        host_tfo_ports_dict = self._find_tfo_port_pairs(own_ports,
                                                        pair_ports,
                                                        connector,
                                                        volume)
        lunnumber = None
        try:
            mapdata, ports, lunnumber = self._map_lun_without_map_file(
                volume,
                connector,
                host_tfo_ports_dict,
                remote=False,
                replica_volume=True,
                timeout=self.req_sec)
        except Exception:
            with excutils.save_and_reraise_exception():
                self._unmap_lun_without_map_file(
                    volume, connector, remote=False,
                    replica_volume=True, timeout=self.req_sec)

        LOG.debug('_map_replication_volume, '
                  'lun number: %(lunnumber)s.',
                  {'lunnumber': mapdata['target_lun']})
        try:
            self._map_lun_without_map_file(
                volume, connector, host_tfo_ports_dict,
                remote=True,
                replica_volume=True,
                lun_no=lunnumber,
                timeout=self.req_sec)
        except Exception:
            with excutils.save_and_reraise_exception():
                self._unmap_lun_without_map_file(volume, connector,
                                                 remote=True,
                                                 replica_volume=True,
                                                 timeout=self.req_sec)
                self._unmap_lun_without_map_file(volume, connector,
                                                 remote=False,
                                                 replica_volume=True,
                                                 timeout=self.req_sec)

        mapdata['target_discoverd'] = True
        mapdata['volume_id'] = volume['id']

        LOG.debug('_map_replication_volume, '
                  'mapdata: %(mapdata)s, '
                  'ports: %(ports)s.',
                  {'mapdata': mapdata, 'ports': ports})
        return mapdata, ports

    def _set_tfog_pair(self, port, host_num):
        """set tfo pair"""
        LOG.debug('_set_tfog_pair, '
                  'port: %(port)s, '
                  'host_num: %(host_num)s.',
                  {'port': port,
                   'host_num': host_num})
        option = {
            'port': port,
            'host-number': host_num
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'set_tfog_pair',
            remote=True, retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            msg = (_('_set_tfog_pair, '
                     'set_tfog_pair failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    @dx_utils.FJDXLockutils('connect', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _map_lun_without_map_file(self, volume, connector,
                                  host_tfo_ports_dict,
                                  remote=False,
                                  replica_volume=False,
                                  lun_no=None,
                                  timeout=None):
        """Normal way to do attach volume."""

        LOG.debug('_map_lun_without_map_file, '
                  'volume id: %(vol_id)s, '
                  'connector: %(connector)s, '
                  'host_tfo_ports_dict: %(ports_dict)s, '
                  'remote: %(remote)s, '
                  'replica_volume: %(replica_volume)s.',
                  {'vol_id': volume['id'],
                   'connector': connector,
                   'ports_dict': host_tfo_ports_dict,
                   'remote': remote,
                   'replica_volume': replica_volume})

        # Get hosts info from connector.
        initiatorlist = self._find_initiator_names(connector)

        if self.protocol == 'FC':
            host_info_list = [x.lower() for x in initiatorlist]
        elif self.protocol == 'iSCSI':
            host_info_list = [initiatorlist[0].lower()]
        LOG.debug('_map_lun_without_map_file, '
                  'host_info_list: %(host_info_list)s.',
                  {'host_info_list': host_info_list})

        port_info_dict = self._get_port_info(remote=remote)
        LOG.debug('_map_lun_without_map_file, '
                  'port_info_dict: %(port_info_dict)s.',
                  {'port_info_dict': port_info_dict})

        # Check informations on ETERNUS.
        host_ag_list = []
        target_ag_list = []  # Exist affinity group list.
        avai_port_list = []  # Available ports for host affinity group.
        initiator_target_map = {}  # For FC zone.
        for host in host_info_list:
            # Get available ports for host affinity group.
            if replica_volume:
                if not remote:
                    avai_ports = host_tfo_ports_dict[host]['avai_own_ports']
                else:
                    avai_ports = host_tfo_ports_dict[host]['avai_pair_ports']
            else:
                if host in host_tfo_ports_dict:
                    avai_ports = host_tfo_ports_dict[host]['ports']
                else:
                    avai_ports = []

            # If there is no available ports, skip to next.
            if not avai_ports:
                continue

            # Create FC zone info for FC.
            if self.protocol == 'FC':
                port_wwpns = []
                for tmp_port in avai_ports:
                    if tmp_port in port_info_dict:
                        port_wwpn = port_info_dict[tmp_port]['port_wwpn']
                        port_wwpns.append(port_wwpn)
                initiator_target_map[host] = port_wwpns

            # Available ports for host affinity group.
            avai_port_list.extend(avai_ports)

            # Check if the host is registered.
            host_num = self._check_if_registered(host, remote=remote)

            host_ag = {
                'host_identifier': None,
                'host_number': None,
                'ports': [],
                'ag_numbers': [],
                'avai_ports': []
            }
            host_ag['host_identifier'] = host
            host_ag['host_number'] = host_num
            host_ag['avai_ports'] = avai_ports

            # Get exist affinity groups.
            if host_num is not None:
                clidata = self._get_ags_info_by_host_number(host_num,
                                                            remote=remote)
                for tmp in clidata:
                    ag_num = tmp.get('AG Num')
                    port = tmp.get('Port')

                    # Need to check if the port is suitable for protocol.
                    if port in avai_ports:
                        host_ag['ag_numbers'].append(ag_num)
                        host_ag['ports'].append(port)
                        if ag_num not in target_ag_list:
                            target_ag_list.append(ag_num)

            host_ag_list.append(host_ag)
            # End of check.

        if not avai_port_list:
            msg = (_('_map_lun_without_map_file, '
                     'volume id: %(vol_id)s, target connector(%(connector)s) '
                     'has no available ports on ETERNUS.')
                   % {'vol_id': volume['id'],
                      'connector': connector})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if len(target_ag_list) > 1:
            msg = (_('_map_lun_without_map_file, '
                     'volume id: %(vol_id)s, target connector(%(connector)s) '
                     'has more than one affinity group with specified ports '
                     'on ETERNUS. Exist affinity group numbers: %(numbers)s.')
                   % {'vol_id': volume['id'],
                      'connector': connector,
                      'numbers': target_ag_list})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Register host on ETERNUS and output host informations.
        for host_ag in host_ag_list:
            host_num = host_ag['host_number']
            host = host_ag['host_identifier']
            if host_num is None:
                host_num = self._register_host(host, remote=remote)
                host_ag['host_number'] = host_num
            LOG.debug('_map_lun_without_map_file, '
                      'host info on ETERNUS: %s.', host_ag)

        if not target_ag_list:
            # Create affinity group and set host affinity.
            tmp_ports = []
            ag_number = self._create_affinity_group(volume, remote=remote)
            lunnumber = 0
            release_dict = {}  # {host number:[port, ...]}.
            try:
                for host_ag in host_ag_list:
                    host_num = host_ag['host_number']
                    ports = host_ag['avai_ports']

                    if self.protocol == 'iSCSI' and not replica_volume:
                        ports = self._filter_idle_ports(ports)
                    # Set host affinity.
                    self._create_host_affinity([str(host_num)],
                                               ports,
                                               ag_number,
                                               remote)
                    tmp_ports.extend(ports)
                    release_dict[host_num] = ports
            except Exception:
                msg = (_('_map_lun_without_map_file, '
                         'remote: %(remote)s, '
                         'host number: %(host_num)s, '
                         'ag number: %(ag_number)s, '
                         'ports: %(ports)s.')
                       % {'remote': remote,
                          'host_num': host_num,
                          'ag_number': ag_number,
                          'ports': ports})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
        else:
            # When target_ag_list == 1, need to check the exist host affinity
            # group. If iSCSI, just add volume to affinity group. If FC, add
            # volume to affinity group first, then check exist ports info
            # with fc_zone ports, make sure that at least create host affinity
            # group with fc_zone ports.
            ag_number = target_ag_list[0]

            lunnumber = self._add_volume_to_affinity_group(ag_number,
                                                           volume,
                                                           remote,
                                                           lunno=lun_no)
            tmp_ports = []
            for host_ag in host_ag_list:
                host_num = host_ag['host_number']
                avai_ports = host_ag['avai_ports']
                exist_ports = host_ag['ports']
                if self.protocol == 'iSCSI' and not replica_volume:
                    filtered_ports = [x for x in exist_ports]
                    if len(exist_ports) < self.need_port_num:
                        avai_ports = self._filter_idle_ports(avai_ports)
                        for avai_port in avai_ports:
                            if len(filtered_ports) < self.need_port_num and \
                                    avai_port not in filtered_ports:
                                filtered_ports.append(avai_port)
                    avai_ports = filtered_ports
                try:
                    # Get missing ports and set host affinity.
                    ports = (
                        [x for x in avai_ports if x not in exist_ports])
                    if ports:
                        self._create_host_affinity(
                            [str(host_num)],
                            ports,
                            ag_number,
                            remote=remote)
                    tmp_ports.extend(avai_ports)
                except Exception:
                    with excutils.save_and_reraise_exception():
                        # Remove LUN from affinity group.
                        self._remove_volume_from_affinity_group(
                            ag_number,
                            lunnumber,
                            remote=remote)

        mapdata = None
        if not remote:
            tmp_ports = list(set(tmp_ports))
            mapdata = self._get_mapdata_ext(volume,
                                            connector,
                                            tmp_ports,
                                            lunnumber)
            if self.protocol == 'FC':
                mapdata['initiator_target_map'] = initiator_target_map
        # Set tfo pair while remote is true.
        if remote:
            for host_ag in host_ag_list:
                host_num = host_ag['host_number']
                ports = host_ag['avai_ports']
                for port in ports:
                    self._set_tfog_pair(port, host_num)
        return mapdata, tmp_ports, lunnumber

    def _get_enternus_fc_zone_info(self, remote=False):
        """Get FC zone info on ETERNUS.

        This method is used for FC attachment to check if the host FC HBA
        has ability to connect to the ETERNUS FC port.
        @return: clidata. A dictionary likes {host WWPN: [port1, port2]}.
        host WWPN: The host FC HBA WWPN.
        port1: xyz: Fixed 3 digits, CM number + CA number + Port number.
        """
        LOG.debug('_get_enternus_fc_zone_info, '
                  'about to discover_host_wwn_names on ETERNUS, '
                  'remote:%(remote)s.',
                  {'remote': remote})

        self.ce_support = (
            self.local_ce_support if not remote else self.remote_ce_support)
        param_dict = {
            'ce_support': self.ce_support
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'discover_host_wwn_names',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **param_dict)

        if rc != 0:
            msg = (_('_get_enternus_fc_zone_info, '
                     'discover_host_wwn_names failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        # Clidata: {host WWPN: [port1, port2]}.
        LOG.debug('_get_enternus_fc_zone_info, '
                  'FC zone info on ETERNUS: %s.', clidata)
        return clidata

    def _filter_idle_ports(self, avai_ports):
        """Filter idle ports."""
        LOG.debug('_filter_idle_ports, '
                  'available ports: %(avai_ports)s.',
                  {'avai_ports': avai_ports})

        need_port_num = self.need_port_num
        if len(avai_ports) <= need_port_num:
            temp_ports = avai_ports
        else:
            port_aglen_dict = {}
            for port in avai_ports:
                clidata = self._get_ags_info_by_port(port)
                port_aglen_dict[port] = len(clidata)
            sort_port_list = sorted(port_aglen_dict.items(),
                                    key=lambda item: item[1])

            temp_ports = [
                sort_port_list[i][0] for i in range(need_port_num)
            ]

        LOG.debug('_filter_idle_ports, '
                  'selected ports: %(temp_ports)s.',
                  {'temp_ports': temp_ports})
        return temp_ports

    def _get_ags_info_by_port(self, port):
        """Get host affinity groups info by target port.

        This is a common method to get all host affinity groups info
        for the target port.

        WARNING: FC host and iSCSI host may have the same
        host number on ETERNUS. You can differ by port type.

        @param port: Target port in ETERNUS.
        @return: clidata. A list likes
        [{Port:x,Host Num:x,Host Name:x,AG Num:x,AG Name:x},...].
        """
        LOG.debug('_get_ags_info_by_port, '
                  'target port: %s.', port)

        option = {
            'port': port,
            'ce_support': self.local_ce_support
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'show_host_affinity',
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            msg = (_('_get_ags_info_by_port, '
                     'show_host_affinity failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        # Clidata: [{Port:x,Host Num:x,Host Name:x,AG Num:x,AG Name:x},...].
        LOG.debug('_get_ags_info_by_port, '
                  'target port number: %(port)s, '
                  'host affinity groups info: %(info)s.',
                  {'port': port,
                   'info': clidata})
        return clidata

    def _check_if_registered(self, host, remote=False):
        """Check if the target host is registered.

        If registered, return host number, else return None.
        @param host: Host IQN(iSCSI) or the WWPN(FC).
        @return: host_num. Host number on ETERNUS.
        """
        LOG.debug('_check_if_registered, '
                  'target host: %(host)s, '
                  'remote: %(remote)s.',
                  {'host': host, 'remote': remote})

        if self.protocol == 'FC':
            cmd = 'show_host_wwn_names'
        elif self.protocol == 'iSCSI':
            cmd = 'show_host_iscsi_names'

        rc, emsg, clidata = self._exec_eternus_cli(cmd,
                                                   remote=remote,
                                                   retry=self.retry_times,
                                                   timeout=self.req_sec)
        if rc != 0:
            msg = (_('_check_if_registered, '
                     '%(cmd)s failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'cmd': cmd,
                      'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        # Clidata: [{Host Name: XXX, IP Version: XXX, Host Num: XXX}, ...].
        host_num = None
        host_ipver = None
        for hostdata in clidata:
            if hostdata['Host Name'].lower() == host:
                if self.protocol == 'FC':
                    host_num = hostdata['Host Num']
                    break
                elif self.protocol == 'iSCSI':
                    if (self.use_ipv6 and
                            hostdata['IP Version'] != CONSTANTS.ISCSI_IPV6):
                        continue
                    elif (not self.use_ipv6 and
                            hostdata['IP Version'] != CONSTANTS.ISCSI_IPV4):
                        continue

                    host_num = hostdata['Host Num']
                    if hostdata['IP Version'] == CONSTANTS.ISCSI_IPV4:
                        host_ipver = 'ipv4'
                    elif hostdata['IP Version'] == CONSTANTS.ISCSI_IPV6:
                        host_ipver = 'ipv6'
                    break

        LOG.debug('_check_if_registered, target host: %(host)s, '
                  'host number on ETERNUS: %(number)s, '
                  'IP Version: %(ipver)s.',
                  {'host': host, 'number': host_num, 'ipver': host_ipver})
        return host_num

    def _get_ags_info_by_host_number(self, host_number, remote=False):
        """Get host affinity groups info by host number.

        This is a common method to get all host affinity groups info
        for the target host.

        WARNING: FC host and iSCSI host may have the same
        host number on ETERNUS. You can differ by port type.

        @param host_number: Host number on ETERNUS.
        @return: clidata. A list likes
        [{Port:x,Host Num:x,Host Name:x,AG Num:x,AG Name:x},...].
        """
        LOG.debug('_get_ags_info_by_host_number, '
                  'target host number: %(host)s, '
                  'remote: %(remote)s.',
                  {'host': host_number, 'remote': remote})

        self.ce_support = (
            self.local_ce_support if not remote else self.remote_ce_support)
        option = {
            'host-number': host_number,
            'ce_support': self.ce_support
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'show_host_affinity',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            msg = (_('_get_ags_info_by_host_number, '
                     'show_host_affinity failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        # Clidata: [{Port:x,Host Num:x,Host Name:x,AG Num:x,AG Name:x},...].
        LOG.debug('_get_ags_info_by_host_number, '
                  'target host number: %(number)s, '
                  'host affinity groups info: %(info)s.',
                  {'number': host_number,
                   'info': clidata})
        return clidata

    def _register_host(self, host, remote=False):
        """Register host and return the host number.

        @param host: Host IQN(iSCSI) or the WWPN(FC).
        @return: host_num. Host number on ETERNUS.
        """
        LOG.debug('_register_host, '
                  'target host: %(host)s, '
                  'remote: %(remote)s.',
                  {'host': host, 'remote': remote})

        if self.protocol == 'FC':
            cmd = 'create_host_wwn_name'
            option = {
                'wwn': host
            }
        elif self.protocol == 'iSCSI':
            cmd = 'create_host_iscsi_name'
            if self.use_ipv6:
                ipver = 'ipv6'
                option = {
                    'iscsi-name': host,
                    'ip-version': ipver
                }
            else:
                ipver = 'ipv4'
                option = {
                    'iscsi-name': host,
                    'ip-version': ipver
                }

        rc, emsg, clidata = self._exec_eternus_cli(
            cmd,
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)

        if rc != 0:
            msg = (_('_register_host, '
                     '%(cmd)s failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'cmd': cmd,
                      'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        # Clidata is the direct result of the command.
        host_num = int(clidata[0], 16)

        LOG.debug('_register_host, target host: %(host)s, '
                  'host number on ETERNUS: %(number)s, '
                  'IP Version: %(ipver)s.',
                  {'host': host, 'number': host_num, 'ipver': ipver})
        return host_num

    def _get_model_supplement_info(self, remote=False):
        """Get model supplement info. """

        LOG.debug('_get_model_supplement_info start.')

        rc, emsg, clidata = self._exec_eternus_cli(
            'show_model_supplement_info',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec)
        if rc != 0:
            msg = (_('_get_model_supplement_info, '
                     '_show_model_supplement_info failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        # Clidata: {machine_info:x,model_info:y}.
        LOG.debug('_get_model_supplement_info, '
                  'model supplement info: %(info)s.',
                  {'info': clidata})
        return clidata

    def _create_affinity_group(self, volume, remote=False):
        """Create a new affinity group on ETERNUS.

        Create a new affinity group with target volume.
        @param volume: Cinder volume.
        @return: ag_num. Affinity group number on ETERNUS.
        """
        LOG.debug('_create_affinity_group, volume id: %s.', volume['id'])

        vol_name = self._get_volume_name(volume)
        option = {
            'volume-name': vol_name, 'lun': 0
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'create_affinity_group',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            msg = (_('_create_affinity_group, '
                     'create_affinity_group failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        ag_num = int(clidata[0], 16)

        LOG.debug('_create_affinity_group, '
                  'volume id: %(vol_id)s, volume name: %(vol_name)s, '
                  'affinity group number: %(ag_num)s.',
                  {'vol_id': volume['id'],
                   'vol_name': vol_name,
                   'ag_num': ag_num})
        return ag_num

    def _create_host_affinity(self, hostnolist, portidlist, agnum,
                              remote=False):
        """Create host affinity on ETERNUS."""
        LOG.debug('_create_host_affinity, '
                  'host numbers: %(hosts)s, ports: %(ports)s, '
                  'affinity group number: %(number)s, '
                  'remote: %(remote)s.',
                  {'hosts': hostnolist,
                   'ports': portidlist,
                   'number': agnum,
                   'remote': remote})

        option = {
            'ag-number': agnum,
            'host-number': ','.join(hostnolist),
            'port': ','.join(portidlist)
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'set_host_affinity',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            msg = (_('_create_host_affinity, '
                     'set_host_affinity failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_create_host_affinity, '
                  'host numbers: %(hosts)s, ports: %(ports)s, '
                  'affinity group number: %(number)s. Success.',
                  {'hosts': hostnolist,
                   'ports': portidlist,
                   'number': agnum})

    def _add_volume_to_affinity_group(self, ag_number, volume, remote=False,
                                      lunno=None):
        """Add target volume to the specified affinity group.

        @param ag_number: Affinity group number on ETERNUS.
        @param volume: Cinder volume.
        @return: lun_no. LUN NO in affinity group.
        """
        vol_id = volume['id']
        LOG.debug('_add_volume_to_affinity_group, '
                  'volume id: %(vol_id)s, affinity group number: %(number)s.',
                  {'vol_id': vol_id, 'number': ag_number})

        # Calculate LUN NO in affinity group.
        option = {
            'ag-number': ag_number
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'show_affinity_group',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            msg = (_('_add_volume_to_affinity_group, '
                     'show_affinity_group failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Clidata: [{Lun Num:XXX, Lun Name:XXX}, ...].
        if not clidata:
            msg = (_('_add_volume_to_affinity_group, '
                     'the target affinity group(%(number)s) '
                     'does not exist on ETERNUS.')
                   % {'number': ag_number})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        lun_no = lunno
        if lun_no is None:
            idx = 0
            used_nos = []
            for data in clidata:
                no = int(data['Lun Num'])
                used_nos.append(no)
                idx = no if idx < no else idx

            lun_no = idx + 1
        if CONSTANTS.MAX_AG_LUN_NO <= lun_no:
            temp = range(CONSTANTS.MAX_AG_LUN_NO)
            unused_num = None
            for t in temp:
                if t not in used_nos:
                    unused_num = t
                    break

            lun_no = unused_num
            if lun_no is None:
                msg = (_('_add_volume_to_affinity_group, '
                         'volume id: %(vol_id)s, '
                         'the volumes in affinity group(%(number)s) '
                         'is full(MAX count: %(count)s).')
                       % {'vol_id': vol_id,
                          'number': ag_number,
                          'count': CONSTANTS.MAX_AG_LUN_NO})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_add_volume_to_affinity_group, '
                  'volume id: %(vol_id)s, '
                  'affinity group number: %(number)s, '
                  'LUN No: %(lun_no)s.',
                  {'vol_id': vol_id,
                   'lun_no': lun_no,
                   'number': ag_number})

        # Set affinity group.
        volumename = self._get_volume_name(volume)
        option = {
            'ag-number': ag_number,
            'volume-name': volumename,
            'lun': lun_no
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'set_affinity_group',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)

        if rc != 0:
            # If error, check if the volume has been mapped.
            option = {
                'ag-number': ag_number
            }
            rc_tmp, emsg_tmp, clidata_tmp = self._exec_eternus_cli(
                'show_affinity_group',
                remote=remote,
                retry=self.retry_times,
                timeout=self.req_sec,
                **option)
            if rc_tmp != 0:
                msg = (_('_add_volume_to_affinity_group, '
                         'show_affinity_group failed, '
                         'Return code: %(rc)lu, '
                         'Error: %(emsg)s, '
                         'Message: %(clidata)s.')
                       % {'rc': rc_tmp,
                          'emsg': emsg_tmp,
                          'clidata': clidata_tmp})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Clidata: [{Lun Num:XXX, Lun Name:XXX}, ...].
            temp_no = None
            for data in clidata_tmp:
                name = data['Lun Name']
                if volumename == name:
                    temp_no = int(data['Lun Num'])
                    break

            if temp_no is not None:
                lun_no = temp_no
            else:
                msg = (_('_add_volume_to_affinity_group, '
                         'set_affinity_group failed, '
                         'Return code: %(rc)lu, '
                         'Error: %(emsg)s, '
                         'Message: %(clidata)s.')
                       % {'rc': rc,
                          'emsg': emsg,
                          'clidata': clidata})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_add_volume_to_affinity_group, '
                  'volume id: %(vol_id)s, '
                  'affinity group number: %(number)s, '
                  'LUN No in affinity group: %(lun_no)s. Success.',
                  {'vol_id': vol_id,
                   'lun_no': lun_no,
                   'number': ag_number})
        return lun_no

    def _remove_volume_from_affinity_group(self, ag_number, lun_no,
                                           remote=False):
        """Remove volume from affinity group.

        @param ag_number: Affinity group number on ETERNUS.
        @param lun_no: LUN NO in affinity group.
        """
        LOG.debug('_remove_volume_from_affinity_group, '
                  'affinity group number: %(number)s, LUN NO: %(no)s.',
                  {'number': ag_number,
                   'no': lun_no})

        option = {
            'ag-number': ag_number,
            'lun': lun_no
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'delete_affinity_group',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            msg = (_('_remove_volume_from_affinity_group, '
                     'delete_affinity_group failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_remove_volume_from_affinity_group, '
                  'affinity group number: %(num)s, LUN NO: %(no)s. Success.',
                  {'num': ag_number,
                   'no': lun_no})

    def _get_mapdata_ext(self, volume, connector, ports, lunnum):
        """Return mapping information."""
        mapdata = None
        multipath = connector.get('multipath', True)

        LOG.debug('_get_mapdata_ext, volume id: %(id)s, '
                  'protocol: %(prtcl)s, multipath: %(mpath)s.',
                  {'id': volume['id'],
                   'prtcl': self.protocol, 'mpath': multipath})

        if self.protocol == 'iSCSI':
            mapdata = self._get_mapdata_iscsi_ext(multipath, ports, lunnum)
        elif self.protocol == 'FC':
            mapdata = self._get_mapdata_fc_ext(ports, lunnum)

        return mapdata

    def _get_mapdata_iscsi_ext(self, multipath, ports, lunnum):
        """Get mapdata for iSCSI."""
        target_portals = []
        target_iqns = []
        target_luns = []

        iscsi_port = self.configuration.iscsi_port

        for port in ports:
            if port in self.port_info_dict:
                target_portal = "%s:%s" % (
                    self.port_info_dict[port]['port_ip'], iscsi_port)
                target_iqn = self.port_info_dict[port]['port_iqn']
                target_lun = lunnum
                target_portals.append(target_portal)
                target_iqns.append(target_iqn)
                target_luns.append(target_lun)

        mapdata = {}
        if multipath:
            mapdata = {
                'target_portals': target_portals,
                'target_iqns': target_iqns,
                'target_luns': target_luns,
            }

        if target_portals and target_iqns and target_luns:
            mapdata['target_portal'] = target_portals[0]
            mapdata['target_iqn'] = target_iqns[0]
            mapdata['target_lun'] = target_luns[0]
        else:
            mapdata = {
                'target_portal': None,
                'target_iqn': None,
                'target_lun': None,
            }
            msg = (_('_get_mapdata_iscsi_ext, '
                     'mapdata: %(mapdata)s, '
                     'all of the ports in config are not used.')
                   % {'mapdata': mapdata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_get_mapdata_iscsi_ext, mapdata: %s.', mapdata)
        return mapdata

    def _get_mapdata_fc_ext(self, ports, lunnum):
        """Get mapdata for FibreChannel."""
        target_wwpn = []

        for port in ports:
            if port in self.port_info_dict:
                target_wwpn.append(self.port_info_dict[port]['port_wwpn'])

        target_lun = lunnum

        mapdata = {
            'target_wwn': target_wwpn,
            'target_lun': target_lun,
        }

        LOG.debug('_get_mapdata_fc_ext, mapdata: %s.', mapdata)
        return mapdata

    @dx_utils.FJDXLockutils('connect', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _unmap_lun_without_map_file(self, volume, connector,
                                    remote=False, replica_volume=False,
                                    timeout=None):
        """Using normal way to detach volume """
        LOG.debug('_unmap_lun_without_map_file, '
                  'volume id: %(vol_id)s, connector: %(connector)s, '
                  'remote: %(remote)s, replica_volume: %(replica_volume)s, '
                  'timeout: %(timeout)s.',
                  {'vol_id': volume['id'],
                   'connector': connector,
                   'remote': remote,
                   'replica_volume': replica_volume,
                   'timeout': timeout})

        volumename = self._get_volume_name(volume)

        # Ag_dict: {ag_number:{host_number:[port1,...]}}
        # Host_dict: {host_number:host_IQN/WWPN}.
        ag_dict, host_dict = (
            self._get_affinity_groups_by_connector(
                connector,
                remote=remote,
                replica_volume=replica_volume))
        ag_numbers = ag_dict.keys()

        init_tgt_map = {}
        if not ag_numbers:
            count = 0
        else:
            count = 0
            for ag_number in ag_numbers:
                option = {
                    'ag-number': ag_number
                }
                rc, emsg, clidata = self._exec_eternus_cli(
                    'show_affinity_group',
                    remote=remote, retry=self.retry_times,
                    timeout=self.req_sec,
                    **option)

                if rc != 0:
                    msg = (_('_unmap_lun_without_map_file, '
                             'show_affinity_group failed. '
                             'Return code: %(rc)lu, '
                             'Error: %(emsg)s, '
                             'Message: %(clidata)s.')
                           % {'rc': rc,
                              'emsg': emsg,
                              'clidata': clidata})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                if not clidata:
                    LOG.warning('_unmap_lun_without_map_file, '
                                'affinity group(%(number)s) is '
                                'not exist on ETERNUS.',
                                {'number': ag_number})
                    continue

                lunnum = None
                map_num = len(clidata)
                for data in clidata:
                    name = data['Lun Name']
                    if volumename == name:
                        lunnum = data['Lun Num']
                        break

                if lunnum is None:
                    count = map_num if map_num > count else count
                    LOG.warning('_unmap_lun_without_map_file, '
                                'there is no volume:%(vol_name)s '
                                'in affinity group:%(number)s.',
                                {'vol_name': volumename,
                                 'number': ag_number})
                    continue

                LOG.debug('_unmap_lun_without_map_file, '
                          'volume id: %(vol_id)s, '
                          'affinity group number: %(number)s, '
                          'LUN No in affinity group: %(lun_no)s.',
                          {'vol_id': volume['id'],
                           'lun_no': lunnum,
                           'number': ag_number})

                # Delete host affinity.
                # If only one volume in affinity group, need to
                # Release host affinity first.
                if map_num == 1:
                    host_ports_dict = ag_dict.get(ag_number, {})
                    host_numbers = host_ports_dict.keys()

                    LOG.debug('_unmap_lun_without_map_file, '
                              'volume id: %(vol_id)s, '
                              'affinity group number: %(number)s, '
                              'LUN No in affinity group: %(lun_no)s. '
                              'Only one volume in affinity group, '
                              'need to release host affinity. '
                              'Host numbers: %(host_numbers)s.',
                              {'vol_id': volume['id'],
                               'lun_no': lunnum,
                               'number': ag_number,
                               'host_numbers': host_numbers})

                    for host_number in host_numbers:
                        ports = host_ports_dict.get(host_number, [])
                        if ports:
                            self._release_host_affinity(host_number,
                                                        ports,
                                                        remote=remote)
                            if self.protocol == 'FC':  # For FC detachment.
                                tmp_host = host_dict.get(host_number, None)
                                if tmp_host:
                                    tmp_ports = init_tgt_map.get(tmp_host, [])
                                    tmp_ports.extend(ports)
                                    init_tgt_map[tmp_host] = tmp_ports

                    self._remove_volume_from_affinity_group(ag_number,
                                                            lunnum,
                                                            remote=remote)
                else:
                    # Remove LUN from affinity group.
                    self._remove_volume_from_affinity_group(
                        ag_number, lunnum, remote=remote)

                map_num = map_num - 1
                count = map_num if map_num > count else count
                LOG.debug('_unmap_lun_without_map_file, '
                          'volume id: %(vol_id)s, '
                          'affinity group number: %(number)s, '
                          'unmap successfully.',
                          {'vol_id': volume['id'],
                           'number': ag_number})

        # Format FC zone info.
        fc_zone_dict = {}
        port_info_dict = self._get_port_info(remote=remote)
        if self.protocol == 'FC':
            for host, ports in init_tgt_map.items():
                ports = list(set(ports))
                port_wwpns = []
                for port in ports:
                    if port in port_info_dict:
                        port_wwpn = port_info_dict[port]['port_wwpn']
                        port_wwpns.append(port_wwpn)
                fc_zone_dict[host] = port_wwpns

        LOG.debug('_unmap_lun_without_map_file, '
                  'volume id: %(vol_id)s, connector: %(connector)s, '
                  'count of mapping volume on the connector: %(count)s.',
                  {'vol_id': volume['id'],
                   'connector': connector,
                   'count': count})
        return count, fc_zone_dict

    def _get_affinity_groups_by_connector(self, connector,
                                          remote=False, replica_volume=False):
        """Get all affinity groups of the target connector."""
        LOG.debug('_get_affinity_groups_by_connector, '
                  'connector: %(connector)s, '
                  'remote: %(remote)s.',
                  {'connector': connector,
                   'remote': remote})

        initiatorlist = self._find_initiator_names(connector)
        host_infos = [x.lower() for x in initiatorlist]

        if not replica_volume:
            _x, ports = self._get_target_port_info()
        elif remote:
            _x, ports = self._filter_available_tfo_port()
        else:
            ports, _x = self._filter_available_tfo_port()

        # Ag_dict: {ag_number:{host_number:[port1,...]}}
        # Host_dict: {host_number:host_IQN/WWPN}.
        ag_dict = {}
        host_dict = {}
        for host in host_infos:
            host_number = self._check_if_registered(host, remote=remote)
            if host_number is not None:
                clidata = self._get_ags_info_by_host_number(host_number,
                                                            remote=remote)
                if clidata:
                    for tmp in clidata:
                        ag = tmp.get('AG Num')
                        port = tmp.get('Port')
                        if port in ports:
                            host_dict[host_number] = host

                            tmp_host_port_dict = ag_dict.get(ag, {})
                            tmp_ports = tmp_host_port_dict.get(
                                host_number, [])
                            tmp_ports.append(port)

                            tmp_host_port_dict[host_number] = tmp_ports
                            ag_dict[ag] = tmp_host_port_dict
                            LOG.debug('_get_affinity_groups_by_connector, '
                                      'host: %(host)s, port: %(port)s, '
                                      'affinity group: %(number)s.',
                                      {'host': host,
                                       'port': port,
                                       'number': ag})

        LOG.debug('_get_affinity_groups_by_connector, '
                  'connector: %(connector)s, '
                  'host dict: %(host_dict)s, '
                  'affinity groups dict: %(ag_dict)s.',
                  {'connector': connector,
                   'host_dict': host_dict,
                   'ag_dict': ag_dict})
        return ag_dict, host_dict

    def _get_target_port_info(self):
        """Return target_portid."""
        lunportlist = []
        target_portlist = []
        target_cliportidlist = []

        lunportinslist = self._enum_eternus_instances(
            'FUJITSU_LUNMappingController',
            retry=self.retry_times)

        for lunportins in lunportinslist:
            lunportlist.append(
                self._conv_port_id(
                    lunportins['CeID'] if 'CeID' in lunportins else None,
                    lunportins['CMSlotNumber'],
                    lunportins['CASlotNumber'],
                    lunportins['PortNo']))

        if self.protocol == 'FC':
            prtcl_endpoint = 'FUJITSU_SCSIProtocolEndpoint'
            connection_type = 2
        elif self.protocol == 'iSCSI':
            prtcl_endpoint = 'FUJITSU_iSCSIProtocolEndpoint'
            connection_type = 7

        try:
            tgtportlist = self._enum_eternus_instances(
                prtcl_endpoint,
                retry=self.retry_times)
        except Exception:
            msg = _('_get_target_port_info, '
                    'EnumerateInstances, '
                    'cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for tgtport in tgtportlist:
            LOG.debug('_get_target_port_info, '
                      'tgtport: %(tgtport)s.',
                      {'tgtport': tgtport})
            cliportid = (
                self._conv_port_id(
                    tgtport['CeID'] if 'CeID' in tgtport else None,
                    tgtport['CMSlotNumber'],
                    tgtport['CASlotNumber'],
                    tgtport['PortNo']))
            if tgtport['ConnectionType'] != connection_type:
                continue
            if tgtport['RAMode'] & 0x7B != 0x00:
                continue
            if 'SCGroupNo' in tgtport:
                continue
            if cliportid in lunportlist:
                continue
            target_portlist.append(tgtport)
            target_cliportidlist.append(cliportid)

            LOG.debug('_get_target_port_info, '
                      'connection type: %(cont)s, '
                      'ramode: %(ramode)s.',
                      {'cont': tgtport['ConnectionType'],
                       'ramode': tgtport['RAMode']})

        LOG.debug('_get_target_port_info, '
                  'target port: %(target_portid)s.',
                  {'target_portid': target_portlist})
        LOG.debug('_get_target_port_info, '
                  'target portid: %(target_portid)s.',
                  {'target_portid': target_cliportidlist})

        if not target_portlist:
            msg = (_('_get_target_port_info, '
                     'protcol: %(protocol)s, '
                     'target_port does not found.')
                   % {'protocol': self.protocol})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return target_portlist, target_cliportidlist

    @staticmethod
    def _conv_port_id(ce_id, cm_slot_num, ca_slot_num, ca_port_num):
        """Convert port info to standard string."""
        if ce_id is None:
            port_id = '%x%x%x' % (cm_slot_num, ca_slot_num, ca_port_num)
        else:
            port_id = '%x%x%x%x' % (ce_id,
                                    cm_slot_num,
                                    ca_slot_num,
                                    ca_port_num)

        return port_id

    def _release_host_affinity(self, host_number, portidlist, remote=False):
        """Release host affinity on ETERNUS."""
        LOG.debug('_release_host_affinity, '
                  'host number: %(host)s, '
                  'ports: %(ports)s, '
                  'remote: %(remote)s.',
                  {'host': host_number,
                   'ports': portidlist,
                   'remote': remote})

        option = {
            'host-number': host_number,
            'port': ','.join(portidlist)
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'release_host_affinity',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)
        if rc != 0:
            data = self._get_ags_info_by_host_number(host_number,
                                                     remote=remote)
            if data:
                for tmp in data:
                    port = tmp.get('Port')
                    if port in portidlist:
                        msg = (_('_release_host_affinity, '
                                 'release_host_affinity failed. '
                                 'Return code: %(rc)lu, '
                                 'Error: %(emsg)s, '
                                 'Message: %(clidata)s.')
                               % {'rc': rc,
                                  'emsg': emsg,
                                  'clidata': clidata})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_release_host_affinity, '
                  'host number: %(host)s, ports: %(ports)s. Success.',
                  {'host': host_number,
                   'ports': portidlist})

    def _map_ordinary_volume(self, volume, connector):
        """Map non-replication volume. """
        LOG.debug('_map_ordinary_volume, '
                  'volume id: %(id)s.',
                  {'id': volume['id']})

        initiator_in_map_file_num = 0
        initiatorlist = self._find_initiator_names(connector)
        if self.host_ag_dict:
            for initiator in initiatorlist:
                hostname = initiator.lower()
                if str(hostname) in self.host_ag_dict:
                    initiator_in_map_file_num += 1

        ports = None
        lunnum = None
        mapdata = None
        # Map lun with mapping file.
        if self.host_ag_dict and initiator_in_map_file_num != 0:

            ports, lungroup = self._get_affinity_group_ext(connector)

            used_ports = [x for x in ports if x in self.port_info_dict]
            if not used_ports:
                msg = (_('_map_ordinary_volume, '
                         'all the ports in cfg-file are not used, '
                         'ports in cfg-file: %(cfg_ports)s, '
                         'ports on ETERNUS: %(e_ports)s.')
                       % {'cfg_ports': ports,
                          'e_ports': self.port_info_dict})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            try:
                lunnum = self._map_lun(volume, connector, lungroup,
                                       timeout=self.req_sec)
            except Exception:
                with excutils.save_and_reraise_exception():
                    self._unmap_lun_ext(volume, connector)

            mapdata = self._get_mapdata_ext(volume, connector, ports, lunnum)
        # Normal way to do attach volume.
        else:
            host_tfo_ports_dict = self._get_ordinary_port(initiatorlist)
            try:
                mapdata, ports, lunmumber = self._map_lun_without_map_file(
                    volume,
                    connector,
                    host_tfo_ports_dict,
                    timeout=self.req_sec)
            except Exception:
                with excutils.save_and_reraise_exception():
                    self._unmap_lun_without_map_file(volume, connector,
                                                     remote=False,
                                                     replica_volume=False,
                                                     timeout=self.req_sec)

        mapdata['target_discoverd'] = True
        mapdata['volume_id'] = volume['id']
        LOG.debug('_map_ordinary_volume, '
                  'mapdata: %(mapdata)s, '
                  'ports: %(ports)s.',
                  {'mapdata': mapdata, 'ports': ports})
        return mapdata, ports

    def _get_ordinary_port(self, initiatorlist):
        _x, port_list = self._get_target_port_info()
        avail_ports = [x for x in port_list if x in self.port_info_dict]
        host_tfo_ports_dict = {}
        if self.protocol == 'FC':
            host_info_list = [x.lower() for x in initiatorlist]
            fc_zone = self._get_enternus_fc_zone_info()
            for host in host_info_list:
                # Get available ports for host affinity group.
                if fc_zone:  # For FC.
                    tmp = fc_zone.get(host, [])
                    tmp_avai_ports = [x for x in tmp if x in avail_ports]
                    host_tfo_ports_dict[host] = {'ports': tmp_avai_ports}
        elif self.protocol == 'iSCSI':
            host_info_list = [initiatorlist[0].lower()]
            for host in host_info_list:
                host_tfo_ports_dict[host] = {'ports': avail_ports}
        return host_tfo_ports_dict

    def _get_affinity_group_ext(self, connector):
        """Get host affinitiy information."""
        initiatorlist = self._find_initiator_names(connector)

        ports = []
        lungroup = None
        for initiator in initiatorlist:
            hostname = initiator.lower()
            if str(hostname) in self.host_ag_dict:
                temp_ports = self.host_ag_dict[hostname]['portlist']
                temp_lungroup = self.host_ag_dict[hostname]['lungroup']
                LOG.debug('_get_affinity_group_ext, '
                          'initiator: %(hostname)s, '
                          'ports: %(ports)s, '
                          'lungroup: %(lungroup)s.',
                          {'hostname': hostname,
                           'ports': temp_ports,
                           'lungroup': temp_lungroup})

                if lungroup and lungroup != temp_lungroup:
                    msg = (_('_get_affinity_group_ext, '
                             'initiator list: %(list)s, '
                             'the specified host(%(connector)s) '
                             'must have the same LUN Group, '
                             'please confirm the configuration file.')
                           % {'list': initiatorlist,
                              'connector': connector})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                ports += temp_ports
                lungroup = temp_lungroup

        ports = list(set(ports))
        if not ports or not lungroup:
            msg = (_('_get_affinity_group_ext, '
                     'initiatorlist: %(initiatorlist)s, '
                     'cannot find port or Lun Group info from config file.')
                   % {'initiatorlist': initiatorlist})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return ports, lungroup

    def _get_map_number(self, connector):
        """Get map info from Eternus."""
        LOG.debug('get_map_number, connector: %s.', connector)

        if not self.host_ag_dict:
            ag_dict, _x = self._get_affinity_groups_by_connector(connector)
            ag_numbers = ag_dict.keys()
            if not ag_numbers:
                return 0
            option = {
                'ag-number': ag_numbers[0]
            }
        else:
            _x, agname = self._get_affinity_group_ext(connector)
            option = {
                'ag-name': agname
            }

        # Get LUN No in affinity group.
        rc, emsg, clidata = self._exec_eternus_cli(
            'show_affinity_group',
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)

        if rc != 0:
            msg = (_('get_map_number, '
                     'show_affinity_group, '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        map_num = len(clidata)

        return map_num

    def _unmap_lun_ext(self, volume, connector):
        """Unmap volume from host."""
        vol_id = volume['id']
        LOG.debug('_unmap_lun_ext, '
                  'volume id: %(vol_id)s, '
                  'connector: %(connector)s.',
                  {'vol_id': vol_id,
                   'connector': connector})

        _x, agname = self._get_affinity_group_ext(connector)

        # Get LUN No in affinity group.
        option = {
            'ag-name': agname
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'show_affinity_group',
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)

        if rc != 0:
            msg = (_('_unmap_lun_ext, '
                     'show_affinity_group, '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not clidata:
            map_num = 0
            LOG.warning('_unmap_lun_ext, '
                        'WARNING!! '
                        'There is no lungroup: %(ag)s in ETERNUS.',
                        {'ag': agname})
            return map_num

        lunnum = None
        map_num = len(clidata)
        volumename = self._get_volume_name(volume)
        for data in clidata:
            name = data['Lun Name']
            if volumename == name:
                lunnum = data['Lun Num']
                break

        if lunnum is None:
            LOG.warning('_unmap_lun_ext, '
                        'WARNING!! '
                        'There is no volume: %(vol)s in lungroup:%(ag)s.',
                        {'vol': volumename,
                         'ag': agname})
            return map_num

        LOG.debug('_unmap_lun_ext, '
                  'volume id: %(vol_id)s, '
                  'affinity group name: %(ag_name)s, '
                  'volume No in affinity group: %(lun_no)s.',
                  {'vol_id': vol_id,
                   'lun_no': lunnum,
                   'ag_name': agname})

        # Remove LUN from affinity group.
        option = {
            'ag-name': agname,
            'lun': lunnum
        }
        rc, emsg, clidata = self._exec_eternus_cli(
            'delete_affinity_group',
            retry=self.retry_times,
            timeout=self.req_sec,
            **option)

        if rc != 0:
            msg = (_('_unmap_lun_ext, '
                     'delete_affinity_group, '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        map_num = map_num - 1

        LOG.debug('_unmap_lun_ext, '
                  'volume id: %(vol_id)s, '
                  'affinity group name: %(ag_name)s, '
                  'unmap successfully.',
                  {'vol_id': vol_id,
                   'ag_name': agname})
        return map_num

    def _unmap_replication_volume(self, volume, connector):
        """Unmap replication volume. """
        LOG.debug('_unmap_replication_volume, volume id: %s.', volume['id'])

        map_num = None
        fc_zone_dict = {}

        lunnumber = self._get_lun_number(volume, connector,
                                         remote=True,
                                         replica_volume=True,
                                         timeout=self.req_sec)
        try:
            map_num, fc_zone_dict = (
                self._unmap_lun_without_map_file(volume,
                                                 connector,
                                                 remote=True,
                                                 replica_volume=True,
                                                 timeout=self.req_sec))
        except Exception:
            with excutils.save_and_reraise_exception():
                own_ports, pair_ports = self._filter_available_tfo_port()
                host_tfo_ports_dict = self._find_tfo_port_pairs(own_ports,
                                                                pair_ports,
                                                                connector,
                                                                volume)
                self._map_lun_without_map_file(volume,
                                               connector,
                                               host_tfo_ports_dict,
                                               remote=True,
                                               replica_volume=True,
                                               lun_no=lunnumber,
                                               timeout=self.req_sec)
        try:
            self._unmap_lun_without_map_file(volume,
                                             connector,
                                             remote=False,
                                             replica_volume=True,
                                             timeout=self.req_sec)
        except Exception:
            with excutils.save_and_reraise_exception():
                own_ports, pair_ports = self._filter_available_tfo_port()
                host_tfo_ports_dict = self._find_tfo_port_pairs(own_ports,
                                                                pair_ports,
                                                                connector,
                                                                volume)

                self._map_lun_without_map_file(volume,
                                               connector,
                                               host_tfo_ports_dict,
                                               remote=False,
                                               replica_volume=True,
                                               lun_no=lunnumber,
                                               timeout=self.req_sec)
                self._map_lun_without_map_file(volume,
                                               connector,
                                               host_tfo_ports_dict,
                                               remote=True,
                                               replica_volume=True,
                                               lun_no=lunnumber,
                                               timeout=self.req_sec)

        LOG.debug('_unmap_replication_volume, '
                  'map_num: %(map_num)s, '
                  'fc_zone_dict: %(fc_zone_dict)s.',
                  {'map_num': map_num, 'fc_zone_dict': fc_zone_dict})
        return map_num, fc_zone_dict

    def _get_lun_number(self, volume,
                        connector,
                        remote=False,
                        replica_volume=False,
                        timeout=None):
        LOG.debug('_get_lun_number, '
                  'volume id: %(vol_id)s, connector: %(connector)s, '
                  'remote: %(remote)s, replica_volume: %(replica_volume)s, '
                  'timeout: %(timeout)s.',
                  {'vol_id': volume['id'],
                   'connector': connector,
                   'remote': remote,
                   'replica_volume': replica_volume,
                   'timeout': timeout})

        volumename = self._get_volume_name(volume)

        ag_dict, host_dict = \
            self._get_affinity_groups_by_connector(
                connector,
                remote=remote,
                replica_volume=replica_volume)

        ag_numbers = ag_dict.keys()
        lunnum = None

        for ag_number in ag_numbers:
            option = {'ag-number': ag_number}
            rc, emsg, clidata = self._exec_eternus_cli(
                'show_affinity_group',
                remote=remote, retry=self.retry_times,
                timeout=self.req_sec,
                **option)

            if rc != 0:
                msg = (_('_get_lun_number, '
                         'show_affinity_group failed. '
                         'Return code: %(rc)lu, '
                         'Error: %(emsg)s, '
                         'Message: %(clidata)s.')
                       % {'rc': rc,
                          'emsg': emsg,
                          'clidata': clidata})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if not clidata:
                LOG.warning('_get_lun_number, '
                            'affinity group(%(number)s) is '
                            'not exist on ETERNUS.',
                            {'number': ag_number})
                continue

            for data in clidata:
                name = data['Lun Name']
                if volumename == name:
                    lunnum = data['Lun Num']
                    break

            if lunnum is None:
                LOG.warning('_get_lun_number, '
                            'there is no volume:%(vol_name)s '
                            'in affinity group:%(number)s.',
                            {'vol_name': volumename,
                             'number': ag_number})
                continue

            LOG.debug('_get_lun_number, '
                      'volume id: %(vol_id)s, '
                      'affinity group number: %(number)s, '
                      'LUN No in affinity group: %(lun_no)s.',
                      {'vol_id': volume['id'],
                       'lun_no': lunnum,
                       'number': ag_number})

        return lunnum

    def _unmap_ordinary_volume(self, volume, connector):
        """Unmap non-replication volume. """
        LOG.debug('_unmap_ordinary_volume, volume id: %s.', volume['id'])

        initiator_in_map_file_num = 0
        initiatorlist = self._find_initiator_names(connector)
        if self.host_ag_dict:
            for initiator in initiatorlist:
                hostname = initiator.lower()
                if str(hostname) in self.host_ag_dict:
                    initiator_in_map_file_num += 1

        map_num = None
        fc_zone_dict = {}
        # Unmap lun with mapping file.
        if self.host_ag_dict and initiator_in_map_file_num != 0:
            try:
                map_num = self._unmap_lun_ext(volume, connector)
            except Exception:
                with excutils.save_and_reraise_exception():
                    _x, lungroup = self._get_affinity_group_ext(connector)
                    self._map_lun(volume, connector, lungroup,
                                  timeout=self.req_sec)
        else:
            try:
                map_num, fc_zone_dict = self._unmap_lun_without_map_file(
                    volume,
                    connector,
                    timeout=self.req_sec)
            except Exception:
                with excutils.save_and_reraise_exception():
                    _x, avail_ports = self._get_target_port_info()
                    host_tfo_ports_dict = {}
                    if self.protocol == 'FC':
                        host_info_list = [x.lower() for x in initiatorlist]
                        for host in host_info_list:
                            host_tfo_ports_dict[host] = {'ports': avail_ports}
                    elif self.protocol == 'iSCSI':
                        host_info_list = [initiatorlist[0].lower()]
                        for host in host_info_list:
                            host_tfo_ports_dict[host] = {'ports': avail_ports}
                    self._map_lun_without_map_file(
                        volume,
                        connector,
                        host_tfo_ports_dict,
                        timeout=self.req_sec)

        LOG.debug('_unmap_ordinary_volume, '
                  'map_num: %(map_num)s, '
                  'fc_zone_dict: %(fc_zone_dict)s.',
                  {'map_num': map_num, 'fc_zone_dict': fc_zone_dict})
        return map_num, fc_zone_dict

    @dx_utils.FJDXLockutils('Set-qos', 'cinder-', True)
    @dx_utils.FJDXTimeout()
    def _set_qos(self, volume, remote=False, use_id=False, timeout=None):
        """Set volume qos using ETERNUS CLI."""
        LOG.debug('_set_qos, remote: %(remote)s, volumeid: %(volumeid)s.',
                  {'remote': remote, 'volumeid': volume['id']})
        try:
            qos_support = self._is_qos_or_format_support('QOS setting')
            # Storage is DX S2 series, qos is not supported.
            if not qos_support:
                return

            qos_specs_dict = self._get_qos_specs(volume)
            # Can not get anything from 'qos_specs_id'.
            if not qos_specs_dict:
                return

            # Get storage version information.
            rc, emsg, clidata = self._exec_eternus_cli(
                'show_enclosure_status',
                remote=remote,
                retry=self.retry_times,
                timeout=self.req_sec)
            if rc != 0:
                msg = (_('_set_qos, '
                         'show_enclosure_status failed. '
                         'Return code: %(rc)lu, '
                         'Error: %(errormsg)s, '
                         'Message: %(clidata)s.')
                       % {'rc': rc,
                          'errormsg': emsg,
                          'clidata': clidata})
                LOG.warning(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            category_dict = {}
            unsupport = []

            # If storage version is before V11L30.
            if clidata['version'] < 'V11L30-0000':
                for key, value in qos_specs_dict.items():
                    if (key in CONSTANTS.FJ_QOS_KEY_BYTES_list or
                            key in CONSTANTS.FJ_QOS_KEY_IOPS_list):
                        msg = (_('_set_qos, Can not support QoS '
                                 'parameter "%(key)s" on firmware version '
                                 '%(version)s.')
                               % {'key': key,
                                  'version': clidata['version']})
                        LOG.warning(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                    if key in CONSTANTS.FJ_QOS_KEY_list:
                        category_dict = self._get_qos_category_by_value(
                            key, value)
                    else:
                        unsupport.append(key)
                if unsupport:
                    LOG.waring('_set_qos, '
                               'Can not support QoS parameter "%s".',
                               unsupport)

            # If storage version is after V11L30.
            if clidata['version'] >= 'V11L30-0000':
                key_dict = self._get_param(qos_specs_dict)
                if not key_dict:
                    return

                # Get total/read/write bandwidth limit.
                category_dict = self._get_qos_category(key_dict,
                                                       remote=remote)

            if category_dict:
                # Set volume qos.
                volumename = self._get_volume_name(volume, use_id=use_id)
                category_dict['volume-name'] = volumename
                rc, errordesc, job = self._exec_eternus_cli(
                    'set_volume_qos',
                    remote=remote,
                    retry=self.retry_times,
                    timeout=self.req_sec,
                    **category_dict)
                if rc != 0:
                    msg = (_('_set_qos, '
                             'set_volume_qos failed. '
                             'Return code: %(rc)lu, '
                             'Error: %(errordesc)s, '
                             'Message: %(job)s.')
                           % {'rc': rc,
                              'errordesc': errordesc,
                              'job': job})
                    LOG.warning(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        except Exception:
            LOG.warning('_set_qos, failed.')

    @staticmethod
    def _get_qos_specs(volume):
        """Get qos specs information from volume information."""
        LOG.debug('_get_qos_specs, volume id: %s.', volume['id'])

        qos_specs_dict = {}
        qos_specs_id = None
        ctxt = None

        volume_type_id = volume.get('volume_type_id')

        if volume_type_id:
            ctxt = context.get_admin_context()
            volume_type = volume_types.get_volume_type(ctxt, volume_type_id)
            qos_specs_id = volume_type.get('qos_specs_id')

        if qos_specs_id:
            qos_specs_dict = (
                qos_specs.get_qos_specs(ctxt, qos_specs_id)['specs'])

        LOG.debug('_get_qos_specs, qos_specs_dict: %s.', qos_specs_dict)
        return qos_specs_dict

    def _is_qos_or_format_support(self, func_name):
        """If storage is DX S2 series, qos or format is not supported."""
        is_support = True

        if self.model_name == CONSTANTS.DX_S2:
            is_support = False
            LOG.warning('%s is not supported for DX S2, '
                        'Skip this process.', func_name)
        return is_support

    @staticmethod
    def _get_qos_category_by_value(key, value):
        """Get qos category using value."""
        LOG.debug('_get_qos_category_by_value, '
                  'key: %(key)s, value: %(value)s.',
                  {'key': key, 'value': value})

        ret = 0

        # Log error method.
        def _get_qos_category_by_value_error():
            """Input value is invalid, log error and raise exception."""
            msg = (_('_get_qos_category_by_value, '
                     'Invalid value is input, '
                     'key: %(key)s, '
                     'value: %(value)s.')
                   % {'key': key,
                      'value': value})
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if key == "maxBWS":
            try:
                digit = int(float(value))
            except Exception:
                _get_qos_category_by_value_error()

            if 800 <= digit:
                ret = 1
            elif 700 <= digit:
                ret = 2
            elif 600 <= digit:
                ret = 3
            elif 500 <= digit:
                ret = 4
            elif 400 <= digit:
                ret = 5
            elif 300 <= digit:
                ret = 6
            elif 200 <= digit:
                ret = 7
            elif 100 <= digit:
                ret = 8
            elif 70 <= digit:
                ret = 9
            elif 40 <= digit:
                ret = 10
            elif 25 <= digit:
                ret = 11
            elif 20 <= digit:
                ret = 12
            elif 15 <= digit:
                ret = 13
            elif 10 <= digit:
                ret = 14
            elif 0 < digit:
                ret = 15
            else:
                _get_qos_category_by_value_error()

        LOG.debug('_get_qos_category_by_value (%s).', ret)

        category_dict = {}
        if ret > 0:
            category_dict = {'bandwidth-limit': ret}

        return category_dict

    def _get_param(self, qos_specs_dict):
        # Get all keys which have been set and its value.
        LOG.debug('_get_param, '
                  'qos_specs_dict: %(qos_specs_dict)s.',
                  {'qos_specs_dict': qos_specs_dict})
        key_dict = {}
        unsupport = []
        for key, value in qos_specs_dict.items():
            if key in CONSTANTS.FJ_QOS_KEY_list:
                msg = (_('_get_param, Can not support QoS '
                         'parameter "%(key)s" on firmware version '
                         'V11L30-0000 or above.')
                       % {'key': key})
                LOG.warning(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if key in CONSTANTS.FJ_QOS_KEY_BYTES_list:
                key_dict[key] = self._check_throughput(key, value)
                # If can not get the corresponding iops,
                # the iops is set to the maximum value.
                iopsStr = key.replace('bytes', 'iops')
                if iopsStr not in qos_specs_dict.keys():
                    key_dict[iopsStr] = CONSTANTS.MAX_IOPS
            elif key in CONSTANTS.FJ_QOS_KEY_IOPS_list:
                key_dict[key] = self._check_iops(key, value)
                # If can not get the corresponding bytes,
                # the bytes is set to the maximum value.
                throughputStr = key.replace('iops', 'bytes')
                if throughputStr not in qos_specs_dict.keys():
                    key_dict[throughputStr] = CONSTANTS.MAX_THROUGHPUT
            else:
                unsupport.append(key)
        if unsupport:
            LOG.warning('_get_param, '
                        'Can not support QoS parameter "%s".', unsupport)

        return key_dict

    def _check_iops(self, key, value):
        """Check input value of IOPS."""
        LOG.debug('_check_iops, key: %(key)s, value: %(value)s.',
                  {'key': key, 'value': value})
        value = int(float(value))
        if value < CONSTANTS.MIN_IOPS or value > CONSTANTS.MAX_IOPS:
            msg = (_('_check_iops, '
                     '%(key)s is out of range.')
                   % {'key': key})
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return int(value)

    def _check_throughput(self, key, value):
        LOG.debug('_check_throughput, key: %(key)s, value: %(value)s.',
                  {'key': key, 'value': value})
        value = float(value) / units.Mi
        if (value < CONSTANTS.MIN_THROUGHPUT or
                value > CONSTANTS.MAX_THROUGHPUT):
            msg = (_('_check_throughput, '
                     '%(key)s is out of range.')
                   % {'key': key})
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return int(value)

    def _get_qos_category(self, key_dict, remote=False):
        """Get qos category by parameters according to the specific volume."""
        LOG.debug('_get_qos_category, '
                  'key_dict: %(key_dict)s.',
                  {'key_dict': key_dict})

        # Get all the bandwidth limits.
        rc, errordesc, bandwidthlist = self._exec_eternus_cli(
            'show_qos_bandwidth_limit',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec)
        if rc != 0:
            msg = (_('_get_qos_category, '
                     'show_qos_bandwidth_limit failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'clidata': bandwidthlist})
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        ret_dict = {}
        for bw in bandwidthlist:
            if 'total_iops_sec' in key_dict.keys():
                if (bw['total_iops_sec'] == key_dict['total_iops_sec'] and
                        bw['total_bytes_sec'] == key_dict['total_bytes_sec']):
                    ret_dict['bandwidth-limit'] = bw['total_limit']
            if 'read_iops_sec' in key_dict.keys():
                if (bw['read_iops_sec'] == key_dict['read_iops_sec'] and
                        bw['read_bytes_sec'] == key_dict['read_bytes_sec']):
                    ret_dict['read-bandwidth-limit'] = bw['read_limit']
            if 'write_iops_sec' in key_dict.keys():
                if (bw['write_iops_sec'] == key_dict['write_iops_sec'] and
                        bw['write_bytes_sec'] == key_dict['write_bytes_sec']):
                    ret_dict['write-bandwidth-limit'] = bw['write_limit']

        # If find all available pairs.
        # len(key_dict) must be 2, 4 or 6
        if len(key_dict) / 2 == len(ret_dict):
            return ret_dict

        rc, errordesc, vqosdatalist = self._exec_eternus_cli(
            'show_volume_qos',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec)
        if rc != 0:
            msg = (_('_get_qos_category, '
                     'show_volume_qos failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'clidata': vqosdatalist})
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Get used total/read/write bandwidth limit.
        totalusedlimits = set()
        readusedlimits = set()
        writeusedlimits = set()
        for vqos in vqosdatalist:
            totalusedlimits.add(vqos['total_limit'])
            readusedlimits.add(vqos['read_limit'])
            writeusedlimits.add(vqos['write_limit'])

        # Get unused total/read/write bandwidth limit.
        totalunusedlimits = list(set(range(1, 16)) - totalusedlimits)
        readunusedlimits = list(set(range(1, 16)) - readusedlimits)
        writeunusedlimits = list(set(range(1, 16)) - writeusedlimits)

        # If there is no same couple, set new qos bandwidth limit.
        if 'total_iops_sec' in key_dict.keys():
            if 'bandwidth-limit' not in ret_dict.keys():
                if len(totalunusedlimits) == 0:
                    msg = _('_get_qos_category, '
                            'There is no available total bandwidth limit.')
                    LOG.warning(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                else:
                    self._set_limit('volume-qos',
                                    totalunusedlimits[0],
                                    key_dict['total_iops_sec'],
                                    key_dict['total_bytes_sec'],
                                    remote=remote)
                    ret_dict['bandwidth-limit'] = totalunusedlimits[0]
        else:
            ret_dict['bandwidth-limit'] = 0

        if 'read_iops_sec' in key_dict.keys():
            if 'read-bandwidth-limit' not in ret_dict.keys():
                if len(readunusedlimits) == 0:
                    msg = _('_get_qos_category, '
                            'There is no available read bandwidth limit.')
                    LOG.warning(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                else:
                    self._set_limit('volume-qos-read',
                                    readunusedlimits[0],
                                    key_dict['read_iops_sec'],
                                    key_dict['read_bytes_sec'],
                                    remote=remote)
                    ret_dict['read-bandwidth-limit'] = readunusedlimits[0]
        else:
            ret_dict['read-bandwidth-limit'] = 0

        if 'write_bytes_sec' in key_dict.keys():
            if 'write-bandwidth-limit' not in ret_dict.keys():
                if len(writeunusedlimits) == 0:
                    msg = _('_get_qos_category, '
                            'There is no available write bandwidth limit.')
                    LOG.warning(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                else:
                    self._set_limit('volume-qos-write',
                                    writeunusedlimits[0],
                                    key_dict['write_iops_sec'],
                                    key_dict['write_bytes_sec'],
                                    remote=remote)
                    ret_dict['write-bandwidth-limit'] = writeunusedlimits[0]
        else:
            ret_dict['write-bandwidth-limit'] = 0

        return ret_dict

    def _set_limit(self, mode, limit, iops, throughput, remote=False):
        LOG.debug('_set_limit, mode: %(mode)s, '
                  'limit: %(limit)s, iops:%(iops)s, '
                  'throughput: %(throughput)s.',
                  {'mode': mode, 'limit': limit,
                   'iops': iops, 'throughput': throughput})
        param_dict = ({'mode': mode,
                       'bandwidth-limit': limit,
                       'iops': iops,
                       'throughput': throughput})

        rc, emsg, clidata = self._exec_eternus_cli(
            'set_qos_bandwidth_limit',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
            **param_dict)

        if rc != 0:
            msg = (_('_set_limit, '
                     'set_qos_bandwidth_limit failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(errormsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'errormsg': emsg,
                      'clidata': clidata})
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    @staticmethod
    def _get_bool(value):
        """Get True/False from String 'True'/'False'."""
        LOG.debug('_get_bool, value: %(value)s.',
                  {'value': value})

        ret = False

        if str(value).lower() == "true":
            ret = True

        return ret

    @dx_utils.FJDXTimeout()
    def failover_host(self, volumes, secondary_id=None,
                      groups=None, timeout=None):
        """Fails over/back the volumes on a host back and forth."""
        LOG.debug('Failover host start, '
                  'failover: %(failover)s, '
                  'secondary_id: %(secondary_id)s.',
                  {'failover': self.failover, 'secondary_id': secondary_id})

        if secondary_id != 'default':
            LOG.debug('Failover host start.')
            if not self.failover:
                if not secondary_id:
                    secondary_id = self.remote_backend_name
                elif self.remote_backend_name != secondary_id:
                    msg = (_('The backend %(backend)s is not '
                             'the right backend.')
                           % {'backend': secondary_id})
                    LOG.error(msg)
                    raise exception.InvalidReplicationTarget(data=msg)
            else:
                msg = (_("Backend %(backend)s is already failed over. "
                         "If you wish to failback, please append "
                         "'--backend_id default' to your command.")
                       % {'backend': self.remote_backend_name})
                LOG.error(msg)
                raise exception.InvalidReplicationTarget(data=msg)
            # Failover volumes to secondary backend
            volume_update_list = self._failover(volumes, secondary_id, groups)
        else:
            LOG.debug('Failback host start.')
            if self.failover:
                secondary_id = None
            else:
                msg = (_("Cannot failback from backend %(backend)s- "
                         "backend not in failed over state. "
                         "If you meant to failover, please "
                         "omit the '--backend_id default' from the command")
                       % {'backend': self.remote_backend_name})
                LOG.error(msg)
                raise exception.InvalidReplicationTarget(data=msg)
            # Failback volumes from secondary backend
            volume_update_list = self._failback(volumes, secondary_id, groups)

        LOG.debug('Failover host complete. ')
        return secondary_id, volume_update_list, []

    def _failover(self, volumes, secondary_id=None, groups=None):
        """Failover volumes to secondary backend."""
        LOG.debug('Failover start, '
                  'secondary_id: %s.', secondary_id)

        success_tfo_ports = []
        volume_update_list = []
        ready_tfog_no = self._get_failover_ready_tfog()
        for tfog_no in ready_tfog_no.keys():
            param_dict = {'tfog-number': tfog_no,
                          'active-mode': 'manual-failover'}
            rc, errordesc, job = self._exec_eternus_cli(
                'forced_tfo_group_activate',
                remote=True,
                timeout=self.req_sec,
                **param_dict)
            if rc != 0:
                msg = (_('_failover, execute Eternus cli: '
                         'forced_tfo_group_activate failed, '
                         'tfo group number: %(tfog_no)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Message: %(job)s.')
                       % {'tfog_no': tfog_no,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})
                LOG.error(msg)
            else:
                success_tfo_ports.append(ready_tfog_no[tfog_no])

        if not success_tfo_ports:
            msg = (_('_failover, execute Eternus cli: '
                     'forced_tfo_group_activate all failed.'))
            LOG.error(msg)
            raise exception.UnableToFailOver(data=msg)

        LOG.debug('_failover success_tfo_ports=%s.',
                  str(success_tfo_ports))
        for volume in volumes:
            if self._is_replicated(volume):
                model_update = self._failover_volume(volume,
                                                     success_tfo_ports)
            else:
                # if volume is snapshot, the follows are wrong.
                model_update = {
                    'volume_id': volume.id,
                    'updates': {'status': 'error',
                                'previous_status': volume['status'],
                                'replication_status': 'not-capable'}}
            volume_update_list.append(model_update)

        self.failover = True
        self.filename = self.remote_eternus_config
        self.remote_filename = self.configuration.cinder_eternus_config_file
        self.local_ce_support, self.remote_ce_support = (
            self.remote_ce_support, self.local_ce_support)

        LOG.debug('_failover volume_update_list=%s.',
                  str(volume_update_list))
        return volume_update_list

    def _failover_volume(self, volume, success_tfo_ports):
        """Failover one volume."""
        LOG.debug('Failover one volume start, '
                  'success_tfo_ports: %(success_tfo_ports)s, '
                  'volume: %(volume)s.',
                  {'success_tfo_ports': str(success_tfo_ports),
                   'volume': volume['id']})

        if volume['status'] == 'in-use':
            used_ports = eval(volume['replication_extended_status'])
            LOG.debug('_failover_volume, used_ports=%s', used_ports)
            for port in used_ports:
                for success_tfo_port in success_tfo_ports:
                    if port in success_tfo_port:
                        local_provider, remote_data = (
                            self._swap_provider_location(volume))
                        model_update = {
                            'volume_id': volume['id'],
                            'updates': {
                                'previous_status': volume['status'],
                                'replication_status': 'failed-over',
                                'replication_driver_data': str(remote_data),
                                'provider_location': str(local_provider)}}
                        return model_update

        model_update = {'volume_id': volume['id'],
                        'updates': {'status': 'error',
                                    'previous_status': volume['status'],
                                    'replication_status': 'failover-error'}}

        LOG.debug('Failover one volume end, '
                  'model_update=%(model_update)s, '
                  'volume: %(volume).',
                  {'model_update': str(model_update),
                   'volume': volume.id})
        return model_update

    @staticmethod
    def _swap_provider_location(volume):
        """Swap remote provider location and local provider location."""
        LOG.debug('_swap_provider_location start, '
                  'volume: %s.', volume['id'])

        local_provider = volume['provider_location']
        remote_data = eval(volume['replication_driver_data'])
        remote_provider = remote_data['provider_location']

        local_provider, remote_provider = remote_provider, local_provider
        remote_data['provider_location'] = remote_provider

        LOG.debug('_swap_provider_location end, volume: %s.', volume['id'])
        return local_provider, remote_data

    def _get_failover_ready_tfog(self):
        """Get tfo groups which are ready to failover from secondary."""
        LOG.debug('_get_failback_ready_tfog start, '
                  'remote_filename=%(remote_filename)s, '
                  'filename=%(filename)s.',
                  {'remote_filename': self.remote_filename,
                   'filename': self.filename})

        ready_tfog_no = {}
        total_tfog_list = self._show_tfo_groups(remote=True)

        LOG.debug("_get_failover_ready_tfog=%s.", str(total_tfog_list))
        for tfog in total_tfog_list:
            if (tfog['type'] == '02' and  # Secondary
                    tfog['name'] == self.tfog_name and
                    tfog['pair_boxid'] == self.local_boxid):
                if tfog['phase'] == '07':  # Normal
                    tfog_port_list = tfog['tfog_port_list']
                    port_list = []
                    for tfo_port in tfog_port_list:
                        port_list.append(tfo_port['pair_port'])
                    ready_tfog_no[tfog['no']] = port_list
                else:
                    msg = (_("Can not failover tfo group: "
                             "tfo group name: %(name)s, "
                             "tfo group status: %(phase)s. ")
                           % {'name': tfog['name'],
                              'phase': tfog['phase']})
                    LOG.warning(msg)

        if not ready_tfog_no:
            msg = (_("no normal tfo group to failover."))
            LOG.error(msg)
            raise exception.UnableToFailOver(data=msg)

        LOG.debug('_get_failover_ready_tfog end.')
        return ready_tfog_no

    def _failback(self, volumes, secondary_id=None, groups=None):
        """Failback volumes to primary backend."""
        LOG.debug('Failback start, secondary_id: %s.', secondary_id)

        success_tfo_ports = []
        volume_update_list = []
        ready_tfog_no = self._get_failback_ready_tfog()
        for tfog_no in ready_tfog_no.keys():
            param_dict = {'tfog-number': tfog_no,
                          'active-mode': 'manual-failback'}
            rc, errordesc, job = \
                self._exec_eternus_cli('forced_tfo_group_activate',
                                       remote=True,
                                       timeout=self.req_sec,
                                       **param_dict)
            if rc != 0:
                msg = (_('_failback, execute Eternus cli: '
                         'forced_tfo_group_activate error, '
                         'tfo group number: %(tfog_no)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Message: %(job)s. ')
                       % {'tfog_no': tfog_no,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})
                LOG.warning(msg)
            else:
                success_tfo_ports.append(ready_tfog_no[tfog_no])

        if not success_tfo_ports:
            msg = (_('_failback, execute Eternus cli:'
                     'forced_tfo_group_activate all failed.'))
            LOG.error(msg)
            raise exception.UnableToFailOver(data=msg)

        for volume in volumes:
            if self._is_replicated(volume):
                model_update = self._failback_volume(volume,
                                                     success_tfo_ports)
            else:
                # if volume is snapshot, the follows are wrong.
                model_update = {
                    'volume_id': volume['id'],
                    'updates': {'status': volume['previous_status'],
                                'previous_status': '',
                                'replication_status': ''}}
            volume_update_list.append(model_update)

        self.filename = self.configuration.cinder_eternus_config_file
        self.remote_filename = self.remote_eternus_config
        self.failover = False
        self.local_ce_support, self.remote_ce_support = (
            self.remote_ce_support, self.local_ce_support)

        LOG.debug('Failback end, volume_update_list=%s.',
                  str(volume_update_list))
        return volume_update_list

    def _failback_volume(self, volume, success_tfo_ports):
        """Failback one volume."""
        LOG.debug('Failback one volume start, '
                  'success_tfo_ports=%(success_tfo_ports)s, '
                  'volume: %(volume)s.',
                  {'success_tfo_ports': str(success_tfo_ports),
                   'volume': volume.id})

        if volume['status'] == 'error':
            pass
        elif volume['status'] == 'in-use':
            used_ports = eval(volume['replication_extended_status'])
            for port in used_ports:
                for success_tfo_port in success_tfo_ports:
                    if port in success_tfo_port:
                        local_provider, remote_data = \
                            self._swap_provider_location(volume)
                        model_update = {
                            'volume_id': volume['id'],
                            'updates': {
                                'previous_status': '',
                                'replication_status': 'enabled',
                                'replication_driver_data': str(remote_data),
                                'provider_location': str(local_provider)}}
                        return model_update

        model_update = {'volume_id': volume.id,
                        'updates': {'status': volume['previous_status'],
                                    'replication_status': 'enabled'}}

        LOG.debug('Failback one volume end. '
                  'model_update=%(model_update)s, '
                  'volume: %(volume).',
                  {'model_update': str(model_update),
                   'volume': volume.id})
        return model_update

    def _get_failback_ready_tfog(self):
        """Get tfo groups which are ready to failback."""
        ready_tfog_no = {}
        LOG.debug('_get_failback_ready_tfog start, '
                  'remote_filename=%(remote_filename)s, '
                  'filename=%(filename)s.',
                  {'remote_filename': self.remote_filename,
                   'filename': self.filename})

        total_tfog_list = self._show_tfo_groups(remote=True)
        LOG.debug('_get_failback_ready_tfog, tfog: %s.', str(total_tfog_list))
        for tfog in total_tfog_list:
            if (tfog['type'] == '01' and  # Primary
                    tfog['status'] == '00' and  # Standby
                    tfog['name'] == self.tfog_name):
                if tfog['phase'] == '05':  # failback ready
                    tfog_port_list = tfog['tfog_port_list']
                    port_list = []
                    for tfo_port in tfog_port_list:
                        port_list.append(tfo_port['own_port'])
                    ready_tfog_no[tfog['no']] = port_list
                else:
                    msg = (_("Can not failback tfo group: "
                             "tfo group name: %(name)s, "
                             "tfo group status: %(phase)s. ")
                           % {'name': tfog['name'],
                              'phase': tfog['phase']})
                    LOG.warning(msg)

        if not ready_tfog_no:
            msg = (_("no tfo group is ready to failback."))
            LOG.error(msg)
            raise exception.UnableToFailOver(data=msg)

        LOG.debug('_get_failback_ready_tfog end.')
        return ready_tfog_no

    def _format_standard_volume(self, volume, remote=False):
        """Format standard volume."""
        LOG.debug('_format_standard_volume, volume id: %(vid)s, '
                  'remote: %(remote)s.',
                  {'vid': volume['id'], 'remote': remote})

        wait_sec = 60
        wait_cnt = 0
        filename = None

        volumename = self._get_volume_name(volume)
        param_dict = {
            'volume-name': volumename
        }

        if remote:
            filename = self.remote_eternus_config

        rc, errordesc, job = self._exec_eternus_cli('format_volume',
                                                    retry=self.retry_times,
                                                    filename=filename,
                                                    timeout=self.req_sec,
                                                    **param_dict)

        if rc != 0:
            msg = (_('_format_standard_volume, '
                     'volumename: %(volumename)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(job)s.')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        while True:
            time.sleep(wait_sec)
            rc, errordesc, job = self._exec_eternus_cli('show_volume_progress',
                                                        retry=self.retry_times,
                                                        filename=filename,
                                                        timeout=self.req_sec,
                                                        **param_dict)

            if rc != 0:
                msg = (_('_format_standard_volume, '
                         'show volume progress error, '
                         'volumename: %(volumename)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Message: %(job)s.')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            wait_sec = int(job)

            if wait_sec == 0:
                LOG.debug('_format_standard_volume, format complete.')
                break
            elif wait_cnt < 3:
                wait_sec /= 3
                wait_cnt += 1

            LOG.debug('_format_standard_volume, format remain time: %s sec.',
                      job)

    def _format_tpv(self, volume, vol_instance, poolname=None, remote=False):
        """Format tpv."""
        LOG.debug('_format_tpv, volume id: %(vid)s, remote: %(remote)s.',
                  {'vid': volume['id'], 'remote': remote})

        filename = None
        volumename = self._get_volume_name(volume)

        if remote:
            filename = self.remote_eternus_config

        if not poolname:
            poolname, _x = self._find_pool_from_volume(vol_instance,
                                                       remote=remote)

        lockname = 'format-' + poolname + six.text_type(remote)

        @dx_utils.FJDXLockutils(lockname, 'cinder-', True)
        @dx_utils.FJDXTimeout()
        def __format_tpv(self, volumename, poolname, filename, timeout=None):
            """Do format tpv using CLI method."""
            wait_sec = 60
            wait_cnt = 0

            format_param_dict = {'volume-name': volumename}

            rc, errordesc, job = self._exec_eternus_cli(
                'format_tpv',
                retry=self.retry_times,
                filename=filename,
                timeout=self.req_sec,
                **format_param_dict)

            if rc != 0:
                msg = (_('_format_tpv, '
                         'volumename: %(volumename)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Message: %(job)s.')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            show_param_dict = {
                'pool-name': poolname
            }
            while True:
                time.sleep(wait_sec)
                rc, errordesc, job = self._exec_eternus_cli(
                    'show_tpv_progress',
                    retry=self.retry_times,
                    filename=filename,
                    timeout=self.req_sec,
                    **show_param_dict)

                if rc != 0:
                    msg = (_('_format_tpv, '
                             'show volume progress error, '
                             'volumename: %(volumename)s, '
                             'Return code: %(rc)lu, '
                             'Error: %(errordesc)s, '
                             'Message: %(job)s.')
                           % {'volumename': volumename,
                              'rc': rc,
                              'errordesc': errordesc,
                              'job': job})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                wait_sec = int(job)

                if wait_sec == 0:
                    LOG.debug('_format_tpv, format complete.')
                    break
                elif wait_cnt < 3:
                    wait_sec /= 3
                    wait_cnt += 1

                LOG.debug('_format_tpv, format wait: %s sec.', job)

        __format_tpv(self, volumename, poolname,
                     filename, timeout=self.req_sec)

    def _delete_volume_base(self, volume, with_format, remote=False):
        LOG.debug('_delete_volume_base, remote: %(remote)s.',
                  {'remote': remote})

        if not remote:
            conn = self.conn
            vol_instance = self._find_lun(volume)
        else:
            conn = self.remote_conn
            vol_instance = self._find_replica_lun(volume)

        if not vol_instance:
            return True

        if self._get_bool(with_format):
            format_support = self._is_qos_or_format_support('Format')
            if format_support:
                poolname, pool = self._find_pool_from_volume(vol_instance,
                                                             remote=remote)
                if not pool:
                    msg = (_('_delete_volume_base, '
                             'vol_instance: %(vol_instance)s, '
                             'pool not found.')
                           % {'vol_instance': vol_instance.path})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                if 'RSP' in pool['InstanceID']:  # Pooltype == RAIDGROUP.
                    self._format_standard_volume(volume, remote=remote)
                else:  # Pooltype == TPPOOL.
                    self._format_tpv(
                        volume, vol_instance, poolname, remote=remote)

        if vol_instance:
            self._delete_volume(volume,
                                conn=conn,
                                remote=remote,
                                timeout=self.req_sec)

        return True

    @dx_utils.FJDXTimeout()
    def _get_manageable_resource_info(self, cinder_resources, resource_type,
                                      marker, limit, offset, sort_keys,
                                      sort_dirs, remote=False, timeout=None):
        """Gets the resources on the backend available for management"""
        """by Cinder."""
        LOG.debug('_get_manageable_resource_info, resource type: %s.',
                  resource_type)

        if remote:
            filename = self.remote_eternus_config
        else:
            filename = self.configuration.cinder_eternus_config_file

        self.conn = self._get_eternus_connection(filename)

        if resource_type == 'volume':
            poolname_list = self._get_drvcfg('EternusPool',
                                             filename=filename,
                                             multiple=True)
        elif resource_type == 'snapshot':
            poolname_list = self._get_drvcfg('EternusSnapPool',
                                             filename=filename,
                                             multiple=True)

        backend_resources = self._find_pools_resources(poolname_list,
                                                       self.conn, filename)

        LOG.debug('_get_manageable_resource_info, pool names: %s.',
                  poolname_list)

        # cinder_current_dict: {'vol_name': {'id': cinder_resource['id'],
        #                       'volume_id': cinder_resource['volume_id']}}
        cinder_current_dict = {}
        for cinder_resource in cinder_resources:
            vol_name = self._get_volume_name(cinder_resource)
            id_dict = {'id': cinder_resource.get('id')}
            if resource_type == 'snapshot':
                id_dict['volume_id'] = cinder_resource.get('volume_id', None)

            cinder_current_dict[vol_name] = id_dict

        manageable_resources = self._decide_manageable(
            resource_type,
            backend_resources,
            cinder_current_dict=cinder_current_dict)

        LOG.debug('_get_manageable_resource_info, manageable resources: %s.',
                  manageable_resources)
        return volume_utils.paginate_entries_list(manageable_resources,
                                                  marker, limit, offset,
                                                  sort_keys, sort_dirs)

    def _find_copysession_by_path(self, vol_instance_path, conn=None):
        """Find copy session on ETERNUS by volume instance path."""
        LOG.debug('_find_copysession_by_path, volume instance path: %s.',
                  vol_instance_path)

        if not conn:
            conn = self.conn

        try:
            cpsessionlist = conn.ReferenceNames(
                vol_instance_path,
                ResultClass='FUJITSU_StorageSynchronized')
        except Exception:
            msg = (_('_find_copysession_by_path, '
                     'ReferenceNames, '
                     'vol_instance_path: %(vol_instance_path)s, '
                     'Cannot connect to ETERNUS.')
                   % {'vol_instance_path': vol_instance_path})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_copysession_by_path, cpsession list: %s.',
                  cpsessionlist)
        return cpsessionlist

    def _get_attached_volumes(self, conn=None):
        """get attached volumes from ETERNUS."""
        LOG.debug('_get_attached_volumes, start.')

        affinity_grouplist = self._find_affinity_group_resources(conn)
        volume_list = []
        for ag in affinity_grouplist:
            try:
                volumes = self._assoc_eternus(
                    ag,
                    conn=conn,
                    AssocClass='FUJITSU_ProtocolControllerForUnit',
                    ResultClass='FUJITSU_StorageVolume')
            except Exception:
                msg = _('_get_attached_volumes, '
                        'cannot connect to ETERNUS.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            volume_list.extend(volumes)

        attached_volname_list = \
            [volume['ElementName'] for volume in volume_list]

        LOG.debug('_get_attached_volumes, '
                  'attached volumes name list: %(attached_volname_list)s.',
                  {'attached_volname_list': attached_volname_list})
        return attached_volname_list

    def _find_affinity_group_resources(self, conn=None):
        """Find affinity group from ETERNUS."""
        LOG.debug('_find_affinity_group_resources, start.')

        affinity_grouplist = []

        try:
            aglist = self._enum_eternus_instance_names(
                'FUJITSU_AffinityGroupController', conn=conn)
        except Exception:
            msg = _('_find_affinity_group_resources, '
                    'cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for ag in aglist:
            try:
                hostaglist = self._assoc_eternus(
                    ag,
                    conn=conn,
                    AssocClass='CIM_AuthorizedTarget',
                    ResultClass='FUJITSU_AuthorizedPrivilege')
            except Exception:
                msg = _('_find_affinity_group_resources, '
                        'Associators: CIM_AuthorizedTarget, '
                        'cannot connect to ETERNUS.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if len(hostaglist) > 0:
                affinity_grouplist.append(ag)

        LOG.debug('_find_affinity_group_resources, '
                  'affinity group list: %(affinity_group)s.',
                  {'affinity_group': affinity_grouplist})
        return affinity_grouplist

    @dx_utils.FJDXTimeout()
    def _revert_to_snapshot(self, volume, snapshot, timeout=None):
        """Revert volume to snapshot."""

        LOG.debug('_revert_to_snapshot, Enter method, '
                  'volume id: %(vid)s, '
                  'snapshot id: %(sid)s. ',
                  {'vid': volume['id'], 'sid': snapshot['id']})

        self.conn = self._get_eternus_connection()

        vol_instance = self._find_lun(volume)
        sdv_instance = self._find_lun(snapshot)

        # Check the existence of volume.
        if not vol_instance:
            msg = (_('_revert_to_snapshot, '
                     'source volume not found on ETERNUS, '
                     'volume: %(volume)s. ')
                   % {'volume': volume})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Check the existence of sdv.
        if not sdv_instance:
            msg = (_('_revert_to_snapshot, '
                     'snapshot volume not found on ETERNUS. '
                     'snapshot: %(snapshot)s. ')
                   % {'snapshot': snapshot})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        checkstart_time = time.time()
        while self._check_if_cascade_copy_volume(snapshot):
            checkend_time = time.time()
            remaining_time = (checkstart_time +
                              self.copy_wait_time) - checkend_time
            if remaining_time <= 0:
                msg = (_('_revert_to_snapshot, '
                         'The source snapshot is cascade copy '
                         'destination volume. '
                         'check start time: %(st)s, '
                         'check end time: %(et)s, '
                         'remaining time: %(rt)s, '
                         'wait time: %(wt)s.')
                       % {'st': checkstart_time,
                          'et': checkend_time,
                          'rt': remaining_time,
                          'wt': self.copy_wait_time})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                wait_time = remaining_time if (
                            remaining_time < CONSTANTS.CHECK_COPY_WAIT_TIME
                ) else CONSTANTS.CHECK_COPY_WAIT_TIME
                LOG.debug('create_cloned_volume, '
                          'check start time: %(st)s, '
                          'check end time: %(et)s, '
                          'remaining time: %(rt)s, '
                          'wait time: %(wt)s.',
                          {'st': checkstart_time,
                           'et': checkend_time,
                           'rt': remaining_time,
                           'wt': self.copy_wait_time})
                time.sleep(wait_time)

        sdvsession = None
        cpsessionlist = self._find_copysession(vol_instance)

        LOG.debug('_revert_to_snapshot, '
                  'cpsessionlist: %(cpsessionlist)s. ',
                  {'cpsessionlist': cpsessionlist})

        for cpsession in cpsessionlist:
            if (cpsession['SystemElement'].keybindings.get('DeviceID') ==
                    vol_instance.path.keybindings.get('DeviceID')):
                if (cpsession['SyncedElement'].keybindings.get('DeviceID') ==
                        sdv_instance.path.keybindings.get('DeviceID')):
                    sdvsession = cpsession
                    break

        if not sdvsession:
            msg = (_('_revert_to_snapshot, '
                     'snapshot volume not found on ETERNUS. '
                     'snapshot: %(snapshot)s. ')
                   % {'snapshot': snapshot})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_revert_to_snapshot, '
                  'sdvsession: %(sdvsession)s. ',
                  {'sdvsession': sdvsession})

        repservice = self._find_eternus_service("FUJITSU_ReplicationService")

        if repservice is None:
            msg = _('_revert_to_snapshot, '
                    'Replication Service not found. ')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Invoke method for restore snapshot
        rc, errordesc, job = self._exec_eternus_service(
            'ModifyReplicaSynchronization',
            repservice,
            Operation=self._pywbem_uint(15, '16'),
            WaitForCopyState=self._pywbem_uint(8, '16'),
            Synchronization=sdvsession)

        if rc != 0:
            msg = (_('_revert_to_snapshot, '
                     '_exec_eternus_service error, '
                     'volume: %(vid)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(job)s.')
                   % {'vid': volume['id'],
                      'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:
            LOG.debug('_revert_to_snapshot, '
                      'successfully. ')
        LOG.debug('_revert_to_snapshot, Exit method. ')

    def _get_snapshot_src(self, vol_instance, cps_list):
        """Return source volume number if vol_instance is a snapshot."""

        vol_no = int(vol_instance['num'], 16)
        for cp in cps_list:
            if cp['Dest Num'] == vol_no and cp['Type'] == 'Snap':
                return cp['Source Num']
            elif cp['Dest Num'] == vol_no and cp['Type'] == 'Snap+':
                # Give up the snap+ type.
                return -1

        return None

    def _get_copy_sessions_list(self, remote=False):
        """Get copy sessions list."""
        LOG.debug('_get_copy_sessions_list, Enter method, '
                  'remote: %(remote)s. ',
                  {'remote': remote})

        rc, emsg, clidata = self._exec_eternus_cli(
            'show_copy_sessions',
            remote=remote,
            retry=self.retry_times,
            timeout=self.req_sec,
        )

        if rc != 0:
            msg = (_('_get_copy_sessions_list, '
                     'get copy sessions failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(emsg)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'emsg': emsg,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        LOG.debug('_get_copy_sessions_list, Exit method, '
                  'copy sessions list: %(clidata)s. ',
                  {'clidata': clidata})

        return clidata

    def _decide_manageable(self, resource_type, volumes, **kwargs):
        LOG.debug('_decide_manageable, volumes: %s.', len(volumes))

        if not self.conn:
            self.conn = self._get_eternus_connection()

        # cinder_current_dict: {'vol_name': {'id': cinder_resource['id'],
        #                       'volume_id': cinder_resource['volume_id']}}
        cinder_current_dict = {}
        # attached_volumes[volumename,...]
        attached_volumes = []
        # cinder_src_info: {volume_no: {'id': cinder_resource['id'],
        #                   'name': volume_name}
        cinder_src_info = {}

        if resource_type == 'volume':
            if self.model_name != CONSTANTS.DX_S2:
                attached_volumes = self._get_attached_volumes(self.conn)
            else:
                LOG.warning('_decide_manageable, attached_volumes '
                            'judgment does not supported on DX S2.')
                attached_volumes = []
        elif resource_type == 'snapshot':
            ctxt = context.get_admin_context()
            Volume_Manager = volume_manager.VolumeManager()
            cinder_resources = Volume_Manager._get_my_volumes(ctxt)
            snp_host = self.host.split('@', 1)[1].split('#')[0]

            for cinder_volume in cinder_resources:
                vol_name = self._get_volume_name(cinder_volume)
                vol_host = cinder_volume['host'].split('@', 1)[1].split('#')[0]

                if (cinder_volume['status'] == 'available' and
                        snp_host == vol_host):
                    vol_info = eval(cinder_volume['provider_location'])
                    if ('keybindings' in vol_info and
                            'DeviceID' in vol_info['keybindings']):
                        num = self._get_volume_no(vol_info['keybindings'])
                    else:
                        num = None
                    vol_dic = {'id': cinder_volume.get('id'), 'name': vol_name}
                    num = int(num, 16)
                    cinder_src_info[num] = vol_dic

        bk_resources = []
        for volume in volumes:
            size_byte = volume['BlockSize'] * \
                        volume['NumberOfBlocks']
            vsize = (size_byte + units.Gi - 1) // units.Gi
            vtype = int(volume['Purpose'][8:10])
            vnum = self._get_volume_no(volume)
            vol_info = {'size': vsize,
                        'type': vtype,
                        'name': volume['ElementName'],
                        'num': vnum,
                        'OtherUsageDescription': '',
                        'IsCompressed': 'False',
                        'IsDeduplicated': 'False'
                        }
            # S2 storage does not support these parameters.
            if self.model_name != CONSTANTS.DX_S2:
                vol_info['OtherUsageDescription'] = \
                    volume['OtherUsageDescription']
                vol_info['IsCompressed'] = volume['IsCompressed']
                vol_info['IsDeduplicated'] = volume['IsDeduplicated']

            bk_resources.append(vol_info)

        if 'cinder_current_dict' in kwargs:
            cinder_current_dict = kwargs['cinder_current_dict']
        else:
            ctxt = context.get_admin_context()
            Volume_Manager = volume_manager.VolumeManager()
            cinder_resources = []
            if resource_type == 'volume':
                cinder_resources = Volume_Manager._get_my_volumes(ctxt)
            elif resource_type == 'snapshot':
                cinder_resources = Volume_Manager._get_my_snapshots(ctxt)

            for cinder_resource in cinder_resources:
                vol_name = self._get_volume_name(cinder_resource)
                id_dict = {'id': cinder_resource.get('id')}
                if resource_type == 'snapshot':
                    id_dict['volume_id'] = cinder_resource.get('volume_id')

                cinder_current_dict[vol_name] = id_dict

        cp_session_list = self._get_copy_sessions_list()

        LOG.debug('_decide_manageable, cinder current dict: %(a)s,'
                  ' cinder src info: %(b)s, copy session list: %(c)s.',
                  {'a': cinder_current_dict,
                   'b': cinder_src_info,
                   'c': cp_session_list})

        manageable_resources = []
        for bk_resource in bk_resources:
            if resource_type == 'volume':
                if not (bk_resource['type'] == CONSTANTS.TYPE_OLU or
                        bk_resource['type'] == CONSTANTS.TYPE_TPPC):
                    continue
                # Skip NAS volume.
                if (bk_resource['type'] == CONSTANTS.TYPE_TPPC and
                        bk_resource['OtherUsageDescription'] != ''):
                    continue
                # Skip Dedup volume.
                if (bk_resource['IsCompressed'] == 'True' or
                        bk_resource['IsDeduplicated'] == 'True'):
                    continue

                if self._get_snapshot_src(bk_resource, cp_session_list):
                    continue
                src_num = int(bk_resource['num'], 16)
                resource_info = {'reference':
                                 {'source-name': bk_resource['name']},
                                 'size': bk_resource['size'],
                                 'cinder_id': None,
                                 'extra_info': {'source-id': src_num}}

                if bk_resource['name'] in cinder_current_dict:
                    resource_info['safe_to_manage'] = False
                    resource_info['reason_not_safe'] = 'Already managed.'
                    resource_info['cinder_id'] = \
                        cinder_current_dict[bk_resource['name']]['id']
                elif bk_resource['name'] in attached_volumes:
                    resource_info['safe_to_manage'] = False
                    resource_info['reason_not_safe'] = \
                        'The volume was attached.'
                else:
                    resource_info['safe_to_manage'] = True
                    resource_info['reason_not_safe'] = None

                manageable_resources.append(resource_info)

            if resource_type == 'snapshot':
                if not (bk_resource['type'] == CONSTANTS.TYPE_SDV or
                        bk_resource['type'] == CONSTANTS.TYPE_TPPC):
                    continue
                src_num = self._get_snapshot_src(bk_resource, cp_session_list)
                if not src_num or src_num == -1:
                    # Abandon non-snap type and snap+ type.
                    continue

                resource_info = {'reference':
                                 {'source-name': bk_resource['name']},
                                 'size': bk_resource['size'],
                                 'cinder_id': None,
                                 'extra_info': None}

                if bk_resource['name'] in cinder_current_dict:
                    resource_info['safe_to_manage'] = False
                    resource_info['reason_not_safe'] = 'Already managed.'
                    resource_info['cinder_id'] = \
                        cinder_current_dict[bk_resource['name']]['id']
                    src_cinder_id = \
                        cinder_current_dict[bk_resource['name']]['volume_id']
                    resource_info['source_reference'] = \
                        {'source-volume-id': src_cinder_id}
                    src_name = None
                    for volinfo in cinder_src_info.values():
                        if volinfo['id'] == src_cinder_id:
                            src_name = volinfo['name']
                    a = 'source_reference'
                    b = 'source-volume-name'
                    resource_info[a][b] = src_name
                else:
                    if src_num in cinder_src_info:
                        resource_info['safe_to_manage'] = True
                        resource_info['reason_not_safe'] = None
                        resource_info['source_reference'] = {
                            'source-volume-name':
                                cinder_src_info[src_num]['name'],
                            'source-volume-id':
                                cinder_src_info[src_num]['id']
                        }
                        # Only judged during manage.
                        # if the specified cinder id is correct.
                        if 'volume_info' in kwargs:
                            volume_info = kwargs['volume_info']
                            src_vid = volume_info['volume_id']
                            if src_vid != cinder_src_info[src_num]['id']:
                                resource_info['safe_to_manage'] = False
                                resource_info['reason_not_safe'] = \
                                    'Source volume is incorrect.'
                    else:
                        resource_info['safe_to_manage'] = False
                        resource_info['source_reference'] = \
                            {'sourcr-volume-num': src_num}
                        resource_info['reason_not_safe'] = \
                            'Source volume is not in this host.'

                manageable_resources.append(resource_info)

        return manageable_resources

    def _check_if_cascade_copy_volume(self, volume):
        """Check if the source volume is a secondary copy volume"""
        LOG.debug('_check_if_cascade_copy_volume, source volume id: %(sid)s.',
                  {'sid': volume['id']})

        cascade_copy_volume = False

        location = eval(volume['provider_location'])
        vnum = self._get_volume_no(location['keybindings'])
        volume_no = int(vnum, 16)
        LOG.debug('_check_if_cascade_copy_volume, source volume num: %(num)s.',
                  {'num': volume_no})

        cp_session_list = self._get_copy_sessions_list()

        for cp in cp_session_list:
            if cp['Dest Num'] == volume_no:
                svolume_no = cp['Source Num']
                for scp in cp_session_list:
                    if scp['Dest Num'] == svolume_no:
                        cascade_copy_volume = True
                        LOG.debug('_check_if_cascade_copy_volume, '
                                  'It is a cascade copy volume. '
                                  'volume id: %(sid)s, num: %(num)s.',
                                  {'sid': volume['id'],
                                   'num': volume_no})
                        break

        return cascade_copy_volume

    def _get_eternus_systemname(self):
        if not self.conn:
            self.conn = self._get_eternus_connection(self.filename)

        try:
            propertylist = ['SystemName']
            ret = self._enum_eternus_instances(
                'FUJITSU_StorageHardwareIDManagementService',
                conn=self.conn,
                PropertyList=propertylist)
        except Exception:
            msg = _('_get_eternus_systemname, EnumerateInstances, '
                    'cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        eternus_systemname = ret[0]['SystemName']

        LOG.debug('_get_eternus_systemname, '
                  'SystemName: %s.', eternus_systemname)

        return eternus_systemname
